#!/usr/bin/perl -w

# Copyright 2019, 2020 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.010;
use strict;
use warnings;
use File::Spec;
use List::Util 'max','min','sum','uniqstr';
use Math::OEIS::Stripped;
use Math::PlanePath::GrayCode;
use Math::PlanePath::Base::Digits
  'bit_split_lowtohigh','digit_join_lowtohigh';
use POSIX 'ceil';

use FindBin;
use lib::abs "$FindBin::Bin/../tools";
use MyFLAT;
$|=1;

# uncomment this to run the ### lines
# use Smart::Comments;


# Kirby A. Baker, George F. McNulty, Walter Taylor, "Growth Problems For
# Avoidable Words", Theoretical Computer Science, volume 69, number 3, 1989,
# pages 319-345.
#
# Gamma+ non-empty 0,1,2,3
#
# h5,3  m=5 z=3 b=2 n=10
# i    dz..d0    cyclic LtoH    e[0]..e[m-1]
# 0     0000       0000 0         01234
# 1     0001       1000 1         51239
# 2     0010       0100 0         06234
# index i in base b=ceil(m^(1/z))
# e[j] = m*d[j] + j
# d[j] <= z
# 5*1+4 == 9
#
# 0 <= d[j] < b
# 0 <= j < m
# e <= m*(b-1) + m-1
#    = m*b - 1
#
# 0 .. n-1 in base b
# z+1 <= m  maybe?
#   z < m
#
sub Omega_b_n {
  my ($m,$z) = @_;
  my $b = ceil($m ** (1/$z));
  my $n = $m*$b;
  $m <= $b**$z or die;
  ($b-1)**$z < $m or die;
  $n - $b*$m <= $b**($z+1) or die;
  return ($b,$n);
}

# Return a list of state table rows (arrayrefs)
#     ( [0,1,2,...],
#       [3,1,2,...],
#     )
sub Omega_table {
  my ($m,$z) = @_;
  my ($b,$n) = Omega_b_n($m,$z);
  my @ret;
  foreach my $i (0 .. $n-1) {
    my @d;
    my $rem = $i;
    foreach (0 .. $z) {
      my $digit = $rem % $b;
      push @d, $digit;
      $rem = ($rem - $digit) / $b;
      $rem==int($rem) or die;
    }
    ### digits: join('',@d)

    $rem == 0 or die;
    for (my $j = 0; @d < $m; $j++) {
      push @d, $d[$j];   # cyclic
    }
    ### cyclic: join('',@d)
    scalar(@d) >= $m or die;

    my @e;
    foreach my $j (0 .. $m-1) {
      my $e = $m*$d[$j] + $j;
      unless ($e < $n) {
        print "oops, too big: m=$m z=$z j=$j d=$d[$j] to e=$e\n";
        die;
      }
      push @e, $e;
    }
    foreach my $j ($m .. $#d) {
      $d[$j] == 0 or die;
    }
    ### e: join('',@e)
    $ret[$i] = \@e;
  }
  return @ret;
}

CHECK {
  # Baker, McNulty, Taylor, table 1 h5,3 column h_{m,z}.
  my @h53 = ('01234',
             '51239',
             '06234',
             '56239',
             '01734',
             '51739',
             '06734',
             '56739',
             '01284',
             '51289');
  my @got = Omega_table(5,3);
  @got = map {join('',@$_)} @got;
  join('__',@got) eq join('__',@h53)  or die;
}

{
  require Math::OEIS::Grep;
  my %seen;
  foreach my $m (2 .. 6) {
    foreach my $z (1 .. $m+2) {
      my ($b,$n) = Omega_b_n($m,$z);
      print "m=$m z=$z is n=$n b=$b\n";
      my $sep = $n > 10 ? ',' : '';
      my @table = Omega_table($m,$z);
      foreach my $row (@table) {
        print "  ",join($sep,@$row),"\n";
      }
      my @h = (0);
      while (@h < 30) {
        print join(',',@h),"\n";
        @h = map { @{$table[$_]} } @h;
      }
      my $str = join(',',@h);
      print "$str\n";
      if ($seen{$str}++) {
        print "DUPLICATE\n";
      } else {
        Math::OEIS::Grep->search (array => \@h,
                                  name => "m=$m, z=$z",
                                  verbose => 0,
                                 );
      }
      print "-------\n";
    }
  }
  exit 0;
}



# A112658
# 0 --> 01
# 1 --> 21
# 2 --> 03
# 3 --> 23
#
# Vec(apply(n->n+1,OEIS_samples("A112658")),20)

my @table = ([0,1],
             [2,1],
             [0,3],
             [2,3]);
sub n_by_table {
  my ($n) = @_;
  my $state = 0;
  foreach my $bit (reverse bit_split_lowtohigh($n)) {
    $state = $table[$state]->[$bit];
  }
  return $state;
}

# starting n=0,
# bit above lowest 0 and parity of its bit position -> num 2,4
# if even, so lowest 0 at least significant -> num 1,3
sub by_bits {
  my ($n) = @_;
  ### n_by_bits(): $n
  if ($n&1) {
    while ($n&1) {
      $n >>= 1;
    }
    return ($n&2) + 1;
  } else {
    return $n&2;
  }
}

{
  my @samples = Math::OEIS::Stripped->anum_to_values('A112658');
  print join('',@samples),"\n";

  my @by_table = map {n_by_table($_)} 0 .. 2**8-1;
  print join('',@by_table[0..20]),"\n";
  join('',@by_table[0..$#samples]) eq join('',@samples)  or die;

  my @by_bits = map {by_bits($_)} 0 .. $#by_table;
  print join('',@by_bits[0..20]),"\n";
  join('',@by_table) eq join('',@by_bits)  or die;

  # print n_by_table(0b11101010101),"\n";
  # print n_by_table(0b111010101010),"\n";
  # print n_by_table(0b1110101010101),"\n";
  #
  # print n_by_table(0b000101010101),"\n";
  # print n_by_table(0b0001010101010),"\n";
  # print n_by_table(0b00010101010101),"\n";

  # my @by_Gray = map {by_Gray($_)} 0 .. $#by_table;
  # print join('',@by_Gray[0..20]),"\n";
  # # join('',@array) eq join('',@by_Gray)  or die;

  # my @by_bits1 = map {n1_by_bits($_)} 0 .. $#by_table;
  # print join('',@by_bits1[0..20]),"\n";
  # join('',@array) eq join('',@by_bits1)  or die;

  my $f = FLAT::DFA->new;
  $f->add_states(4);
  $f->MyFLAT::set_name('high to low');
  $f->set_starting(0);
  $f->set_accepting(1);
  foreach my $from (0 .. 3) {
    foreach my $bit (0,1) {
      my $to = $table[$from]->[$bit];
      $f->add_transition($from,$to, $bit);
    }
  }
  $f->MyFLAT::view;

  # low to high
  # 0 = any..00
  # 1 = any..001111
  # 2 = any..01
  # 3 = any..101111
  # any..00 = num 1
  # num 2
  # .....10 even 1s = num 4
  # .....00 odd 1s = num 4
  # num 4
  # .....00 even 1s = num 4
  # .....10 odd 1s = num 4
  # ......101  = num 4
  $f = $f->MyFLAT::reverse
    ->MyFLAT::set_name('low to high');
  # $f = $f->MyFLAT::minimize(verbose=>1)
  $f->MyFLAT::view;

  exit 0;
}

sub to_Gray_reflected {
  my ($n) = @_;
  my $digits = [ bit_split_lowtohigh($n,2) ];
  Math::PlanePath::GrayCode::_digits_to_gray_reflected($digits,2);
  return digit_join_lowtohigh($digits,2);
}
sub from_Gray_reflected {
  my ($n) = @_;
  my $digits = [ bit_split_lowtohigh($n,2) ];
  Math::PlanePath::GrayCode::_digits_from_gray_reflected($digits,2);
  return digit_join_lowtohigh($digits,2);
}
