/* Copyright 2015, 2016, 2017, 2018 Kevin Ryde

   This file is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 3, or (at your option) any later
   version.

   This file is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.

   You should have received a copy of the GNU General Public License along
   with this file.  See the file COPYING.  If not, see
   <http://www.gnu.org/licenses/>.
*/

/* Some dragon graph experiments using Nauty.
   All likely to change wildly etc.
*/

#include <assert.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <nauty/nauty.h>
#include <nauty/nausparse.h>
#include <nauty/naututil.h>
#include <nauty/naugroup.h>
#include <nauty/gtools.h>
#include <nauty/gutils.h>

#include "nautyextra.h"
#include "nautyextra-wip.h"

#define WANT_DEBUG  0  /* set to 1 for debug prints */
#if WANT_DEBUG
#define DEBUG(expr)  expr
#else
#define DEBUG(expr)
#endif

/* GP-DEFINE  read("dragon-defines.gp"); */


/* ------------------------------------------------------------------------- */
/* Misc Generic Helpers */

#define numberof(array)  (sizeof(array)/sizeof((array)[0]))

/* bit position of the highest 1 bit of an int n>=0 */
static const int  int_highbit_pos = sizeof(int)*CHAR_BIT - 2;

/* static const int  int_highbit = 1 << int_highbit_pos; */

/* Return a bit mask "000011111" which is the lowest 0-bit or 1-bit
   (respectively) of n and all the bits below that position. */
#define DRAGON_MASK_LOWEST_ZERO_AND_BELOW(n)   ((n) ^ ((n) + 1))
#define DRAGON_MASK_LOWEST_ONE_AND_BELOW(n)    ((n) ^ ((n) - 1))

/* Return a bit mask "000100000" which is the bit above the lowest 0-bit of
   n. */
#define DRAGON_MASK_BIT_ABOVE_LOWEST_ZERO(n)    \
  (DRAGON_MASK_LOWEST_ZERO_AND_BELOW(n) + 1)

/* Return a bit mask "000100000" which is the bit above the lowest 1-bit of
   n. */
#define DRAGON_MASK_BIT_ABOVE_LOWEST_ONE(n)    \
  (DRAGON_MASK_LOWEST_ONE_AND_BELOW(n) + 1)

/* Change (x,y) to (-y,x) which is a rotation by +90 degrees. */
#define DRAGON_XY_ROTATE_PLUS90(x,y)            \
  do {                                          \
    int DRAGON_XY_ROTATE_PLUS90__tmp = (x);     \
    (x) = -(y);                                 \
    (y) = DRAGON_XY_ROTATE_PLUS90__tmp;         \
  } while (0)

/* Change (x,y) to (y,-x) which is a rotation by -90 degrees. */
#define DRAGON_XY_ROTATE_MINUS90(x,y)           \
  do {                                          \
    int DRAGON_XY_ROTATE_MINUS90__tmp = (x);    \
    (x) = (y);                                  \
    (y) = - DRAGON_XY_ROTATE_MINUS90__tmp;      \
  } while (0)

/* ------------------------------------------------------------------------- */
/* Dragon Sequence Values */

void
dragon_overflow (void)
{
  fprintf(stderr, "dragon recurrence overflow\n");
  abort();
}

#define DRAGON_RECURRENCE_ITERATE(v0,v1,v2)             \
  do {                                                  \
    int __DRAGON_RECURRENCE_ITERATE__v3 = 2*(v0);       \
    if (__DRAGON_RECURRENCE_ITERATE__v3 < (v0)) {       \
      dragon_overflow();                                \
    }                                                   \
    __DRAGON_RECURRENCE_ITERATE__v3 += (v2);            \
    if (__DRAGON_RECURRENCE_ITERATE__v3 < (v2)) {       \
      dragon_overflow();                                \
    }                                                   \
    (v0) = (v1);                                        \
    (v1) = (v2);                                        \
    (v2) = __DRAGON_RECURRENCE_ITERATE__v3;             \
  } while (0)

/* dragon_LQ(k) returns the number of left boundary squares on dragon curve k.
   1, 1, 2, 4, 6, 10, 18, 30, 50, 86, 146, 246, 418, 710
*/
int
dragon_LQ (int k)
{
  int lq0 = 1;
  int lq1 = 1;
  int lq2 = 2;
  while (k-- > 0) {
    DRAGON_RECURRENCE_ITERATE(lq0,lq1,lq2);
  }
  return lq0;
}

static int
dragon_dJA (int k)
{
  int dja0 = 0;
  int dja1 = 0;
  int dja2 = 1;
  while (k-- > 0) {
    DRAGON_RECURRENCE_ITERATE(dja0,dja1,dja2);
  }
  return dja0;
}

/* 2*8 - 4 + 2*2 - 2*1 == 14
   2*14 - 8 + 2*4 - 2*2 == 24

   R(k+4) = 2*R(k+3) - R(k+2) + 2*R(k+1) - 2*R(k)
*/
#define DRAGON_R_ITERATE(r0,r1,r2,r3)                                   \
  do {                                                                  \
    int __DRAGON_R_ITERATE__r10;                                        \
    int __DRAGON_R_ITERATE__r4 = 2*(r3);                                \
    if (__DRAGON_R_ITERATE__r4 < (r3)) { dragon_overflow(); }           \
                                                                        \
    __DRAGON_R_ITERATE__r4 -= (r2);                                     \
                                                                        \
    __DRAGON_R_ITERATE__r10 = ((r1) - (r0)) << 1;                       \
                                                                        \
    __DRAGON_R_ITERATE__r4 += __DRAGON_R_ITERATE__r10;                  \
    if (__DRAGON_R_ITERATE__r4 < __DRAGON_R_ITERATE__r10) {             \
      dragon_overflow();                                                \
    }                                                                   \
                                                                        \
    (r0) = (r1);                                                        \
    (r1) = (r2);                                                        \
    (r2) = (r3);                                                        \
    (r3) = __DRAGON_R_ITERATE__r4;                                      \
  } while (0)

/*
  GP-Test  vector(100,k,k--; 2*R(k+3) >= R(k+2)) == vector(100,k,k--; 1)

  2*(R(k+1)-R(k)) less than existing value R(k+3) so no overflow at "2*"
  GP-Test  vector(100,k,k--; 2*(R(k+1) - R(k+0)) <= R(k+3)) == vector(100,k,k--; 1)
*/

/* dragon_R(k) returns the right boundary length of dragon curve k.
   1, 2, 4, 8, 16, 28, 48, 84, 144, 244, 416, 708, 1200, 2036, ...
*/
int
dragon_R (int k)
{
  if (k == 0) { return 1; }
  int r0 = 2;
  int r1 = 4;
  int r2 = 8;
  int r3 = 16;
  while (--k > 0) {
    DRAGON_R_ITERATE(r0,r1,r2,r3);
  }
  return r0;
}
/* dragon_B(k) returns the boundary length of dragon curve k.
   2, 4, 8, 16, 28, 48, 84, 144, 244, 416, 708, 1200, ...
*/
int
dragon_B (int k)
{
  return dragon_R(k+1);
}

/* dragon_A(k) returns the area enclosed by dragon curve k.
   0, 0, 0, 0, 1, 4, 11, 28, 67, 152, 335, 724, 1539, ...
*/
int
dragon_A (int k)
{
  int a = 1 << k;
  if (a <= 0) { dragon_overflow(); } /* FIXME: not all CPUs overflow like this */
  return (a - (dragon_B(k) >> 1)) >> 1;
}
/* dragon_AL(k) returns the area enclosed by the left side of dragon curve k.
   0, 0, 0, 0, 1, 3, 7, 17, 39, 85, 183, 389, 815, ...
*/
int
dragon_AL (int k)
{
  if (k==0) { return 0; }
  return dragon_A(k) - dragon_A(k-1);
}

/* dragon_P(k) returns the number distinct points visited by dragon curve k.
   2, 3, 5, 9, 16, 29, 54, 101, 190, 361, 690, 1325, 2558, ...
*/
int
dragon_P (int k)
{
  int p = 1 << k;
  if (p <= 0) { dragon_overflow(); }   /* FIXME: not all CPUs overflow like this */
  return p+1 - dragon_A(k);
}

/* dragon_BlobN(k) returns the point number of the start of blob k.
   The first blob is k=4.
   7,14,28,53,103,206,412,821,1639,3278,6556,13109
*/
int
dragon_BlobN (int k)
{
  static const int offset[] = { 3, 6, 12, 9 };
  int n = 1 << (k+1);
  if (n <= 0) { dragon_overflow(); }   /* FIXME: not all CPUs overflow like this */
  n += offset[k&3];
  if (n <= 0) { dragon_overflow(); }   /* FIXME: not all CPUs overflow like this */
  return n / 5;
}

/* dragon_BlobP(k) returns the number distinct points visited by dragon blob k.
   4, 9, 17, 34, 68, 133, 257, 498, 972, ...
*/
int
dragon_BlobP (int k)
{
  /* 2/10*2^k - 1/10*vector_modulo([17, 9, 23, 31],k) + 1/2 * dJA(k+2) */
  static const int offset[] = { 17, 9, 23, 31 };

  if (k < 4) return 0;
  int p = 1 << (k+1);
  if (p <= 0) { dragon_overflow(); }   /* FIXME: not all CPUs overflow like this */
  p -= offset[k&3];
  p += 5*dragon_dJA(k+2);
  if (p <= 0) { dragon_overflow(); }   /* FIXME: not all CPUs overflow like this */
  return p / 10;
}

/* dragon_TS(k) returns the number of single-visited points in twindragon k.
*/
int
dragon_TS (int k)
{
  return dragon_LQ(k+3);
}
/* dragon_TD(k) returns the number of double-visited points in twindragon k.
*/
int
dragon_TD (int k)
{
  return dragon_AL(k+3);
}
/* dragon_TP(k) returns the number of distinct points visited by
   twindragon k.
*/
int
dragon_TP (int k)
{
  return dragon_TS(k) + dragon_TD(k);
}

/* dragon_TAdiameter(k) returns the diameter of twindragon area tree k. */
int
dragon_TAdiameter (int k)
{
  return dragon_LQ(k+1) - 1;
}

/* dragon_TAW(k) returns the Wiener index of twindragon area tree k.
   my(k=2); ((2*LQ(k) + 38*LQ(k+1) + 9*LQ(k+2) - 34) * 2^k - 24) / 136
*/
unsigned long long
dragon_TAW(int k)
{
  int t;
  if (k == 0) { return 0; }
  t = ((2*dragon_LQ(k) + 38*dragon_LQ(k+1) + 9*dragon_LQ(k+2) - 34) << k) - 24;
  if (t % 136) {
    fprintf(stderr,
            "dragon_TAW() k=%d t=%d not a multiple of 136\n",
            k, t);
    abort();
  }
  t /= 136;
  return ((unsigned long long) t) << (k-1);
}

/* Return true if the dragon curve turns right at point n.
   First point with a turn is n=1.  */
int
dragon_turn_is_right (int n)
{
  return (n & DRAGON_MASK_BIT_ABOVE_LOWEST_ONE(n));
}
/* Return true if the dragon curve turns right after segment n.
   First segment is n=0.  */
int
dragon_next_turn_is_right (int n)
{
  return (n & DRAGON_MASK_BIT_ABOVE_LOWEST_ZERO(n));
}

/* ------------------------------------------------------------------------- */
/* Twindragon Area Tree */

/* Return the vertex above n in the twindragon area tree, or -1 if no edge
   in that direction.
*/
int
dragon_TAVertexToUpper (int n)
{
  int mask = DRAGON_MASK_LOWEST_ONE_AND_BELOW(n);
  if(n & (mask+1)) return -1;
  else             return n + mask;
}
/* Return the vertex below n in the twindragon area tree, or -1 if no edge
   in that direction.
*/
int
dragon_TAVertexToLower (int n)
{
  int mask = DRAGON_MASK_LOWEST_ZERO_AND_BELOW(n);
  if(n & (mask+1)) return n - mask;
  else             return -1;
}
/* Return the vertex right of n in the twindragon area tree, or -1 if no
   edge in that direction
*/
int
dragon_TAVertexToRight (int n)
{
  if (n&1) return -1;
  else     return n ^ 1;
}
/* Return the vertex left of n in the twindragon area tree, or -1 if no edge
   in that direction.
*/
int
dragon_TAVertexToLeft (int n)
{
  if (n&1) return n ^ 1;
  else     return -1;
}

/* Return the vertex in direction dir mod 4 from n in the twin alternate
   area tree, or -1 if no edge in that direction.
 */
int
dragon_TAstepDir (int n, int dir)
{
  switch (dir & 3) {
  case 0:  return dragon_TAVertexToRight(n);  /* dir=0 */
  case 1:  return dragon_TAVertexToUpper(n);  /* dir=1 */
  case 2:  return dragon_TAVertexToLeft(n);   /* dir=2 */
  default: return dragon_TAVertexToLower(n);  /* dir=3 */
  }
}

/* new_dir = dragon_TAparentDir_state_table[dir][bit]
   for bits of n taken high to low
*/
const int dragon_TAparentDir_state_table[4][2]
  = { { 0, 1 },     /* 0=right */
      { 1, 2 },     /* 1=upper */
      { 3, 2 },     /* 2=left */
      { 0, 3 } };   /* 3=lower */

const int dragon_TAparentDir_starting_dir = 1;

/* Return direction 0,1,2,3 which is the parent of vertex n>=1 of the
   twindragon area tree.  n=0 is the root (so has no parent).
*/
int
dragon_TAparentDir (int n)
{
  assert (n >= 0);
  if (n==0) return -1;
  int dir = dragon_TAparentDir_starting_dir;
  /* printf("dragon_TAparentDir() n=%d\n", n); */
  int pos;
  for (pos = int_highbit_pos; pos >= 0; pos--) {
    int bit = (n >> pos) & 1;
    /* printf(" dir=%d pos=%d bit %d\n", dir, pos, bit); */
    dir = dragon_TAparentDir_state_table[dir][bit];
  }
  /* printf(" final dir=%d\n", dir); */
  return dir;
}

/* Return the vertex which is parent of n in the twindragon area tree.
   n=0 is the root and has no parent.  For it return -1.
*/
int
dragon_TAparent (int n)
{
  if (n == 0) { return -1; }
  int dir = dragon_TAparentDir(n);
  assert (dir >= 0);
  n = dragon_TAstepDir(n, dir);
  assert (n >= 0);
  return n;
}

/* ------------------------------------------------------------------------- */
/* Dragon Curve Graph */

/* dragon_graph(g, m,n, k) sets graph g to dragon curve k.
   n should be the number of vertices dragon_P(k).
   g should be already allocated to size m,n.

   dragon_blob_graph(g, m,n, k) similarly, but a graph just of blob k from
   the curve (the middle biggest blob of curve k).
   n should be the number of vertices dragon_BlobP(k).

   The current implementation follows the curve by its bit above lowest 0
   turn sequence and does a linear search for preceding x,y to notice a
   point revisited.  This is not very efficient but is easy and is fast
   enough since any subsequent real work on the graph takes much longer.
*/
int dragon_graph_range_last_v;
void
dragon_graph_range(graph *g, int m,int n, int start,int end)
{
  struct { int x; int y; } *v_to_xy;
  int i;
  int from = -1;
  int dx = 1;
  int dy = 0;
  int xy_upto = 0;
  int x = 0;
  int y = 0;
  DEBUG(printf("dragon_graph_range n=%d  start=%d end=%d\n", n, start,end));

  v_to_xy = ALLOCS(n, sizeof(v_to_xy[0]));
  if (v_to_xy == NULL) alloc_error("dragon_graph()");

  nauty_check(WORDSIZE, m,n, NAUTYVERSIONID);
  EMPTYGRAPH(g, m,n);

  for (i = 0; i <= end; i++) {
    int to;
    if (i >= start) {
      DEBUG(printf(" i=%d xy= %d,%d  dxdy=%d,%d\n", i, x,y, dx,dy));
      for (to = 0; to < xy_upto; to++) {
        if (v_to_xy[to].x == x && v_to_xy[to].y == y) {
          goto GOT_TO;
        }
      }
      to = xy_upto++;
      DEBUG(printf(" new location, vertex %d\n", to));
      if (to >= n) {
        fprintf(stderr, "dragon_graph() oops, at i=%d (end=%d) to=%d xy=%d,%d exceeds size n=%d\n",
                i, end, to, x,y, n);
        abort();
      }
      v_to_xy[to].x = x;
      v_to_xy[to].y = y;
    GOT_TO:
      if (from >= 0) {
        DEBUG(printf(" edge %d to %d\n", from,to));
        ADDONEEDGE(g, from, to, m);
      }
      from = to;
      dragon_graph_range_last_v = to;
    } else {
      DEBUG(printf(" skip i=%d xy= %d,%d\n", i, x,y));
    }

    x += dx;
    y += dy;
    if (dragon_next_turn_is_right(i))
      DRAGON_XY_ROTATE_MINUS90(dx,dy);      /* right */
    else
      DRAGON_XY_ROTATE_PLUS90(dx,dy);       /* left */
  }

  if (xy_upto != n) {
    fprintf(stderr, "oops, dragon graph start=%d end=%d xy_upto=%d but vertices n=%d\n",
            start,end, xy_upto, n);
    abort();
  }
  FREES(v_to_xy);
}
void
dragon_graph(graph *g, int m, int n, int k)
{
  dragon_graph_range(g, m,n, 0, 1<<k);
}
void
dragon_blob_graph(graph *g, int m, int n, int k)
{
  if (k<4) {
    assert(n==0);
    return;
  }
  dragon_graph_range(g, m,n, dragon_BlobN(k), dragon_BlobN(k+1)-3);
}

/* ------------------------------------------------------------------------- */
/* Twindragon Graph */

/* dragon_make_twindragon_graph(g, m,n, k) sets graph g to twindragon k.
   n should be the number of vertices dragon_TP(k).
   g should be already allocated to size m,n.
*/
void
dragon_make_twindragon_graph(graph *g, int m, int n, int k)
{
  struct { int x; int y; } *v_to_xy;
  int rev, i;
  int limit = 2 << k;
  int from;
  int dx = 1;
  int dy = 0;
  int xy_upto = 1;
  int x = 0;
  int y = 0;

  v_to_xy = ALLOCS(n, sizeof(v_to_xy[0]));
  if (v_to_xy == NULL) alloc_error("dragon_make_twindragon_graph()");
  v_to_xy[0].x = 0;
  v_to_xy[0].y = 0;

  nauty_check(WORDSIZE, m,n, NAUTYVERSIONID);
  EMPTYGRAPH(g, m,n);

  from = 0;
  for (rev = 0; rev <= 1; rev++) {
    for (i = 1; i <= limit; i++) {
      int to;
      x += dx;
      y += dy;
      DEBUG(printf(" i=%d xy= %d,%d\n", i, x,y));
      for (to = 0; to < xy_upto; to++) {
        if (v_to_xy[to].x == x && v_to_xy[to].y == y) {
          goto GOT_TO;
        }
      }
      to = xy_upto++;
      if (to >= n) {
        fprintf(stderr, "oops, dragon_make_twindragon_graph() i=%d(of %d) to=%d exceeds size %d\n",
                i, limit, to, n);
        abort();
      }
      v_to_xy[to].x = x;
      v_to_xy[to].y = y;
    GOT_TO:
      /* printf("edge %d to %d at %d,%d rev=%d\n", from, to, x, y, rev); */
      ADDONEEDGE(g, from, to, m);

      if (dragon_turn_is_right(i))
        DRAGON_XY_ROTATE_MINUS90(dx,dy);     /* right */
      else
        DRAGON_XY_ROTATE_PLUS90(dx,dy);     /* left */

      from = to;
    }
    dx = -1;  /* second curve rotate 180 */
    dy = 0;
  }

  if (xy_upto != n) {
    fprintf(stderr, "oops, dragon_make_twindragon_graph() k=%d xy_upto=%d but vertices n=%d\n",
            k, xy_upto, n);
    abort();
  }
  FREES (v_to_xy);
}

/* ------------------------------------------------------------------------- */
/* Twindragon Area Tree */

/* dragon_TA_tree_n(k) returns the number of vertices in twindragon
   area tree k.  This is simply 2^k.

   dragon_make_TA_tree(g, m,n) sets graph g to twindragon area tree k.
   n should be the number of vertices as per dragon_TA_tree_n(k).
   g should be already allocated to size m,n.

   The twindragon area tree is a tree, so sparse, so usually
   dragon_make_TA_tree_sg() below is better, when sparsegraph suits.
*/
int
dragon_TA_tree_n(int k)
{
  int n = 1 << k;
  if (n <= 0) { dragon_overflow(); }
  return n;
}
void
dragon_make_TA_tree(graph *g, int m, int n)
{
  int from;

  assert (n >= 0);
  nauty_check(WORDSIZE, m,n, NAUTYVERSIONID);
  EMPTYGRAPH(g, m,n);

  for (from = 0; from < n; from++) {

    /* horizontal from=xx1 edge to=xx0 */
    if (from & 1) {
      int to = from ^ 1;
      ADDONEEDGE(g, from,to, m);
    }

    /* to lower */
    if (from < n - 1) {    /* no zeros in from=1111 */
      int mask = DRAGON_MASK_LOWEST_ZERO_AND_BELOW(from);
      if (from & (mask+1)) {  /* from=...10111 */
        int to = from - mask; /*   to=...01000 */
        ADDONEEDGE(g, from,to, m);
      }
    }
  }
}

/* dragon_make_TA_tree_sg(sg,k) sets sparsegraph sg to twindragon area tree
   level k.

   sg must be SG_INIT but any previous contents are cleared and realloced as
   necessary for k.

   Edges are added in both directions between vertices, as usual for an
   undirected sparsegraph.

   ENHANCE-ME: Maybe a directed option for up to root=0.
*/
void
dragon_make_TA_tree_sg(sparsegraph *sg, int k)
{
  assert (k >= 0);
  int num_vertices = dragon_TA_tree_n(k);
  int num_edges = num_vertices - 1;

  int nde = 2*num_edges;
  SG_ALLOC(*sg, num_vertices, nde, "malloc");
  sg->nv = num_vertices;
  sg->nde = nde;

  int *dd = sg->d;
  size_t *vv = sg->v;
  int *ee = sg->e;
  int epos = 0;
  int from;
  for (from = 0; from < num_vertices; from++) {
    *vv++ = epos;
    int from_degree = 0;

    /* horizontal from=...0 edge to=...1 or the other way around */
    if (k) {
      int to = from ^ 1;
      ee[epos++] = to;
      from_degree++;
    }

    /* to lower */
    if (from < num_vertices - 1) {
      int mask = DRAGON_MASK_LOWEST_ZERO_AND_BELOW(from);
      if (from & (mask+1)) {    /* from=...10111 */
        int to = from - mask;   /*   to=...01000 */
        ee[epos++] = to;
        from_degree++;
      }
    }

    /* to upper */
    if (from > 0) {   /* no 1s in from=00000 */
      int mask = DRAGON_MASK_LOWEST_ONE_AND_BELOW(from);
      if ((from & (mask+1)) == 0) {   /* from=...01000 */
        int to = from + mask;         /*   to=...10111 */
        if (to < num_vertices) {
          ee[epos++] = to;
          from_degree++;
          DEBUG(printf("%d -> %d  upper\n", from, to));
        }
      }
    }

    *dd++ = from_degree;
  }
  assert (epos == nde);
}

/* dragon_make_TA_tree_vpar() sets vpar[1..2^k] to twindragon area tree k.
   The vertex numbers are the area tree numbering 0 .. 2^k-1 but +1.
*/
void
dragon_make_TA_tree_vpar(int *vpar, int k)
{
  assert (k >= 0);
  int n_max = dragon_TA_tree_n(k);
  int n;
  vpar[0] = 0;
  for (n = 1; n <= n_max; n++)
    vpar[n] = dragon_TAparent(n-1) + 1;
}

/* ------------------------------------------------------------------------- */

void
test (void)
{
  int bad = 0;
  printf ("int_highbit_pos = %d\n", int_highbit_pos);
  {
    /* dragon_TAparentDir */
    int want[] = { 2, 3, 2, 0, 3, 3, 2, 0, 1, 0, 3, 0, 3, 3, 2, 0, 1 };
    int i;
    for (i = 0; i < sizeof(want)/sizeof(want[0]); i++) {
      int n = i+1;
      int got = dragon_TAparentDir(n);
      if (! (got == want[i])) {
        printf ("dragon_TAparentDir(%d) want %d got %d\n", n, want[i], got);
        bad++;
      }
    }
  }
  {
    /* dragon_TAparent */
    int want[] = { 0, 1, 2, 5, 2, 5, 6, 9, 10, 11, 4, 13, 10, 13, 14, 17, 18 };
    int i;
    for (i = 0; i < sizeof(want)/sizeof(want[0]); i++) {
      int n = i+1;
      int got = dragon_TAparent(n);
      if (! (got == want[i])) {
        printf ("dragon_TAparent(%d) want %d got %d\n", n, want[i], got);
        bad++;
      }
    }
  }
  {
    /* dragon_make_TA_tree_sg() vs dragon_make_TA_tree_vpar() */
    const int k_max = 8;
    int vpar[(1 << k_max)+1];
    SG_DECL(sg);
    SG_INIT(sg);
    int k;
    for (k=0; k < 8; k++) {
      dragon_make_TA_tree_sg(&sg,k);
      dragon_make_TA_tree_vpar(vpar,k);
      int n = dragon_TA_tree_n(k);
      if (! nautyextra_is_isomorphic_sg_vpar(&sg,vpar,n)) {
        printf ("dragon_make_TA_tree_sg() vs dragon_make_TA_tree_vpar() k=%d\n", k);
        bad++;
      }
    }
    SG_FREE(sg);
  }
  if (bad) abort();
}
void
writeautom (int *p, int n)
{
  int i;
  printf("autom ");
  for (i = 0; i < n; i++)
    printf(" %2d",p[i]);
  printf("\n");
}

void
show_edges (graph *g, int m,int n)
{
  set *row = (set*) g;
  int from, to;
  printf("edges");
  for (from=0; from < n-1; from++, row+=m) {
    for (to=from+1; to < n; to++) {
      if (ISELEMENT(row,to))
        printf(" %d-%d", from,to);
    }
  }
  putchar('\n');
}

void
try_make_TA_graph (void)
{
  static DEFAULTOPTIONS_SPARSEGRAPH(options);
  statsblk stats;

  DYNALLSTAT(int, lab,    lab_sz);
  DYNALLSTAT(int, ptn,    ptn_sz);
  DYNALLSTAT(int, orbits, orbits_sz);
  DYNALLSTAT(int, map,    map_sz);
  SG_DECL(ta);
  SG_DECL(canon_ta);
  SG_DECL(canon_ta2);

  int k = 5;

  options.getcanon = TRUE;
  options.userautomproc = groupautomproc;
  options.userlevelproc = grouplevelproc;

  printf("make tree\n");
  dragon_make_TA_tree_sg(&ta, k);
  printf("n = %d\n", ta.nv);
  printf("dilation %d\n", nautyextra_dilation_sg(&ta));

  printf("\n");
  printf("as generated\n");
  nautyextra_print_sg(&ta);

  int num_vertices = ta.nv;
  DYNALLOC1(int, lab,    lab_sz,    num_vertices, "malloc");
  DYNALLOC1(int, ptn,    ptn_sz,    num_vertices, "malloc");
  DYNALLOC1(int, orbits, orbits_sz, num_vertices, "malloc");
  DYNALLOC1(int, map,    map_sz,    num_vertices, "malloc");

  printf("\nlabel ta\n");
  sparsenauty(&ta,lab,ptn,orbits,&options,&stats,&canon_ta);

  printf("canonical\n");
  nautyextra_print_sg(&canon_ta);

  sortlists_sg(&ta);

  {
    int num_edges = ta.nde;
    int m = SETWORDSNEEDED(num_vertices);
    printf("fcanon m=%d\n", m);
    /* int loops = loopcount(&ta, m, num_vertices); */
    /* printf("loops %d\n", loops); */
    SG_ALLOC(canon_ta2, num_vertices, num_edges, "canon_ta2");
    fcanonise_inv_sg(&ta, m, num_vertices, &canon_ta2,
                     NULL,  /* fmt colours */
                     NULL,  /* distances_sg,  */ /* invarproc */
                     1,     /* mininvarlevel */
                     1,     /* maxinvarlevel */
                     3,     /* invararg */
                     0);    /* digraph */
    sortlists_sg(&canon_ta2);
  }

  printf("putgraph_sg()\n");
  putgraph_sg(stdout, &ta, 79);  /* line length */

  printf("\n");
  writeline(stdout,SPARSE6_HEADER);
  writes6_sg(stdout,&ta);
  writeline(stdout,SPARSE6_HEADER);
  writes6_sg(stdout,&canon_ta);
  writeline(stdout,SPARSE6_HEADER);
  writes6_sg(stdout,&canon_ta2);

  printf("\n");
  writeline(stdout,GRAPH6_HEADER);
  writeg6_sg(stdout,&ta);
  writeline(stdout,GRAPH6_HEADER);
  writeg6_sg(stdout,&canon_ta);
  writeline(stdout,GRAPH6_HEADER);
  writeg6_sg(stdout,&canon_ta2);

  printf("aresame  %d %d\n",
         aresame_sg(&ta,&canon_ta),
         aresame_sg(&canon_ta,&canon_ta2));


  printf ("\nGenerators\n");
  printf ("automorphism group size = ");
  writegroupsize(stdout, stats.grpsize1, stats.grpsize2);
  printf ("\n");
  printf ("num orbits    %d\n", stats.numorbits);
  printf ("  orbits      ");
  int i;
  for (i = 0; i < canon_ta.nv; i++) printf(" %d",orbits[i]);
  printf("\n");
  printf ("errstatus     %d\n", stats.errstatus);
  printf ("numnodes      %lu\n", stats.numnodes);
  printf ("numbadleaves  %lu\n", stats.numbadleaves);
  printf ("maxlevel      %d\n", stats.maxlevel);

  {
    grouprec *group = groupptr(FALSE);
    if (! group) {
      printf("no grouprec\n");
    } else {
      printf("grouprec %p\n", group);
      makecosetreps(group);
      allgroup(group,writeautom);
    }
  }
}

void
try_TAdiameter (void)
{
  DYNALLSTAT(graph, ta, ta_sz);
  int k;

  for (k=0; k < 10; k++) {
    int tadiameter = dragon_TAdiameter(k);
    int n = dragon_TA_tree_n(k);
    int m = SETWORDSNEEDED(n);
    int radius, diameter;

    DYNALLOC2(graph,ta,ta_sz,m,n,"malloc");
    dragon_make_TA_tree(ta, m,n);
    diamstats(ta, m,n, &radius, &diameter);

    printf("k=%d  radius %d  diameter %d %d\n",
           k, radius, diameter, tadiameter);
  }
}

void
try_TAW (void)
{
  DYNALLSTAT(graph, ta, ta_sz);
  int k;

  for (k=0; k < 10; k++) {
    int n = dragon_TA_tree_n(k);
    int m = SETWORDSNEEDED(n);
    unsigned long long TAW_calc;
    unsigned long long TAW_formula = dragon_TAW(k);

    DYNALLOC2(graph,ta,ta_sz,m,n,"malloc");
    dragon_make_TA_tree(ta, m,n);
    struct nautyextra_Wiener_info info;
    nautyextra_Wiener_index(ta, m,n, 0, &info);
    TAW_calc = info.W;

    printf("k=%d  Wiener index %Ld %Ld\n", k, TAW_calc, TAW_formula);
  }
}

void
try_sequences (void)
{
  int n, k;
  printf("turn right n=1: ");
  for (n=1; n < 10; n++) {
    printf(" %d,", dragon_turn_is_right(n));
  }
  printf("\n");

  for (k=0; k < 10; k++) {
    int R = dragon_R(k);
    int B = dragon_B(k);
    int A = dragon_A(k);
    int P = dragon_P(k);
    int TP = dragon_TP(k);
    int BlobN = dragon_BlobN(k);
    int BlobP = dragon_BlobP(k);
    printf("k=%d  R=%d B=%d A=%d P=%d TP=%d BlobN=%d BlobP=%d\n", k, R, B, A, P, TP, BlobN, BlobP);
  }
}

/* Dragon diameter 1,2,4,8, 12,18, 26,36, 52,70, 102,136, 200,266, ...
   Twindragon diameter 2, 2, 4, 8, 12, 20, 28, 40, 56, 80, ...

   Twindragon Wiener
   8,40,212,1240,6216,32028,161992,852836,4351024,23381168,121657192,663496008,3508777680,19441583316,104018948832
   TW(0) = 4*(2*1 + 2) / 2 = 8
   TW(1) = ( 2*(2*1 + 2 + 2*3 + 4) + 4*(2*1 + 3*2 + 3) + 4*1+2*2 ) / 2 = 40
   grows roughly 5.282725^k = 4*1.32068 = 2*2.641362
*/
void
try_dragon_diameter (void)
{
  DYNALLSTAT(graph, g, g_sz);
  int k;

  for (k=5; k <= 10; k++) {
    int radius = 0, diameter = 0, cycles = 0, diameter_formula = 0;
    unsigned long long W = 0;

    /* int n = dragon_P(k); */
    /* int m = SETWORDSNEEDED(n); */
    /* DYNALLOC2(graph,g,g_sz,m,n,"malloc"); */
    /* dragon_graph(g, m,n, k); */

    int n = dragon_TP(k);
    int m = SETWORDSNEEDED(n);
    DYNALLOC2(graph,g,g_sz,m,n,"malloc");
    dragon_make_twindragon_graph(g, m,n, k);

    diamstats(g, m,n, &radius, &diameter);
    /* cycles = cyclecount(g, m,n); */
    /* W = nautyextra_Wiener_index(g, m,n); */

    printf("k=%d n=%d  radius %d  diameter %d %d  cycles %d  W=%Ld\n",
           k, n, radius, diameter, diameter_formula, cycles, W);
  }
}

void
try_blob_diameter (void)
{
  DYNALLSTAT(graph, g, g_sz);
  int k;

  for (k=4; k <= 10; k++) {
    int radius = 0, diameter = 0;
    int start_ecc = 0, start_ecc_destination = 0;
    int end_ecc = 0, end_ecc_destination = 0;
    unsigned long long W = 0;

    int n = dragon_BlobP(k);
    int m = SETWORDSNEEDED(n);
    DYNALLOC2(graph,g,g_sz,m,n,"malloc");
    dragon_blob_graph(g, m,n, k);


    diamstats(g, m,n, &radius, &diameter);
    /* W = nautyextra_Wiener_index(g, m,n); */
    start_ecc = nautyextra_eccentricity_of_vertex(g, m,n, 0,
                                                  &start_ecc_destination);
    end_ecc = nautyextra_eccentricity_of_vertex(g, m,n,
                                                dragon_graph_range_last_v,
                                                &end_ecc_destination);

    printf("k=%d n=%d  radius %d  diameter %d  W=%Ld  start ecc %d (to %d)  end ecc %d (%d to %d)\n",
           k, n, radius, diameter, W,
           start_ecc, start_ecc_destination,
           end_ecc, dragon_graph_range_last_v, end_ecc_destination);
    /* if (n <= 8) { */
    /*   writeg6(stdout,g,m,n); */
    /*   writes6(stdout,g,m,n); */
    /*   show_edges(g,m,n); */
    /* } */

    {
      int v = 0, i;
      int *dist;
      dist = (int*) ALLOCS(n,sizeof(int));
      if (dist == NULL) alloc_error("for find_dist()");
      find_dist(g, m,n, v, dist);
      for (i = 0; i < n; i++) {
        if (dist[i] == start_ecc) {
          printf("  start_ecc attained by %d\n", i);
        }
      }
      FREES(dist);
    }
  }
}

/*
  math-image --wx --path=DragonCurve --values=Lines --scale=20

         *----*       k=5
         |    |       cycles = 4
    *----*----*
    |    |
    *----*
    |    |
    *----*

                      *----*       k=6
                      |    |       cycles = 9
  *----*    *----*----*----*
  |    |    |    |    |
  *----*----*----*----*
       |    |    |    |
       *----*    *----*
*/
void
try_blob_cyclecount (void)
{
  DYNALLSTAT(graph, g, g_sz);
  int k;

  for (k=4; k <= 10; k++) {
    int n = dragon_BlobP(k);
    if (n > 32) break;

    int m = SETWORDSNEEDED(n);
    DYNALLOC2(graph,g,g_sz,m,n,"malloc");
    dragon_blob_graph(g, m,n, k);

    long cycles = cyclecount(g,m,n);

    printf("k=%d n=%d  cyclecount %ld\n", k,n, cycles);
  }
}

/* TA
   indnum = 2^(k-1)
   count indnum
   not in OEIS: 1, 2, 3,  8,   28,  300, 56144, 2217746112, 3488404844698242048
   .            0  1  2   3     4     5

               2---3   level => 2 = path-4  indnum=2 sets=2
               |       indnum by 0,2  \ complements
           0---1                 1,3  /

               6---7     level => 3         indnum=4 sets=3
               |
           4---5         indnum by 0,2 4,6  \ complements
               |                   1,3 5,7  /
               2---3               1,3 4,6  extra
               |
           0---1


   count_indsets
   not in OEIS: 2, 3, 8, 60, 3024, 7872192, 53686957178880
   .            0  1  2   3     4     5

   count indomsets (maximal indsets)
   not in OEIS: 1,2,3,8,76,5456,28285632

   to_binary(7872192)        ==  11110000001111011000000
   to_binary(53686957178880) ==  1100001101001111111000000000110000000000000000
*/
void
try_TAindnum (void)
{
  DYNALLSTAT(int,vpar,vpar_sz);
  SG_DECL(sg);
  SG_INIT(sg);
  int k;
  for (k=0; k < 12; k++) {
    int n = dragon_TA_tree_n(k);
    DYNALLOC1(int,vpar,vpar_sz,n+1,"try");
    dragon_make_TA_tree_vpar(vpar, k);

    if (0) {
      printf("k=%d n=%d\n", k,n);
    }

    if (0) {
      int indnum = nautyextra_indnum_vpar(vpar,n);
      printf("    indnum %d\n", indnum);
    }

    if (0) {
      struct nautyextra_indnum_and_count  indnum_and_count;
      nautyextra_indnum_and_count_vpar (vpar,n,&indnum_and_count);
      /* printf("    indnum %d  count %Lu\n", indnum_and_count.indnum, indnum_and_count.count); */
      printf("%Lu,", indnum_and_count.count);
    }

    if (1) {
      unsigned long long count_indsets = nautyextra_indsets_count_vpar(vpar,n);
      /* printf("    count_indsets %Lu\n", count_indsets); */
      printf("%Lu, ", count_indsets);
    }
    if (0) {
      unsigned long long count_indomsets
        = nautyextra_indomsets_count_vpar(vpar,n);
      printf("    count_indomsets %Lu\n", count_indomsets);
    }
  }
  SG_FREE(sg);
}

/* TA domnum = 13/32*2^k + 1, initial 1,1,2,4,7
                0 1 2 3 4  5
   not in OEIS: 1,1,2,4,7,14,27,53,105,209,417,833
   A168596 = 14 onwards
   apply(to_binary,[1,1,2,4,7,14,27,53,105,209,417,833,1665,3329,6657,13313,26625,53249,106497,212993])
   110100000000001
   TAdomnum(k) = 2*TAdomnum(k-1) - 1
   connection vertex leaf, its required is already dominated, so on join
   choose one of the two now non-leaf

   TA domnum count
                0 1 2  3  4     5
   not in OEIS: 2,2,4,16,96,16256
   96    == 2^5 * 3
   16256 == 2^7 * 127
*/
void
try_TAdomnum (void)
{
  DYNALLSTAT(int,vpar,vpar_sz);
  printf("TA domnum\n");
  int k;
  for (k=0; k < 10; k++) {
    int n = dragon_TA_tree_n(k);
    /* printf("k=%d n=%d\n", k,n); */

    DYNALLOC1(int,vpar,vpar_sz,n+1,"try");
    dragon_make_TA_tree_vpar(vpar, k);

    int domnum = nautyextra_domnum_vpar(vpar,n);
    printf("%d,", domnum);
  }
  printf("\n");

  printf("TA domnum count\n");
  SG_DECL(sg);
  SG_INIT(sg);
  for (k=0; k < 10; k++) {
    struct nautyextra_domnum_and_count  info;
    dragon_make_TA_tree_sg(&sg, k);
    nautyextra_domnum_and_count_sg(&sg, &info);
    printf("%Lu,", info.count);
  }
  SG_FREE(sg);
}

/* TA domsets

   domsets
   not in OEIS: 1,3,9,81,9153,104026113,14081844418010793
   GP-Test  9153 == 3^4*113

   minimal domsets
   not in OEIS: 1,2,4,16,196,45136,1663247296,2247330711026085888

   factor(1663247296)
   factor(2247330711026085888)
*/
void
try_TAdomsets (void)
{
  DYNALLSTAT(int,vpar,vpar_sz);
  SG_DECL(sg);
  SG_INIT(sg);
  int k;
  {
    for (k=0; k <= 6; k++) {
      int n = dragon_TA_tree_n(k);
      DYNALLOC1(int,vpar,vpar_sz,n+1,"try");
      dragon_make_TA_tree_vpar(vpar, k);

      /* printf("k=%d n=%d\n", k,n); */

      unsigned long long count = nautyextra_domsets_count_vpar(vpar,n);
      printf("k=%d n=%d  domsets %Lu\n", k, n, count);
      /* printf("%d,", count); */
    }
    printf("\n");
  }
  {
    for (k=0; k <= 7; k++) {
      int n = dragon_TA_tree_n(k);
      DYNALLOC1(int,vpar,vpar_sz,n+1,"try");
      dragon_make_TA_tree_vpar(vpar, k);

      /* printf("k=%d n=%d\n", k,n); */

      unsigned long long count = nautyextra_minimal_domsets_count_vpar(vpar,n);
      printf("k=%d n=%d  minimal domsets %Lu\n", k, n, count);
      /* printf("%d,", count); */
    }
    printf("\n");
  }
  SG_FREE(sg);
}


/* #define CLEAR_LOWEST_ONE(n) \ */
/*   (n & ~DRAGON_MASK_LOWEST_ONE_AND_BELOW(n)) */

/* dragon graph
   k      0,1,2,3,4, 5, 6
   indnum 1,2,3,5,9,16,29
   count  2,1,1,1,1, 1, 1
   indomnum 0,0
   count    1,1

   blob
   k      4,5,6, 7
   indnum 2,5,9,18,
   count  2,1,1, 1,

   k=5     *---.
           |   |
       *---.---*  indnum=5
       |   |
       .---*
       |   |
       *---.
*/
void
try_dragon_graph_indnum (void)
{
  DYNALLSTAT(graph,g,g_sz);
  SG_DECL(sg);
  SG_INIT(sg);
  int k;
  for (k=0; k <= 12; k++) {
    int n = dragon_P(k);
    int m = SETWORDSNEEDED(n);
    printf("k=%d m,n=%d,%d dragon\n", k,m,n);
    DYNALLOC2(graph,g,g_sz,m,n,"malloc");
    dragon_graph(g, m,n, k);
    nauty_to_sg(g,&sg,m,n);

    /* int n = dragon_BlobP(k); */
    /* int m = SETWORDSNEEDED(n); */
    /* printf("k=%d m,n=%d,%d blob\n", k,m,n); */
    /* DYNALLOC2(graph,g,g_sz,m,n,"malloc"); */
    /* dragon_blob_graph(g, m,n, k); */

    {
      struct nautyextra_indnum_and_count  info;
      nautyextra_indnum_and_count(g,m,n,&info);
      printf("    indnum %d count %Lu\n", info.indnum, info.count);
    }
    {
      struct nautyextra_indomnum_and_count  info;
      nautyextra_indomnum_and_count_sg(&sg,&info);
      printf("    indomnum %d count %Lu\n", info.indomnum, info.count);
    }
  }
  SG_FREE(sg);
}

/* twindragon graph
   k      0,1,2, 3, 4, 5, 6
   indnum 2,4,7,13,25,
   count  2,1,1, 1, 1,
   
   indomnum 2,3,5, 9,25,
   count    2,3,5,13, 1,
*/
void
try_twindragon_graph_indnum (void)
{
  DYNALLSTAT(graph,g,g_sz);
  SG_DECL(sg);
  SG_INIT(sg);
  int k;
  for (k=0; k <= 12; k++) {
    int n = dragon_TP(k);
    int m = SETWORDSNEEDED(n);
    printf("k=%d m=%d, n=%d\n", k,m,n);
    DYNALLOC2(graph,g,g_sz,m,n,"malloc");
    dragon_make_twindragon_graph(g, m,n, k);
    nauty_to_sg(g,&sg,m,n);

    {
      struct nautyextra_indnum_and_count  info;
      nautyextra_indnum_and_count(g,m,n,&info);
      printf("    indnum %d count %Lu\n", info.indnum, info.count);
    }
    {
      struct nautyextra_indomnum_and_count  info;
      nautyextra_indomnum_and_count_sg(&sg,&info);
      printf("    indomnum %d count %Lu\n", info.indomnum, info.count);
    }
  }
  SG_FREE(sg);
}


/* TA matchnum = 2^(k-1) in 1 way
*/
void
try_TAmatchnum (void)
{
  DYNALLSTAT(int,vpar,vpar_sz);
  printf("TA matchnum\n");
  int k;
  for (k=0; k < 10; k++) {
    int n = dragon_TA_tree_n(k);
    DYNALLOC1(int,vpar,vpar_sz,n+1,"try");
    dragon_make_TA_tree_vpar(vpar, k);

    int matchnum = nautyextra_matchnum_vpar(vpar,n);
    printf("%d,", matchnum);
  }
  printf("\n");

  printf("TA matchnum count\n");
  for (k=0; k < 10; k++) {
    int n = dragon_TA_tree_n(k);
    DYNALLOC1(int,vpar,vpar_sz,n+1,"try");
    dragon_make_TA_tree_vpar(vpar, k);
    struct nautyextra_matchnum_and_count  info;
    nautyextra_matchnum_and_count_vpar(vpar,n, &info);
    printf("%Lu,", info.count);
  }
}

/* TA totdomnum
*/
void
try_TAtotdomnum (void)
{
  SG_DECL(sg);
  SG_INIT(sg);

  {
    const int k = 6;
    const int n = (1<<k);
    int vpar[n+1];
    printf("TA totdomnum k=%d n=%d\n", k, n);
    dragon_make_TA_tree_vpar(vpar, k);

    nautyextra_vpar_to_sg (vpar, n, &sg, 0); /* undirected */
    assert(sg.nv == n);

    int totdomnum = nautyextra_totdomnum_vpar(vpar,n);
    printf("k=%d totdomnum %d\n", k, totdomnum);

    struct nautyextra_totdomnum_and_count info;
    nautyextra_totdomnum_and_count_vpar(vpar,n, &info);
    printf("totdomnum %d count %Lu\n", info.totdomnum, info.count);

    static const setword k5[] = {
      1713791862, 1713792870, 1717986918, 1722179958, 1722180966, 1726375014,
      1854300534, 1854301542, 1858495590
    };
    static const int swap_from[] = { 16, 17 };
    static const int swap_to[]   = { 46, 47 };

    int i;
    for (i=0; i<numberof(k5); i++) {
      assert(n/2==32);
      int size = setsize((set*)&k5[i], 1);
      if (size != totdomnum/2) {
        printf ("i=%d size %d want %d  %X %s\n",
                i, size, totdomnum/2, k5[i], nautyextra_set_str(&k5[i],n/2));
        abort();
      }
    }

    int m = SETWORDSNEEDED(n);
    assert(m==2);
    int count = 0;
    for (i=0; i<numberof(k5); i++) {
      int j;
      for (j=0; j<numberof(k5); j++) {
        int swap;
        for (swap=0; swap <= 3; swap++) {
          setword set[2];
          set[0] = k5[i];
          set[1] = k5[j];
          if (swap & 1) {
            FLIPELEMENT(set,swap_from[0]);
            FLIPELEMENT(set,swap_to[0]);
          }
          if (swap & 2) {
            FLIPELEMENT(set,swap_from[1]);
            FLIPELEMENT(set,swap_to[1]);
          }

          int size = setsize(set,m);
          if (size != totdomnum) {
            printf ("i=%d j=%d swap=%d  size %d want %d   %s\n",
                    i,j,swap,
                    size, totdomnum, nautyextra_set_str(set,n));
            abort();
          }

          int is_totdomset = nautyextra_is_totdomset_sg (&sg, set);
          count += is_totdomset;
          if (! is_totdomset) {
            printf("not i=%d j=%d swap=%d   %s\n",
                   i, j, swap, nautyextra_set_str(set,n));
          }
        }
      }
    }
    printf ("k=6 combinations %d\n", count);
  }

  {
    const int k = 5;
    const int n = (1<<k);
    int vpar[n+1];
    printf("TA totdomnum k=%d n=%d\n", k, n);
    dragon_make_TA_tree_vpar(vpar, k);
    nautyextra_vpar_to_sg (vpar, n, &sg, 0); /* undirected */

    int totdomnum = nautyextra_totdomnum_vpar(vpar,n);
    printf("totdomnum %d\n", totdomnum);

    struct nautyextra_totdomnum_and_count info;
    nautyextra_totdomnum_and_count_vpar(vpar,n, &info);
    printf("totdomnum %d count %Lu\n", info.totdomnum, info.count);

    setword inc = BITT[n-1];
    setword ss = 0;
    do {
      if (POPCOUNT(ss) != totdomnum) continue;
      if (nautyextra_is_totdomset_sg (&sg, &ss)) {
        printf("%s\n", nautyextra_set_str(&ss,n));
        /* setword base = 1717986918; */
        /* setword  xor = base ^ ss; */
        /* printf("xor %s\n", nautyextra_set_str(&xor,n)); */
      }
    } while (ss += inc);
  }
  {
    DYNALLSTAT(int,vpar,vpar_sz);
    int k;
    for (k=0; k < 10; k++) {
      int n = dragon_TA_tree_n(k);
      DYNALLOC1(int,vpar,vpar_sz,n+1,"try");
      dragon_make_TA_tree_vpar(vpar, k);

      struct nautyextra_totdomnum_and_count info;
      nautyextra_totdomnum_and_count_vpar(vpar,n, &info);
      printf("k=%d totdomnum %d count %Lu\n", k, info.totdomnum, info.count);
    }
    printf("\n");
  }
  SG_FREE(sg);
}

/* ------------------------------------------------------------------------- */

int
main (void)
{
  setbuf(stdout,NULL);
  test();

  try_make_TA_graph();
  return 0;
  try_dragon_graph_indnum();
  try_twindragon_graph_indnum();
  try_TAtotdomnum();
  try_TAdomsets();
  try_TAmatchnum();
  try_TAdomnum();
  try_TAindnum();
  try_blob_cyclecount();
  try_dragon_diameter();
  try_sequences();
  try_blob_diameter();
  try_TAW();
  try_TAdiameter();
  return 0;
}
