#!/usr/bin/perl -w

# Copyright 2014, 2015, 2016, 2017, 2019 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.010;
use strict;
use Carp 'croak';
use File::Spec;
use List::Util 'max','min','sum','pairmap';
use Math::BaseCnv 'cnv';
use Math::BigInt try => 'GMP';
use Math::Complex 'pi';
use Math::Libm 'hypot';
use Math::Trig 'deg2rad', 'rad2deg';
use POSIX 'floor';

use Math::PlanePath::DragonCurve 118; # v.118 for xyxy_to_n_either()
use Math::PlanePath::Base::Digits
  'bit_split_lowtohigh', 'digit_split_lowtohigh',
  'digit_join_lowtohigh', 'round_down_pow';
use Math::PlanePath::Base::Generic
  'round_nearest';

# use Math::PlanePath;
# *_divrem_mutate = \&Math::PlanePath::_divrem_mutate;
# unshift @INC, '/usr/share/perl5';

use FindBin;
use lib::abs "$FindBin::Bin/../tools";
use MyDragon ':all';
our @ipc;
$|=1;

# uncomment this to run the ### lines
# use Smart::Comments;

{
  # N with high part equal to X, seeking point ffixed fpoint(f)=f

  # k=32
  #                               ^ N=1011 S
  #                               |
  #     N=10000 W   N=1101 S      |
  #  <----- *         * <---------* N=1100 W
  #         | 00      ^ 01       10
  #         |         |
  #         v         |
  # N=1111  *
  #   N          next X 010 and 011, from N=1100 to 1100_10  y=1 S
  #                                                 1100_11  W
  #                                  from N=1101 to 1101_00  S

  # (1/b - 1/b^2) == 1/2            \\ 11
  # (1/b - 1/b^2 - 1/b^4) == 3/4    \\ 110x
  # (1/b - 1/b^2 - 1/b^3 - 1/b^4 - 1/b^5 ) == 9/8 + 1/8*I     \\ 1111x
  # (1/b - 1/b^2 - 1/b^3 - 1/b^4 - 1/b^5 - 1/b^6 - 1/b^8 ) == 17/16 \\ 1111 1101

  my $path = Math::PlanePath::DragonCurve->new;
  my $k = 0;
  my @n_list = (Math::BigInt->new(0));
  for (;;) {
    $k += 2;
    last if $k >= 200;

    my $h = $k/2;
    my @new_n_list;
    foreach my $n_parent (@n_list) {
      ### n_parent: "$n_parent"
      my $n = 4*$n_parent;
      for (my $i = 1; $i <= 4; $i++,$n++) {
        my ($x,$y) = $path->n_to_xy($n);
        my ($dx,$dy) = $path->n_to_dxdy($n);
        ($x,$y) = transform_to_horizontal($x,$y, $k);
        ($dx,$dy) = transform_to_horizontal($dx,$dy, $k);
        my $xe = $x + $dx;
        my $ye = $y + $dy;
        my $n2 = cnv($n,10,2);
        my $ntop = $n >> $h;
        ### consider: "n=$n2  $x,$y--$xe,$ye  ntop=$ntop"
        next unless $y == 0 || $ye == 0;

        # ($ntop - $x) <= 1 || abs($ntop - $xe) <= 1
        if ($ntop==$x || $ntop==$x-1 || $ntop==$xe || $ntop==$xe-1) {
          if ($n > 7 && $n2 !~ /^111111/ && $k%8==0) {
            my $x2 = cnv($x,10,2);
            if ($n&1) { $dx=-$dx; } # side of expansion
            my $HV = $dy==0 ? ($dx<0 ? 'W' : 'E') : '';
            printf "k=%-3d  n=%*s\n       x=$x2 y=$y  dxdy=$dx,$dy   $HV\n",
              $k,$k,$n2;
          }
          push @new_n_list, $n;
        }
      }
    }
    print "\n";
    @n_list = @new_n_list;
  }
  exit 0;
}

{
  # feasible N, doubling
  # 11010000111111010000000000001101000000000000000000000000110100001111111...
  # 110100001111

  my $path = Math::PlanePath::DragonCurve->new;
  my $k = 4;
  my @x_list = (Math::BigInt->new(0b1101));
  while (@x_list) {
    print "k=$k x_list count ",scalar(@x_list),"\n";
    my $n_hi = Math::BigInt->new(1) << (2*$k);
    last if $k > 8512;

    my @new_x_list;
    foreach my $x (@x_list) {
      my $x2 = $x->to_base(2);
      printf "k=%-3d  x=%*s\n", $k, $k,$x2;

      my @n_list;
      foreach my $xend (0, 1) {
        my $y = Math::BigInt->new(0);
        my ($x_un,$y_un) = transform_to_horizontal($x+$xend,$y, -2*$k);
        ### transformed: "$x_un, $y_un"
        my @this = $path->xy_to_n_list($x_un,$y_un);
        print "     xend=$xend this ",join(', ',@this),"\n";
        push @n_list, @this;
      }
      @n_list = sort {$a<=>$b} @n_list;
      print "     visits ",scalar(@n_list),"\n";
      @n_list = List::Util::uniqstr(@n_list);
      print "     visits ",scalar(@n_list)," distinct\n";
      @n_list = map {$_,$_-1} @n_list;
      @n_list = grep {$_>0} @n_list;
      @n_list = grep {$_<$n_hi} @n_list;
      @n_list = sort {$a<=>$b} @n_list;
      print "     visits and prev ",scalar(@n_list),"\n";

      foreach my $n (@n_list) {
        my $xn = $n >> $k;
        my $desc;
        if ($xn == $x) {
          $desc = "YES";
          push @new_x_list, $n;
        } else {
          # my $xn2 = cnv($xn,10,2);
          # $desc = "NO  (xn=$xn2)";
          $desc = "NO";
        }
        my $n2 = $n->to_base(2);
        $n2 = sprintf '%*s', 2*$k, $n2;
        substr($n2,-$k,0, '_');
        print "       n=$n2  $desc\n";
      }
    }
    $k <<= 1;
    @x_list = List::Util::uniqstr(@new_x_list);
    print "\n";
  }
  exit 0;
}

{
  my $str = <<HERE;
1101
0000

11111101
000000000000
1101
000000000000000000000000

1101
0000
111111111111111111111111111111111111111111111101
0000

1101
00000000
11111111111111111111111111111111111111111111111111111111111111111111111111111111111111111101
0000

11111101
0000
1101
0000000000000000

111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111101
0000
11111101
000000000000

1101
00000000
1101
HERE
  my $p = 0;
  while ($str =~ /(0+)|(1+01)/g) {
    if (defined $1) { $p += length($1); }
    if (defined $2) { 
      my $len = length($2);
      print "[$p,$len]\n";
      $p += $len;
    }
  }
  exit 0;
}



{
  # X segment equal to its N high part, fixed point ffixed
  # 11010000111111010000000000001101000000000000000000000000110100001111111...
  # 110100001111
  #
  # k=24  x=110100001111  n= 110100001111_110011111100 xend=0 noffset=-1  YES
  # k=24  x=110100001111  n= 110100001111_110011111101 xend=0 noffset= 0  AGAIN
  # k=24  x=110100001111  n= 110100001111_110011111111 xend=1 noffset=-1  AGAIN
  # k=24  x=110100001111  n= 110100001111_110100000000 xend=1 noffset= 0  AGAIN
  #
  #          n+1             m-1                   |              |
  #           |               ^                     ----           ----
  #           v               |                         |              |
  #         / n                 \                       |              |
  # n-1 --->    <-----+-----   m ---> m+1           ----*    +    *---/ /---
  #            /            \                 |    |       0   1  |    |
  #           ^ arm 3         |               |    |              |    |
  #           |               |                ----                ----
  #    x=110100001111     x=110100010000
  #         n odd             m even
  #
  #                b 1- -    1---
  #                  110100001111
  # n-1==from_binary(110100001111110011111100)
  # n   =from_binary(110100001111110011111101);
  # n+1==from_binary(110100001111110011111110)
  # dragon_point(n-1) == from_binary(110100001110)
  # dragon_point(n)   == from_binary(110100001111)
  # dragon_point(n+1) == from_binary(110100001111) + I
  # dragon_unpoint(from_binary(110100001111),  1) == [3, 49479423]
  # dragon_unpoint(from_binary(110100001111), -1) == [0, n]
  # dir(n-1) % 4 == 0
  # dir(n) % 4 == 1
  # turn(n) == 1
  # dir(n+1) % 4 == 0
  # turn(n+1) == -1
  # dragon_point(n+1) == from_binary(110100001111) + I
  #
  #                  110100001111
  # m-1==from_binary(110100001111110011111111)
  # m   =from_binary(110100001111110100000000);
  # m+1==from_binary(110100001111110100000001)
  # dragon_unpoint(from_binary(110100010000),  1) == [0, m]
  # dragon_unpoint(from_binary(110100010000), -1) == [3, 49479424]
  # turn(m) == 1
  # dir(m) % 4 == 0
  # dragon_point(m-1) == from_binary(110100010000) + I
  # dragon_point(m)   == from_binary(110100010000)
  # dragon_point(m+1) == from_binary(110100010001)

  # k=500
  # x=1101000011111101000000000000110100000000000000000000000011010000111111111111111111111111111111111111111111111101000011010000000011111111111111111111111111111111111111111111111111111111111111111111111111111111111111111101000011111101000011010000000000
  # n=1101000011111101000000000000110100000000000000000000000011010000111111111111111111111111111111111111111111111101000011010000000011111111111111111111111111111111111111111111111111111111111111111111111111111111111111111101000011111101000011010000000000_0000001111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111010000111111010000000000001101000000001101000100000000000000000001
  # xend=1 noffset= 0  AGAIN

  # k=500
  # x=1101 0000
  #   11111101 0000 0000 0000
  #   1101 0000 0000 0000 0000 0000 0000
  #   1101 0000
  #   111111111111111111111111111111111111111111111101 0000
  #   1101 00000000
  #   11111111111111111111111111111111111111111111111111111111111111111111111111111111111111111101 0000
  #   11111101 0000
  #   1101 0000000000
  # n=1101     0000
  #   11111101 0000 0000 0000
  #   1101     0000 0000 0000 0000 0000 0000
  #   1101     0000
  #   111111111111111111111111111111111111111111111101 0000
  #   1101     0000 0000
  #   11111111111111111111111111111111111111111111111111111111111111111111111111111111111111111101 0000
  #   11111101 0000
  #   1101     0000 0000 00_00 0000
  #   111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111101 0000
  #   11111101 0000 0000 0000
  #   1101     0000 0000
  #   1101000100000000000000000001
  # hex digits F, D, 0
  #

  # k=8  n=11001111  x=1100 y=1  dxdy=0,-1
  # k=8  n=11010000  x=1100 y=0  dxdy=1,0   E
  # k=8  n=11010001  x=1101 y=0  dxdy=0,1

  # k=20  n=11010000111111001111  x=1101000100 y=-1  dxdy=0,1
  # k=20  n=11010000111111010000  x=1101000100 y=0  dxdy=-1,0   W
  # k=20  n=11010000111111010001  x=1101000011 y=0  dxdy=0,-1


  my $path = Math::PlanePath::DragonCurve->new;
  my $k = 0;
  my $n_hi = Math::BigInt->new(1);
  my @x_list = (Math::BigInt->new(0));
  for (;;) {
    ### x_list: join(', ',@x_list)
    # print "x_list ",join(', ',@x_list),"\n";
    print "x_list count ",scalar(@x_list),"\n";
    $k += 2;
    $n_hi <<= 2;
    last if $k > 500;

    my $h = $k/2;
    my @new_x_list;
    my %seen_x;
    my $yes = 0;
    foreach my $x_parent (@x_list) {
      ### x_parent: "$x_parent"
      next if $k==12 && ($x_parent <= 2 || $x_parent >= 30);
    BIT: foreach my $bit (0,1) {
        my $xstart = 2*$x_parent + $bit;
        foreach my $xend (0, 1) {
          my $x = $xstart + $xend;
          my $y = Math::BigInt->new(0);
          my @n_list = $path->xy_to_n_list(transform_to_horizontal($x,$y, -$k));
          ### centreline x: "$x"
          ### untransformed: join(', ',transform_to_horizontal($x,$y, -$k))
          ### n_list: join(', ',@n_list)
          @n_list = grep {$_<=$n_hi} @n_list;
          foreach my $n (@n_list) {
            foreach my $noffset (-1, 0) {
              my $n = $n + $noffset;
              next if $n < 0;
              my $xn = $n >> $h;

              my $desc;
              if ($xn == $xstart) {
                $desc = "YES";
                $yes++;
                if ($seen_x{$xstart}++) {
                  $desc = "AGAIN";
                } else {
                  push @new_x_list, $xstart;
                }
              } else {
                # my $xn2 = cnv($xn,10,2);
                # $desc = "NO  (xn=$xn2)";
                $desc = "NO";
              }
              my $xstart2 = cnv($xstart,10,2);
              my $n2 = cnv($n,10,2);
              $n2 = sprintf '%*s', $k+1, $n2;
              substr($n2,$h+1,0, '_');
              if ($k % 4 == 0) {
                printf "k=%-3d  x=%*s\n      n=$n2 xend=$xend noffset=%2d  $desc\n",
                  $k, $h,$xstart2, $noffset;
              }
            }
          }
        }
      }
    }
    print "yes n count $yes\n";
    print "\n";
    @x_list = sort {$a<=>$b} @new_x_list;
  }
  exit 0;
}


{
  # centreline in binary, count

  foreach my $type ('point','segment') {
    my $f = CentrelinePred_flat($type);
    $f = $f->MyFLAT::minimize(verbose=>1);
    my @counts = $f->MyFLAT::FLAT_count_contains(20);
    say join(', ',@counts);
  }
  exit 0;
}
{
  {
    my $f = CentrelinePred_flat();
    $f = $f->intersect(fRpred_flat());
    $f = $f->MyFLAT::minimize(verbose=>1);
    # $f->MyFLAT::view;
    $f = $f->MyFLAT::blocks(4);
    $f = $f->MyFLAT::FLAT_rename(symbols_func => sub {
                                   my ($binary) = @_;
                                   return cnv($binary,2,16);
                                 });
    $f->MyFLAT::view;
  }
  exit 0;
}

{
  # points on centre line predicate - MyFSM

  $|=1;

  my %table = (h    => {0=>'h0', 1=>'h1'},
               h0   => {0=>'s', 1=>    0},
               h1   => {0=> 'e', 1=>'hrev' },
               s    => {0=>'s0', 1=>'s1'},
               s0   => {0=>'h', 1=>'erev'},
               s1   => {0=>  0,  1=>     0 },
               e    => {0=>'e0', 1=>'e1'},
               e0   => {0=>  0, 1=>    0},
               e1   => {0=>  0,  1=>'srev' },
               hrev  => {0=>'hrev0', 1=>'hrev1'},
               hrev0 => {0=>'h', 1=>'erev'},
               hrev1 => {0=>  0,  1=>'srev' },
               srev  => {0=>'srev0', 1=>'srev1'},
               srev0 => {0=>  0, 1=>    0},
               srev1 => {0=> 'e', 1=>'hrev' },
               erev  => {0=>'erev0', 1=>'erev1'},
               erev0 => {0=>'s', 1=>    0},
               erev1 => {0=>  0,  1=>     0 },
               0    => {0=>0, 1=>0 },
              );
  my %accepting = (
                   h    => 1,
                   hrev => 1,
                   s    => 1,
                   erev => 1,
                   0    => 0,
                  );
  require MyFSM;
  my $fsm = MyFSM->new(table => \%table,
                       initial => 'h',
                       accepting => \%accepting,
                       title => 'CentrelinePred',
                       flow => 'east',
                      );
  $fsm->simplify (verbose => 1);
  # $fsm = $fsm->reverse;
  # $fsm->simplify (verbose => 1);
  # $fsm->renumber_states;
  $fsm->view;
  # $fsm->print_tikz;
  exit 0;
}

{
  # n=11010000 11111101 00000000 00001101 00000000 00000000 00000000 11010000 11111111 11111111 11111111 11111111 11111111 11111101 00001101 00000000 11111111111111111111111111111111111111111111111111111111111111111111111111111111111111111101000011111101000011010000000000_0000001111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111010000111111010000000000001101000000001101000100000000000000000001
  # b^-1-b^-2 -b^-4 == 3/4        .11
  # b^-8*(b^-1-b^-2-b^-3-b^-4-b^-5-b^-6 -b^-8) == 17/256  \\ 1 0001

  # dragon_fpoint(from_binary(1100)/2^4) == 1/2
  # dragon_fpoint(from_binary(1101)/2^4) == 3/4
  # dragon_fpoint(from_binary(0000000011111101)/2^16) == 17/256
  # my(k=4);dragon_fpoint((2^k-3)/2^k) ==  3/4
  # my(k=8);dragon_fpoint((2^k-3)/2^k) == 17/16
  # my(k=16);dragon_fpoint((2^k-3)/2^k) == 257/256
  # my(k=64);dragon_fpoint((2^k-3)/2^k) == 1+1/2^32
  # 3/4+17/256 == 209/256  \\  1101 0001
  exit 0;
}




{
  # ffixed is on boundary
  my $str = '110100001111110100000000000011010000000000000000000000001101000011111111111111111111111111111111111111111111110100001101000';
  foreach my $len (0 .. length($str)) {
    my $sub = substr($str,0,$len);
    print "$len  ",fRpred_flat()->contains($sub) ? 1 : 0, "  $sub\n";
  }
  exit 0;
}

{
  # N high equal X by path

  my $path = Math::PlanePath::DragonCurve->new;
  for (my $k=0; $k < 28; $k+=2) {
    my ($n_lo, $n_hi) = $path->level_to_n_range($k);
    my $h = $k/2;
    foreach my $n ($n_lo .. $n_hi) {
      my ($x,$y) = $path->n_to_xy($n);
      my ($dx,$dy) = $path->n_to_dxdy($n);
      ($x,$y) = transform_to_horizontal($x,$y, $k);
      ($dx,$dy) = transform_to_horizontal($dx,$dy, $k);
      next unless $y == 0 || $y+$dy == 0;
      if (($n >> $h) == $x) {
        my $n2 = cnv($n,10,2);
        my $x2 = cnv($x,10,2);
        printf "k=$k  n=%*s  x=$x2\n", $k,$n2;
      }
    }
    print "\n";
  }
  exit 0;
}

{
  # points on centre line predicate by base-4

  #                                   @          @
  #        |s         |e              |erev      |srev
  #        |          |               |          |        e    = N at start
  #    ----*---h----->*---        ----*<---hrev--@---     hrev = N at arrow
  #                                                       srev = N at arrow
  #        *      *<----*               *----->*               points on start
  #        |      ^                     |      ^
  #        v      |                     v      |
  #        *<-----*              *----->*      *

  my %table = (h    => {0=>'s', 1=>    0,  2=> 'e', 3=>'hrev' },
               s    => {0=>'h', 1=>'erev', 2=>  0,  3=>     0 },
               e    => {0=>  0, 1=>    0,  2=>  0,  3=>'srev' },
               hrev => {0=>'h', 1=>'erev', 2=>  0,  3=>'srev' },
               srev => {0=>  0, 1=>    0,  2=> 'e', 3=>'hrev' },
               erev => {0=>'s', 1=>    0,  2=>  0,  3=>     0 },
               0    => {0=>0, 1=>0, 2=>0, 3=>0 },
              );
  my %accepting = (
                   h    => 1,
                   hrev => 1,
                   s    => 1,
                   erev => 1,
                   0    => 0,
                  );
  require MyFSM;
  my $fsm = MyFSM->new(table => \%table,
                       initial => 'h',
                       accepting => \%accepting,
                       title => 'CentrelinePred',
                       flow => 'east',
                      );
  $fsm->simplify (verbose => 1);
  # $fsm = $fsm->reverse;
  # $fsm->simplify (verbose => 1);
  # $fsm->renumber_states;
  $fsm->view;
  $fsm->print_tikz;


  my $path = Math::PlanePath::DragonCurve->new;
  for (my $k=0; $k < 16; $k+=1) {
    my ($n_lo, $n_hi) = $path->level_to_n_range($k);
    foreach my $n ($n_lo .. $n_hi-1) {
      my ($x,$y) = $path->n_to_xy($n);
      ($x,$y) = transform_to_horizontal($x,$y, $k);
      my $want = ($y==0) ? 1 : 0;
      my $got = CentrelinePred($n,$k) ? 1 : 0;
      if ($got != $want) {
        my $n2 = cnv($n,10,2);
        print "k=$k  n=$n2  got=$got want=$want\n";
      }
    }
  }
  exit 0;


  sub CentrelinePred {
    my ($n, $k) = @_;
    ### CentrelinePred(): "n=$n  k=$k"
    if ($k % 2) {
      if ($n % 2) { return 0; }
      $n /= 2;
      $k--;
    }
    my $state = 'h';
    my @digits = digit_split_lowtohigh($n,4);
    $#digits = $k/2-1;
    ### digits: reverse @digits
    foreach my $digit (reverse @digits) {
      $state = $table{$state}->{$digit||0} // die;
      $state || return 0;
    }
    ### final: $state
    return $accepting{$state};
  }
}

{
  # N on centreline with N nearly = X, cf fpoint(f)==f

  my $path = Math::PlanePath::DragonCurve->new;
  foreach my $k (8 * 5) {
    my ($n_lo, $n_hi) = $path->level_to_n_range($k);
    my ($x_hi,$y_hi) = $path->n_to_xy($n_hi);
    print "end at $x_hi,$y_hi\n";
    for (my $x = int(0.8*$x_hi); $x < $x_hi; $x++) {
      my $n = $path->xy_to_n($x,0) // next;
      $n < $n_hi or next;
      my $nx = $n *$x_hi/$n_hi;
      if (abs($nx-$x) < 3) {
        my $xf = $x / $x_hi;
        print "x=$x  ",($n/$n_hi)/($x/$x_hi),"  $nx  xf=$xf\n";
        my $n2 = cnv($n,10,2);
        my $x2 = cnv($x,10,2);
        print "  n=$n2\n";
        print "  x=$x2\n";
      }
    }
  }
  exit 0;
}


