#!/usr/bin/perl -w

# Copyright 2020 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.



# x-f extents depend on x and f of the sub-curve.
# Think once f amounts smaller than x then fall into end-most unit square.



use 5.010;
use strict;
use List::Util 'min','max';
use Math::PlanePath::DragonCurve 117; # v.117 for level_to_n_range()

use FindBin;
use lib "$FindBin::Bin/../tools";
use MyDragon 'dir', 'points_str';
use MyPlanar;
use MyPrintwrap ':all';

# uncomment this to run the ### lines
# use Smart::Comments;


# 2**$num_blocks_pow2 many blocks drawn across the horizontal axis.
# 2**$block_width_pow2 subcurve segments considered in each block.
# Subcurves help get a more accurate x-f.
#
my $num_blocks_pow2 = 2;
my $block_width_pow2 = 0;


print "  % generated by tools/$FindBin::Script\n";

my $level = $num_blocks_pow2 + $block_width_pow2;
$level % 2 == 0 or die "want level even";
my $level_rot = $level/2;   # back to horizontal
my $scale = 2**($level/2);

my $path = Math::PlanePath::DragonCurve->new;
my ($n_lo,$n_hi) = $path->level_to_n_range($level);

# truncate to at most 7 digits of precision (8 including decimal point)
sub decimals {
  my ($x) = @_;
  $x = sprintf '%.8f', $x;
  $x =~ s/(\d)\.0*$/$1/;   # integer no trailing 0s or decimal point
  $x =~ s/(\.\d*?)0+/$1/;    # trailing 0s after another digit
  $x =~ s/([0-9.]{8})[0-9]+$/$1/;
  return $x;
}
### decimals: decimals('1.25')
decimals('1') eq '1' or die;
decimals('1.25') eq '1.25' or die;
decimals('999999.125') eq '999999.1' or die;

# dir + (4 if reverse)
my @xf_extent_max60 = ( 16,  24,   16,   9,    28,  16,    9,  12);
my @xf_extent_min60 = (-28, -76, -129, -72,   -16, -84, -136, -69);
# GP-Test  16/60 == 4/15
# GP-Test  -28/60 == -7/15

my @xf_lo;
my @xf_hi;
my @y_lo;
my @y_hi;
my $f_scale = 60 / 2**($level/2);
foreach my $n ($n_lo .. $n_hi-1) {
  my ($x,$y) = $path->n_to_xy($n);
  my $dir = dir($n) % 4;
  my $seg = MyDragon::seg_normalize("$x,$y,$dir");

  my $hull_unrot60 = MyDragon::seg_to_hull6_points($seg);
  $hull_unrot60 = MyPlanar::points_scale($hull_unrot60, 10);
  my $hull60 = MyPlanar::points_rotate_plus90($hull_unrot60, -$level_rot);

  my ($x1,$y1, $x2,$y2) = MyPlanar::points_bbox_xyxy($hull60);
  my $block = $n >> $block_width_pow2;
  $y_lo[$block] = min($y_lo[$block]//$y1, $y1);
  $y_hi[$block] = max($y_hi[$block]//$y2, $y2);

  ($x,$y) = MyDragon::xy_rotate_plus90($x,$y, -$level_rot);
  my $T = ($dir - $level_rot)%4 + 4*($n%2);
  my $f = $n*$f_scale;
  my $xf = 60*$x - $f;
  $xf_lo[$block] = min($xf_lo[$block]//$xf,  $xf + $xf_extent_min60[$T]);
  $xf_hi[$block] = max($xf_hi[$block]//$xf,  $xf + $xf_extent_max60[$T]);

  ### $n
  ### $block
  ### $seg
  ### x60 rot: 60*$x
  ### $f
  ### $xf
  ### $T
  ### xf this lo: $xf + $xf_extent_min60[$T]
  ### xf this hi: $xf + $xf_extent_max60[$T]
  ### hull y1: $y1
  ### hull y2: $y2
  ### hull60 unrot: points_str($hull_unrot60)
  ### hull60 horizontal: points_str($hull60)

  if (0) {
    # diagnostic to see hulls around segments
    $Printwrap = 2;
    printwrap("\\draw");
    printwrap_poly($hull60);
    printwrap(";\n");
    printwrap("\\draw ($x1,$y1) rectangle ($x2,$y2);\n");
  }

  if ($n == $n_hi-1) {
    my ($x,$y) = $path->n_to_xy($n+1);
    ($x,$y) = MyDragon::xy_rotate_plus90($x,$y, -$level_rot);
    $f+$f_scale == $x*60 or die;
  }
}
scalar(@xf_lo) == 1<<$num_blocks_pow2 or die;
scalar(@xf_hi) == 1<<$num_blocks_pow2 or die;
scalar(@y_lo) == 1<<$num_blocks_pow2 or die;
scalar(@y_hi) == 1<<$num_blocks_pow2 or die;

### @y_hi
### @y_lo

### @xf_hi
### @xf_lo

sub lohi_to_points {
  my ($alo,$ahi) = @_;
  return [(map { [$_, $alo->[$_]], [$_+1, $alo->[$_]] } 0 .. $#$alo),
          (map { [$_+1, $ahi->[$_]], [$_, $ahi->[$_]] } reverse(0 .. $#$ahi))];
}

my $xf_points = lohi_to_points(\@xf_lo, \@xf_hi);
my $y_points  = lohi_to_points(\@y_lo,  \@y_hi);
### $xf_points
### $y_points

my $xf_gpc = MyPlanar::points_to_gpc($xf_points);
my $y_gpc  = MyPlanar::points_to_gpc($y_points);
my $both_gpc = Math::Geometry::Planar::GpcClip('INTERSECTION',$xf_gpc,$y_gpc);

# $xf_gpc = Math::Geometry::Planar::GpcClip('DIFFERENCE',
#                                           $xf_gpc, $both_gpc);
$y_gpc = Math::Geometry::Planar::GpcClip('DIFFERENCE',
                                         $y_gpc, $both_gpc);

# avoid TeX dimension overflow
my $h_scale_pow2 = min(8, $num_blocks_pow2);
my $h_scale_mul = 1/2**(  $num_blocks_pow2 - $h_scale_pow2);

my $v_scale_pow2 = min(3, $level_rot);
my $v_scale_mul = 1/2**(  $level_rot - $v_scale_pow2);

### $h_scale_pow2
### $h_scale_mul
### $v_scale_pow2
### $v_scale_mul

my $total_num_points = 0;
print "  \\begin{scope}[xscale={1/2^$h_scale_pow2},yscale={1/(60*2^$v_scale_pow2)},even odd rule]\n";
foreach my $elem (
                  # ['y',    $y_gpc],
                  # ['both', $both_gpc],
                  ['x-f',  $xf_gpc],
                 ) {
  my ($name,$gpc) = @$elem;
  my @contours = Math::Geometry::Planar::Gpc2Polygons($gpc);
  my $total_holes = 0;
  print "  \\begin{scope}[my $name]\n";
  foreach my $contour (@contours) {  # Math::Geometry::Planar contour
    ### $contour
    # Second and subsequent polygon of the contour are holes.
    # Think their points already cycle the opposite way as required by pgf
    # default \fill[nonzero rule], but using \fill[even odd rule] works either
    # way.
    $Printwrap = 4;
    printwrap(" \\fill");
    $Printwrap += 2;
    my $polygons = $contour->polygons;  # arrayref of points arrayrefs
    $total_holes += scalar(@$polygons)-1;
    my $separator = '';
    foreach my $points (@$polygons) {
      ### $points
      my $scaled_points = MyPlanar::points_xyscale($points,
                                                   $h_scale_mul, $v_scale_mul);
      foreach my $p (@$scaled_points) {
        $p->[1] = decimals($p->[1]);
      }
      printwrap($separator);
      $separator = "\n";
      printwrap_poly($scaled_points);
      $total_num_points += scalar(@$points);
    }
    printwrap(";\n");
  }
  $total_holes == 0 or die;
  print "  \\end{scope}\n";
}
print "  \\end{scope}\n";

print "  % total num points $total_num_points\n";
{
  my ($x1,$y1, $x2,$y2) = MyPlanar::points_bbox_xyxy($xf_points);
  my $scaled_x1 = $x1 * $h_scale_mul;
  my $scaled_x2 = $x2 * $h_scale_mul;
  print "  % f extents $x1 .. $x2 scaled $scaled_x1 .. $scaled_x2\n";
}
{
  my ($x1,$y1, $x2,$y2) = MyPlanar::points_bbox_xyxy($xf_points);
  my $scaled_y1 = $y1 * $v_scale_mul;
  my $scaled_y2 = $y2 * $v_scale_mul;
  print "  % xf extents $y1 .. $y2 scaled $scaled_y1 .. $scaled_y2\n";
}
{
  my ($x1,$y1, $x2,$y2) = MyPlanar::points_bbox_xyxy($y_points);
  my $scaled_y1 = $y1 * $v_scale_mul;
  my $scaled_y2 = $y2 * $v_scale_mul;
  print "  % y extents $y1 .. $y2 scaled $scaled_y1 .. $scaled_y2\n";
}

print "  % end generated\n";
exit 0;

__END__

foreach my $t (@xf_lo, @xf_hi) {
  if ($t) {
    $t = sprintf '%.6f', $t;  # mutate arrays
    $t =~ s/(\.\d)0+/$1/;
    $t =~ s/^0(\.\d)/$1/;
  }
}

my $join = ' ';
foreach my $block (0 .. $#t_lo) {
  my $f1 = ($block  )*$h_scale_mul;
  my $f2 = ($block+1)*$h_scale_mul;
  printwrap("$join($f1,$t_lo[$block])");
  printwrap("--($f2,$t_lo[$block])");
  $join = '--';
}
printwrap_newline();
printwrap(" [shift={(line-width-hi)}]");

$join = ' --';
foreach my $block (reverse 0 .. $#t_hi) {
  my $f1 = ($block  )*$h_scale_mul;
  my $f2 = ($block+1)*$h_scale_mul;
  printwrap("$join($f2,$t_hi[$block])");
  printwrap("--($f1,$t_hi[$block])");
  $join = '--';
}
printwrap("--cycle;\n");
