#!/usr/bin/perl -w

# Copyright 2016, 2017, 2019, 2020 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  See the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.

use strict;
use 5.006;
use Math::PlanePath::DragonCurve;
use Test::More;

use FindBin;
use lib $FindBin::Bin;
use MyDragon ':all';

use MyTestHelpers;
BEGIN { MyTestHelpers::nowarnings(); }

plan tests => 1222;

# uncomment this to run the ### lines
# use Smart::Comments;


#------------------------------------------------------------------------------
# xy_to_gp_str()

is(xy_to_gp_str(0,0), '0');
is(xy_to_gp_str(1,0), '1');
is(xy_to_gp_str(6,0), '6');
is(xy_to_gp_str(-6,0), '-6');

is(xy_to_gp_str(0,1), 'I');
is(xy_to_gp_str(0,-1), '-I');
is(xy_to_gp_str(0,6), '6*I');
is(xy_to_gp_str(0,-6), '-6*I');

is(xy_to_gp_str(8,1), '8+I');
is(xy_to_gp_str(8,-1), '8-I');
is(xy_to_gp_str(8,6), '8+6*I');
is(xy_to_gp_str(8,-6), '8-6*I');


#------------------------------------------------------------------------------
# turn(), TurnLpred(), TurnRpred()

foreach my $n (1 .. 100) {
  my $t = turn($n);
  is (TurnLpred($n), $t==1 ? 1 : 0);
  is (TurnRpred($n), $t==-1 ? 1 : 0);
}

#------------------------------------------------------------------------------
# bitlength()

foreach my $i (0 .. 100) {
  my $str = sprintf '%b', $i;
  if ($str eq '0') { $str = ''; }
  is (bitlength($i), length($str));
}


#------------------------------------------------------------------------------
# ffixed_string()

is (ffixed_string(), '');
is (ffixed_string(max_m => 0), '');
is (ffixed_string(max_m => 1), '1101');
is (ffixed_string(max_m => 2), '11010000');
is (ffixed_string(max_m => 3), '1101000011111101');
is (ffixed_string(num_runs => 0), '');
is (ffixed_string(num_runs => 1), '1101');
is (ffixed_string(num_runs => 2), '1101000011111101');
foreach my $num_bits (0 .. 50) {
  is (length(ffixed_string(num_bits => $num_bits)), $num_bits);
}

is (ffixed_string(num_runs => 1, num_bits => 3), '1101');
is (ffixed_string(max_m => 1, num_bits => 3), '1101');
is (ffixed_string(max_m => 2, num_runs => 1, num_bits => 3), '11010000');


#------------------------------------------------------------------------------
# Rpred()

{
  my $path = DragonCurve_path();
  my $bad = 0;
  foreach my $n (0 .. 1024) {
    my $by_func = Rpred($n) ?1:0;
    my $by_path = $path->_UNDOCUMENTED__n_segment_is_right_boundary($n) ?1:0;
    unless ($by_func == $by_path) {
      $bad++;
    }
  }
  is($bad, 0);
}


#------------------------------------------------------------------------------
# fVisits non-overlap of s3, s4, s7, s11 with m

{
  require MyPlanar;
  my $m_points = $hull6_points;
  my $m_gpc = MyPlanar::points_to_gpc($hull6_points);
  foreach my $s (3,4,7,11) {
    my $seg = $fVisits_adjacent_list[$s];
    my $seg_points = MyDragon::seg_to_hull6_points($seg);
    my $seg_gpc = MyPlanar::points_to_gpc($seg_points);
    ok( ! MyPlanar::gpc_has_overlap($m_gpc, $seg_gpc),
        "s$s no area overlap of m");
    is(points_num_common($m_points, $seg_points), 1,
        "s$s one point in common");

    my @ints = poly_segments_intersection($m_points, $seg_points);
    ### @ints
    is(scalar(@ints), 1);
    my $int = $ints[0];
    ok(MyPlanar::poly_is_point($int), "s$s intersection segment is one point");
  }
}

sub poly_segments_intersection {
  my ($points1, $points2) = @_;
  my @ret;
  foreach my $i (0 .. $#$points1) {
    foreach my $j (0 .. $#$points2) {
      if (my $int = MyPlanar::segment_intersection_segment
          ([$points1->[$i-1], $points1->[$i],
            $points2->[$j-1], $points2->[$j]])) {
        unless (@ret && points_eq($ret[-1],$int)) {
          push @ret, $int;
        }
      }
    }
  }
  return @ret;
}

sub points_eq {
  my ($points1, $points2) = @_;
  unless ($#$points1 == $#$points2) { return 0; }
  foreach my $i (0 .. $#$points1) {
    unless (MyPlanar::point_eq($points1->[$i], $points2->[$i])) {
      return 0;
    }
  }
  return 1;
}

# $points1 and $points2 are arrayrefs [ [$x,$y], ... ].
# Return the number of points in $points1 which are also in $points2.
sub points_num_common {
  my ($points1, $points2) = @_;
  my $ret = 0;
  foreach my $elem1 (@$points1) {
    if (grep {MyPlanar::point_eq($elem1,$_)} @$points2) {
      $ret++;
    }
  }
  return $ret;
}


#------------------------------------------------------------------------------
# 7ths Sevenths

{
  my $count = 0;
  foreach my $conf (@triple_7ths_segs) {
    next unless defined $conf;
    ok ($conf->contains('0,0,0'));
    $count++;
  }
  is ($count, 6);
}

{
  # $triple_7ths_and_next_segs  union at positions before and after
  my $u = $triple_7ths_union_segs;
  my $r1 = $triple_7ths_union_segs->rotate(1);
  my $r3 = $triple_7ths_union_segs->rotate(-1);
  $u = $u->union($r1,
                 $r3,
                 $r1->translate(1,-1),
                 $r3->translate(1,1));
  ok ($u->equals($triple_7ths_and_next_segs))
  # $u->print;
}
{
  my $c = $triple_7ths_and_next_within;
  is ($c->num_segs, 15);
  ok ( $c->contains('0,0,0'));

  foreach my $bit (0,1) {
    my $e = $c->expand($bit);
    # $e->print;
    ok ($c->is_subset($triple_7ths_and_next_within));
  }
}

MyDragon::triple_7ths_surround_init();



{
  # Everything 3 segments away

  # conf 1 away side=0
  #    0,0,0 0,0,3 0,0,1 1,-1,1 1,1,3
  #    *     *
  #    ^     |
  #    |     |
  #    |     v
  #    O====>*
  #    |     ^
  #    |     |
  #    v     |
  #    *     *

  # conf 2 away side=0
  #    0,0,0 0,0,3 0,0,1 1,-1,1 1,1,3 0,0,2 -1,-1,0 1,-1,2 1,1,2 -1,1,0 1,-1,0 2,0,2 1,1,0
  #    *---->*<----*---->*
  #          ^     |
  #          |     |
  #          |     v
  #    *<----O====>*<----*
  #          |     ^
  #          |     |
  #          v     |
  #    *---->*<----*---->*

  # conf 3 away side=0
  #    0,0,0 0,0,3 0,0,1 1,-1,1 1,1,3 0,0,2 -1,-1,0 1,-1,2 1,1,2 -1,1,0 1,-1,0 2,0,2 1,1,0 -1,1,3 -1,-1,1 -1,-1,3 0,-2,1 1,-1,3 1,1,1 0,2,3 -1,1,1 2,-2,1 2,0,3 2,0,1 2,2,3
  #    *     *     *     *
  #    ^     |     ^     |
  #    |     |     |     |
  #    |     v     |     v
  #    *---->*<----*---->*
  #    |     ^     |     ^
  #    |     |     |     |
  #    v     |     v     |
  #    *<----O====>*<----*
  #    ^     |     ^     |
  #    |     |     |     |
  #    |     v     |     v
  #    *---->*<----*---->*
  #    |     ^     |     ^
  #    |     |     |     |
  #    v     |     v     |
  #    *     *     *     *

  my $u = MyDragonConf->initial;
  foreach my $i (1 .. 3) {
    # $u->print;
    $u = $u->union_in_positions($conf_before_and_after)
      ->set_name("$i away");
  }
  # $u->print;

  # does not expand within self
  # foreach my $bit (0,1) {
  #   ok($u->is_subset( $u->expand($bit) ),
  #      "$u->{'name'}, bit=$bit");
  # }
}

{
  # Everything 3 segments away

  # forward 3                                                   *           
  #                                                             |           
  #                                                             v           
  #                                        *--->*               *<---*      
  #                                        |    ^                    |      
  #                                        v    |                    v      
  # after   t e !t     *===>*    *    *===>*    *    *===>*     *===>*      
  #                         ^    |                        ^   
  #                         |    v                        |   
  #                         *--->*                   *<---*   
  #                                                  ^        
  #                                                  |        
  #                                                  *        
  #                          
  # e1 o e2   before
  #
  #    *    *===>*          *===>*      
  #    ^    |               |    
  #    |    v               v           
  #    *--->*               *<---*            
  #                              |
  #         *===>*               v
  #         |                    *
  #         v
  #    *--->*           
  #    |
  #    v
  #    *

  my $u = MyDragonConf->initial;

  foreach my $i (1 .. 3) {
    # $u->print;
    $u = $u->union_in_positions($conf_before_and_after)
      ->set_name("$i away");
  }
  # $u->print;
}



#------------------------------------------------------------------------------
# $hullstouch_segs

# check that the expand(0) and expand(1) of $conf have all of $conf as subsets
sub ok_conf_expand_within_self {
  my ($conf) = @_;
  foreach my $bit (0,1) {
    ok($conf->is_subset( $conf->expand($bit) ),
       "$conf->{'name'}, bit=$bit");
  }
}

ok_conf_expand_within_self($hullstouch_segs);

{
  my $top_points = MyDragon::seg_to_hull6_points('1,1,0');
  is_deeply($top_points,
            [ [ 5+6, -3+6],
              [ 7+6,  0+6],
              [ 7+6,  1+6],
              [ 6+6,  2+6],
              [ 4+6,  2+6],
              [-1+6,  1+6],
              [-2+6,  0+6],
              [-2+6, -2+6],
              [ 0+6, -4+6],  # equals [6,2]
              [ 4+6, -4+6],
            ]);
  ok (MyPlanar::points_has_boundary_touch($hull6_points, $top_points));
  ok (MyPlanar::points_has_boundary_touch($hull6_points, [ [6,2],[7,2],[6,3] ]));
}

{
  # $hullstouch_segs versus search for which hulls touch or overlap

  # $hullstouch_segs->print;
  require MyPlanar;
  my $m_points = $hull6_points;
  my $m_gpc = MyPlanar::points_to_gpc($hull6_points);
  my %segs;
  foreach my $x (-4 .. 4) {
    foreach my $y (-4 .. 4) {
      foreach my $dir (0 .. 3) {
        next unless ($x+$y)%2 == 0;
        my $seg = MyDragon::seg_normalize("$x,$y,$dir");
        my $seg_points = MyDragon::seg_to_hull6_points($seg);
        if (MyPlanar::gpc_has_overlap($m_gpc,
                                      MyPlanar::points_to_gpc($seg_points))
            || MyPlanar::points_has_boundary_touch($m_points, $seg_points)) {
          $segs{$seg} = undef;
        }
      }
    }
  }
  my $s = MyDragonConf->initial (segs => [keys %segs], side => 0);
  # $s->print;
  ok($s->equals($hullstouch_segs));
}

ok ($triple_15ths_union_segs->is_subset($hullstouch_segs),
   '15ths union is a subset of the hullstouch');
{
  # extra end segment e, remark in dragon.tex
  my $diff = $hullstouch_segs->difference($triple_15ths_union_segs);
  is ($diff->num_segs, 1,
      'hullstouch is 1 extra segment over 15ths union');
  ok($diff->equals(MyDragonConf->initial (segs => ['2,0,2'], side => 0)),
     'hullstouch extra over 15ths union is end segment 2,0,2');
}


#------------------------------------------------------------------------------
# Lnext_k()

{
  my $bad = 0;
  foreach my $k (0 .. 8) {
    my $Lnext = 0;
    foreach my $n (0 .. 2**$k-1) {
      if (Lpred_k($k,$n)) {
        if ($Lnext == $n) {
          $Lnext = Lnext_k($k,$n);
        } else {
          $bad++;
        }
      }
    }
    if (defined $Lnext) {
      $bad++;
    }
  }
  is ($bad, 0, "Lnext_k() vs Lpred_k()");
}


#------------------------------------------------------------------------------
# Rnext()

{
  my $Rnext = 0;
  my $bad = 0;
  foreach my $n (0 .. 2**14) {
    if (Rpred($n)) {
      if ($Rnext == $n) {
        $Rnext = Rnext($n);
      } else {
        $bad++;
      }
    }
  }
  is ($bad, 0, "Rnext() vs Rpred()");
}


#------------------------------------------------------------------------------
# MyDragonConf

{
  my $s = MyDragonConf->initial (name => '');
  is($s->num_segs, 1);
  ok($s->equals($s));

  foreach my $bit (0,1) {
    my $e = $s->expand($bit);
    is($e->num_segs, 2);
  }
  my $s_expand_0 = $s->expand(0);
  my $s_expand_1 = $s->expand(1);
  ok(! $s_expand_0->equals($s));
  ok(! $s_expand_1->equals($s));
  is($s_expand_0->{'name'}, '0');
  is($s_expand_1->{'name'}, '1');

  #   bit=0             bit=1
  #           2
  #           |          side=1
  #           v      end    start
  #   0 ----> 1       0 ----> 1
  # start    end              ^
  #    side=0                 |
  #                           2
  is($s_expand_0->key, 'side=0 0,0,0 1,1,3');
  is($s_expand_1->key, 'side=1 0,0,0 1,-1,1');

  # ok(  $s->is_boundary(0, $R_square));
  # ok(! $s->is_boundary(1, $R_square));
  # ok(! $s->is_boundary(0, $L_square));
  # ok(  $s->is_boundary(1, $L_square));

  # foreach my $c ($R_square, $L_square) {
  #     $c->print;
  #   foreach my $bit (0, 1) {
  #     $c->expand($bit)->print;
  #     ok( $c->is_subset( $c->expand($bit) ),
  #       "$c->{name} bit=$bit");
  #   }
  # }

  my $B = $R_square_segs->union($L_square_segs)->set_name('B squares');
  # $B->print;
  foreach my $bit (0, 1) {
    # $B->expand($bit)->print;
    ok( $B->is_subset( $B->expand($bit) ),  "$B->{name} bit=$bit");
  }
}

#              |
#              |
#              v
#        O---->
#        |
#        |
#  *     v     *
is (MyDragon::seg_transform_to('1,1,3', '1,1,3'), '0,0,0');
is (MyDragon::seg_transform_to('0,0,0', '1,1,3'), '1,-1,1');
is (MyDragon::seg_transform_to('0,0,3', '1,1,3'), '1,-1,0');

{
  my $s = MyDragonConf->next_positions(0);
  is($s->key, 'side=0 0,0,0');
}
{
  # *     *     *     *
  #       ^     |
  #       |     |
  #       |     v
  # *<----O====>*<----*
  #       |     ^
  #       |     |
  #       v     |
  # *     *     *     *
  my $s = MyDragonConf->next_positions(1);
  is($s->key, 'side=0 0,0,0 0,0,1 0,0,2 0,0,3 1,-1,1 1,1,3 2,0,2');
  is($s->num_segs, 7);
}
{
  # *     *     *     *     *     *     *     *
  #                   ^     |
  #                   |     |
  #                   |     v
  # *     *     *<----*---->*<----*     *     *
  #             ^     |     ^     |
  #             |     |     |     |
  #             |     v     |     v
  # *     *<----*---->*<----*---->*<----*     *
  #       ^     |     ^     |     ^     |
  #       |     |     |     |     |     |
  #       |     v     |     v     |     v
  # *<----*---->*<----O====>*<----*---->*<----*
  #       |     ^     |     ^     |     ^
  #       |     |     |     |     |     |
  #       v     |     v     |     v     |
  # *     *<----*---->*<----*---->*<----*     *
  #             |     ^     |     ^
  #             |     |     |     |
  #             v     |     v     |
  # *     *     *<----*---->*<----*     *     *
  #                   |     ^
  #                   |     |
  #                   v     |
  # *     *     *     *     *     *     *     *
  my $s = MyDragonConf->next_positions(3);
  is($s->num_segs, 47);
}

{
  # MyDragonCOnf transform_to_seg()
  my $s = MyDragonConf->initial (name => '', side => 1);
  my $t = $s->transform_to_seg('0,0,0');
  is ($t->{'side'}, 1);
}



{
  my $s = MyDragonConf->initial (name => '',
                                 want_next_segs => 8);
  my $s_expand_0 = $s->expand(0);
  is($s_expand_0->key, 'side=0 0,0,0 1,1,3 next=1,1,3');

  my $s_expand_1 = $s->expand(1);
  is($s_expand_1->key, 'side=1 0,0,0 1,-1,1');

  #   4
  #   |
  #   v
  #   3 <---- 2                2 ---> 1
  #           |                  \  / ^\
  #           v                   vv  |/
  #   0 ----> 1                       0

  my $s_expand_00 = $s_expand_0->expand(0);
  is($s_expand_00->key, 'side=0 0,0,0 0,2,3 1,1,2 1,1,3 next=1,1,3 1,1,2 0,2,3');

  my $s_expand_01 = $s_expand_0->expand(1);
  is($s_expand_01->key, 'side=1 -1,-1,0 0,0,0 0,0,3 1,-1,1 next=0,0,3 -1,-1,0');

  my $s_expand_10 = $s_expand_1->expand(0);
  is($s_expand_10->key, 'side=0 0,0,0 0,0,1 1,-1,1 1,1,2 next=1,-1,1');

  my $s_expand_11 = $s_expand_1->expand(1);
  is($s_expand_11->key, 'side=1 0,0,0 1,1,0 1,1,3 2,0,1');

  is($s_expand_00->{'name'}, '00');
  is($s_expand_01->{'name'}, '01');
  is($s_expand_10->{'name'}, '10');
  is($s_expand_11->{'name'}, '11');

  my $s_expand_101 = $s_expand_10->expand(1);
}

{
  my $s = MyDragonConf->initial (name => '',
                                 want_next_segs => 1);
  my $s_expand_0 = $s->expand(0);
  is($s_expand_0->key, 'side=0 0,0,0 1,1,3 next=1,1,3');

  my $s_expand_1 = $s->expand(1);
  is($s_expand_1->key, 'side=1 0,0,0 1,-1,1');

  #   4
  #   |
  #   v
  #   3 <---- 2                2 ---> 1
  #           |                  \  / ^\
  #           v                   vv  |/
  #   0 ----> 1                       0

  my $s_expand_00 = $s_expand_0->expand(0);
  is($s_expand_00->key, 'side=0 0,0,0 0,2,3 1,1,2 1,1,3 next=1,1,3');

  my $s_expand_01 = $s_expand_0->expand(1);
  is($s_expand_01->key, 'side=1 -1,-1,0 0,0,0 0,0,3 1,-1,1 next=0,0,3');

  my $s_expand_10 = $s_expand_1->expand(0);
  is($s_expand_10->key, 'side=0 0,0,0 0,0,1 1,-1,1 1,1,2 next=1,-1,1');

  my $s_expand_11 = $s_expand_1->expand(1);
  is($s_expand_11->key, 'side=1 0,0,0 1,1,0 1,1,3 2,0,1');

  is($s_expand_00->{'name'}, '00');
  is($s_expand_01->{'name'}, '01');
  is($s_expand_10->{'name'}, '10');
  is($s_expand_11->{'name'}, '11');

  my $s_expand_101 = $s_expand_10->expand(1);
}

# MyDragonConf->initial->print;
# MyDragonConf->initial->expand(0)->print;


sub n_to_conf_by_path {
  my ($k, $on, %options) = @_;
  my $want_next_segs = ($options{'want_next_segs'} ||= 0);
  ### n_to_conf_by_path(): "k=$k on=$on want_next_segs=$want_next_segs"

  my $path = DragonCurve_path();
  my ($ox,$oy) = $path->n_to_xy($on & 1 ? $on+1 : $on);
  my $odir = MyDragon::xy_to_dir4($path->n_to_dxdy($on));
  if ($on & 1) { $odir += 2; }
  (($ox+$oy)&1) == 0 or die "oops";
  ### origin: "$ox,$oy odir $odir"

  my $transform = sub {
    my ($x,$y, $dir) = @_;
    ### transform: "$x,$y dir $dir"
    if (($x+$y) & 1) {
      $x += $MyDragon::dir4_to_dx[$dir];
      $y += $MyDragon::dir4_to_dy[$dir];
      $dir += 2;
      ### from end: "$x,$y dir $dir"
    }
    $x -= $ox;
    $y -= $oy;
    ($x,$y) = MyDragon::xy_rotate_plus90($x,$y, -$odir);
    $dir = ($dir - $odir) % 4;
    ### shift and rotate: "$x,$y dir $dir"
    return "$x,$y,$dir";
  };

  my ($n_lo, $n_hi) = $path->level_to_n_range($k);
  my @segs;
  my @next_segs;
  foreach my $n ($n_lo .. $n_hi-1) {
    my ($x,$y) = $path->n_to_xy($n);
    my $dir = MyDragon::xy_to_dir4($path->n_to_dxdy($n));
    my $seg = $transform->($x,$y, $dir);
    ### $n
    ### $seg
    if ($n > $on && @next_segs < $want_next_segs) {
      push @next_segs,  $seg;
    }
    push @segs, $seg;
  }
  ### @next_segs

  return MyDragonConf->new
    (name      => "N=$on",
     side      => $on & 1,
     segs      => \@segs,
     next_segs => \@next_segs,
     %options);
}
{
  #   8   5---4
  #   |   ^   |
  #   |   |   |
  #   7<--6   3---2
  #               |
  #           0---1
# n=5
# conf: side=1 next=2,0,1 -1,-1,3 0,-2,1 0,-2,2 0,0,0 0,0,1 1,-1,1 1,-1,2 1,1,2
# path: side=1 next=0,0,1 -1,-1,3 0,-2,1 0,-2,2 0,0,0 0,0,1 1,-1,1 1,-1,2 1,1,2

  my $want_next_segs = 16;
  my @confs = (MyDragonConf->initial (want_next_segs => $want_next_segs));
  my $bad = 0;
  K: foreach my $k (0 .. 1) {
    foreach my $n (0 .. 2**$k-1) {
      my $conf_by_path = n_to_conf_by_path($k,$n,
                                           want_next_segs => $want_next_segs);
      my $str_by_path = $conf_by_path->key;
      my $str         = $confs[$n]->key;
      unless ($str_by_path eq $str) {
        print "wrong at k=$k n=$n\n";
        print "conf: $str\n";
        print "path: $str_by_path\n";
        last K if ++$bad > 3;
      }
    }
    @confs = map {$_->expand(0), $_->expand(1)} @confs;
  }
  is ($bad, 0);
}

is_deeply([$MyDragon::triple_7_9_13_segs->triple_15ths_list],
          [7,9]);
is_deeply([$MyDragon::triple_7_9_13_segs->triple_15ths_diffs],
          [2]);

{
  my @confs = (MyDragonConf->initial (want_next_segs => 1,
                                      within => $triple_15ths_and_next_within));
  @confs = map {$_->expand(0), $_->expand(1)} @confs;
  is_deeply ([$confs[0]->triple_15ths_list], [13]);
  is_deeply ([$confs[0]->triple_15ths_diffs], []);
  is_deeply ([$confs[0]->next_segs], ['1,1,3']);
  is_deeply ([$confs[0]->triple_15ths_diff_to_next], [15-13 + 15-9]);

  is ($confs[1]->{'side'}, 1);
  is_deeply ([$confs[1]->triple_15ths_list], [7,9]);  # not ordered
  is_deeply ([$confs[1]->triple_15ths_diffs], [2]);
}


#------------------------------------------------------------------------------
# 5ths Fifths

{
  is ($triple_5ths_nexts_segs->num_segs, 10);
  is ($triple_5ths_nexts_segs->{'side'}, 0);

  is ($triple_5ths_and_next_within->num_segs, 26);
  is ($triple_5ths_and_next_within->{'side'}, 0);

  foreach my $bit (0,1) {
    my $e = $triple_5ths_and_next_within->expand($bit);
    ok ($triple_5ths_and_next_within->is_subset($e));

    # $e->print;
    # $triple_5ths_and_next_within->difference($e)->print;
  }
}

is_deeply([$MyDragon::triple_1_3_11_segs->triple_5ths_list],
          [1]);
is_deeply([$MyDragon::triple_2_6_8_segs->triple_5ths_list],
          [2]);
is_deeply([$MyDragon::triple_7_9_13_segs->triple_5ths_list],
          [3]);
is_deeply([$MyDragon::triple_4_12_14_segs->triple_5ths_list],
          [3,4]);     # 4 contains 3 as a segment subset

{
  my @confs = (MyDragonConf->initial (want_next_segs => 3,
                                      within => $triple_5ths_and_next_within));
  @confs = map {$_->expand(0), $_->expand(1)} @confs;
  is_deeply ([$confs[0]->triple_5ths_list], []);
  is_deeply ([$confs[0]->triple_5ths_diffs], []);
  is_deeply ([$confs[0]->next_segs], ['1,1,3']);
  is_deeply ([$confs[0]->triple_5ths_diff_to_next], []);

  is ($confs[1]->{'side'}, 1);
  is_deeply ([$confs[1]->triple_5ths_list], [3]);
  is_deeply ([$confs[1]->triple_5ths_diffs], []);

  @confs = map {$_->expand(0), $_->expand(1)} @confs;
  @confs = map {$_->expand(0), $_->expand(1)} @confs;
  is (scalar(@confs), 8);

  ### x:  [map {[$_->triple_5ths_list]} @confs]
  is_deeply([map {[$_->triple_5ths_list]} @confs],
            [ [], [3], [3,4], [], [], [2,3,4], [3], [] ]);
  #           ->  <-    ->    <-  ->   <-      ->
  #           0    1     2     3   4    5
  is_deeply ([$confs[0]->triple_5ths_diffs], []);
  is_deeply ([$confs[0]->triple_5ths_diff_to_next], []);

  is_deeply ([$confs[1]->triple_5ths_diffs], []);
  is_deeply ([$confs[1]->triple_5ths_diff_to_next], [6]);

  is_deeply ([$confs[2]->triple_5ths_diffs], [1]);
  is_deeply ([$confs[2]->triple_5ths_diff_to_next], [1+5+5+1]);

  is_deeply ([$confs[3]->triple_5ths_diffs], []);
  is_deeply ([$confs[4]->triple_5ths_diffs], []);
  is_deeply ([$confs[5]->triple_5ths_diffs], [1,1]);
  is_deeply ([$confs[5]->triple_5ths_list], [2,3,4]);
  is_deeply ([$confs[5]->triple_5ths_diff_to_next], [2 + 3]);
}

MyDragon::triple_5ths_surround_init();


#------------------------------------------------------------------------------
# 15ths Fifteenths

{
  # conf triple 15ths union side=0
  #    0,0,0 0,0,1 0,0,3 0,0,2 1,-1,2 -1,-1,1 1,-1,3 1,-1,1 -1,-1,0 1,1,3 1,1,0
  #    *     *     *---->*
  #          ^     |
  #          |     |
  #          |     v
  #    *<----O====>*     *
  #    ^     |     ^
  #    |     |     |
  #    |     v     |
  #    *---->*<----*     *
  #                |
  #                |
  #                v
  #    *     *     *     *

  # conf 15ths union and extra side=0
  #    0,0,0 0,0,1 0,0,3 0,0,2 1,-1,2 -1,-1,1 1,-1,3 1,-1,1 -1,-1,0 1,1,3 1,1,0 2,0,2
  #    *     *     *---->*
  #          ^     |
  #          |     |
  #          |     v
  #    *<----O====>*<----*
  #    ^     |     ^
  #    |     |     |
  #    |     v     |
  #    *---->*<----*     *
  #                |
  #                |
  #                v
  #    *     *     *     *


  my $c = $triple_15ths_union_segs;
  is ($c->num_segs, 11);
  ok (  $c->contains('0,0,0'));
  ok (  $c->contains('1,-1,2'));

  # extra at "e"
  my @extra_segs = ('2,0,2');
  foreach my $extra_seg (@extra_segs) {
    ok (! $c->contains($extra_seg));
  }
  $c = $c->union (MyDragonConf->new(segs => \@extra_segs))
    ->set_name('15ths union and extra');
  is ($c->num_segs, 12);

  foreach my $bit (0,1) {
    my $e = $c->expand($bit);
    # $e->print;
    ok ($c->is_subset($e));
  }
}
{
  # $triple_15ths_and_next_segs  union at positions before and after
  is ($triple_15ths_and_next_segs->num_segs, 25);
  is ($triple_15ths_and_next_within->num_segs, 27);

  foreach my $bit (0,1) {
    my $e = $triple_15ths_and_next_within->expand($bit);
    ok ($triple_15ths_and_next_within->is_subset($e));

    # $e->print;
    # $c->difference($e)->print;
  }

  my $u = $triple_15ths_union_segs;
  my $r1 = $triple_15ths_union_segs->rotate(1);
  my $r3 = $triple_15ths_union_segs->rotate(-1);
  $u = $u->union($r1,
                 $r3,
                 $r1->translate(1,-1),
                 $r3->translate(1,1));
  ok ($u->equals($triple_15ths_and_next_segs));
}
{
  # @triple_15ths_segs

  my $count = 0;
  foreach my $i (0 .. $#triple_15ths_segs) {
    my $c = $triple_15ths_segs[$i];
    if (defined $c) {
      $count++;
      ok($c->contains('0,0,0'), "triple_15ths_segs i=$i");
    }
  }
  is ($count, 12);
}

MyDragon::triple_15ths_surround_init();


#------------------------------------------------------------------------------
# TAparent

{
  my @TAparentDir_samples = (2, 3, 2, 0, 3, 3, 2, 0, 1, 0, 3, 0, 3, 3, 2, 0, 1);
  foreach my $i (0 .. $#TAparentDir_samples) {
    my $n = $i + 1;
    is (TAparentDir($n), $TAparentDir_samples[$i], "TAparentDir");
  }
}

{
  my @TAtospineDir_samples = (0, 1, 1, 2, 1, 2, 3, 2, 1, 2, 3, 2, 0, 3, 3, 2, 1, 2);
  foreach my $n (0 .. $#TAtospineDir_samples) {
    is (TAtospineDir($n), $TAtospineDir_samples[$n], "TAtospineDir");
  }
}
{
  my @TAtospine_samples = (1,2,5,2,11,4,5,6,23,8,9,10,13,10,13,14,47,16);
  foreach my $n (0 .. $#TAtospine_samples) {
    is (TAtospine($n), $TAtospine_samples[$n]);
  }
}


#------------------------------------------------------------------------------
# NdragonToPlus()

{
  my @NdragonToPlus_samples = (0,1,3,5,7,6,11,13,15,14,12,10,23,22,27,29);
  foreach my $n (0 .. $#NdragonToPlus_samples) {
    is (NdragonToPlus($n), $NdragonToPlus_samples[$n], "NdragonToPlus n=$n");
  }
}


#------------------------------------------------------------------------------
# Paper()

{
  my @Paper_samples = (1,1,0,1,1,0,0,1,1,1,0,0,1,0,0);
  foreach my $i (0 .. $#Paper_samples) {
    my $n = $i+1;
    is (Paper($n), $Paper_samples[$i],
        'Paper() vs samples');
    is (Paper($n), TurnLpred($n),
        'Paper() is TurnLpred()');
  }
}


#------------------------------------------------------------------------------
# various measures

{
  my @R_samples = (1, 2, 4, 8, 16, 28, 48, 84, 144, 244, 416, 708, 1200, 2036);
  foreach my $k (0 .. $#R_samples) {
    is (R($k), $R_samples[$k], "R");
  }
}
{
  my @L_samples = (1, 2, 4, 8, 12, 20, 36, 60, 100, 172, 292, 492, 836, 1420);
  foreach my $k (0 .. $#L_samples) {
    is (L($k), $L_samples[$k], "L");
  }
}
{
  my @B_samples = (2, 4, 8, 16, 28, 48, 84, 144, 244, 416, 708, 1200);
  foreach my $k (0 .. $#B_samples) {
    is (B($k), $B_samples[$k], "B");
  }
}

{
  my @RQ_samples = (1, 2, 3, 5, 9, 15, 25, 43, 73, 123, 209, 355, 601, 1019, 1729);
  foreach my $k (0 .. $#RQ_samples) {
    is (RQ($k), $RQ_samples[$k], "RQ");
  }
}
{
  foreach my $k (0 .. 20) {
    is (BQ3o($k), JA($k));
  }
}
{
  my @LQ_samples = (1, 1, 2, 4, 6, 10, 18, 30, 50, 86, 146, 246, 418, 710);
  foreach my $k (0 .. $#LQ_samples) {
    is (LQ($k), $LQ_samples[$k], "LQ");
  }
}

{
  my @A_samples = (0, 0, 0, 0, 1, 4, 11, 28, 67, 152, 335, 724, 1539);
  foreach my $k (0 .. $#A_samples) {
    is (A($k), $A_samples[$k], "A");
  }
}
{
  my @AR_samples = (0, 0, 0, 0, 0, 1, 4, 11, 28, 67, 152, 335, 724);
  foreach my $k (0 .. $#AR_samples) {
    is (AR($k), $AR_samples[$k], "AR");
  }
}
{
  my @AL_samples = (0, 0, 0, 0, 1, 3, 7, 17, 39, 85, 183, 389, 815);
  foreach my $k (0 .. $#AL_samples) {
    is (AL($k), $AL_samples[$k], "AL");
  }
}

{
  #                 0  1  2  3  4
  my @JA_samples = (0, 0, 0, 1, 2, 3, 6, 11, 18, 31, 54, 91, 154, 263, 446, 755);
  foreach my $k (0 .. $#JA_samples) {
    is (JA($k), $JA_samples[$k], "JA");
  }
}
{
  #                 0  1  2  3  4
  my @JN_samples = (1, 2, 4, 7, 14, 28, 55, 110, 220, 439, 878);
  foreach my $k (0 .. $#JN_samples) {
    is (JN($k), $JN_samples[$k], "JN");
  }
}
{
  #                      0  1  2  3  4
  my @JNother_samples = (1, 2, 4, 11, 22, 44, 91, 182, 364, 731, 1462);
  foreach my $k (0 .. $#JNother_samples) {
    is (JNother($k), $JNother_samples[$k], "JN");
  }
}

{
  my @S_samples = (2, 3, 5, 9, 15, 25, 43, 73, 123, 209, 355, 601, 1019, 1729);
  foreach my $k (0 .. $#S_samples) {
    is (S($k), $S_samples[$k], "S");
  }
}
{
  my @P_samples = (2, 3, 5, 9, 16, 29, 54, 101, 190, 361, 690, 1325, 2558, 4961);
  foreach my $k (0 .. $#P_samples) {
    is (P($k), $P_samples[$k], "P");
  }
}
{
  my @Dpred_samples =
    ('', # 0
     '', # 1
     '', # 2
     '', # 3
     '7,11', # 4
     '7,11,14,15,19,21,22,25', # 5
     '7,11,14,15,19,21,22,25,28,30,31,35,38,39,42,43,44,45,49,50,53,57');
  foreach my $k (0 .. $#Dpred_samples) {
    my $got = join(',', grep {Dpred_k($k,$_)} 0 .. (1<<$k)-1);
    my $want = $Dpred_samples[$k];
    is($got, $want, "Dpred k=$k");
  }
}


#------------------------------------------------------------------------------
# centroid

{
  my ($x,$y) = GLQ_xy(0);
  is($x, .5);
  is($y, .5);
}
{
  my ($x,$y) = GLQ_xy(2);
  # GLQ_xy(2): "$x,$y"
  is ($x, (.5 + -.5)/2);
  is ($y, (.5 + 1.5)/2);
}

{
  my @sG_real_samples = (1, 1, 2, 5, 10, 19, 38, 77, 154, 307, 614);
  foreach my $k (0 .. $#sG_real_samples) {
    my $got = sG_real($k);
    my $want = $sG_real_samples[$k];
    is($got, $want, "sG_real k=$k");
  }
}
{
  my @sG_imag_samples = (0, 0, 1, 2, 3, 6, 13, 26, 51, 102, 205);
  foreach my $k (0 .. $#sG_imag_samples) {
    my $got = sG_imag($k);
    my $want = $sG_imag_samples[$k];
    is($got, $want, "sG_imag k=$k");
  }
}

#------------------------------------------------------------------------------
# convex hull

{ my ($x,$y) = P1_xy(0); is($x,1); is($y,0); }
{ my ($x,$y) = P2_xy(0); is($x,0); is($y,0); }
{ my ($x,$y) = P3_xy(0); is($x,0); is($y,0); }
{ my ($x,$y) = P6_xy(0); is($x,1); is($y,0); }
{ my ($x,$y) = P7_xy(0); is($x,1); is($y,0); }

{
  my @P1N_samples = (1, 1, 2, 4, 9, 17, 34, 68, 137, 273, 546, 1092);
  foreach my $k (0 .. $#P1N_samples) {
    my $got = P1N($k);
    my $want = $P1N_samples[$k];
    is($got, $want, "P1N k=$k");
  }
}
{
  my @P5N_samples = (1/2,0,0,0,1,1,2,4,9,17,34,68,137,273,546,1092,2185,4369,8738,17476);
  foreach my $k (0 .. $#P5N_samples) {
    my $got = P5N($k);
    my $want = $P5N_samples[$k];
    is($got, $want, "P5N k=$k");
  }
}

{
  my @HBV_samples = (2,3,5,7,9,11,16,20,24,28,38,46,54,62,82,98);
  foreach my $k (0 .. $#HBV_samples) {
    my $got = HBV($k);
    my $want = $HBV_samples[$k];
    is($got, $want, "HBV k=$k");
  }
}

{
  my $path = DragonCurve_path();
  foreach my $k (0 .. 20) {
    my $n = P1N($k);
    is ($n, int($n));
    my ($x,$y) = $path->n_to_xy($n);
    is_deeply([$x,$y], [P1_xy($k)]);
  }
}


#------------------------------------------------------------------------------
# blob measures

{
  my @BlobP_samples = (0, 0, 0, 0, 4, 9, 17, 34, 68, 133, 257, 498, 972);
  foreach my $k (0 .. $#BlobP_samples) {
    is (BlobP($k), $BlobP_samples[$k], "BlobP");
  }
}
{
  my @BlobS_samples = (0, 0, 0, 0, 3, 6, 11, 20, 35, 62, 107, 180, 307, 526);
  foreach my $k (0 .. $#BlobS_samples) {
    is (BlobS($k), $BlobS_samples[$k], "BlobS");
  }
}
{
  my @BlobPrecedingP_samples = (7,13,24,43,79,149,284,543,1043,2017);
  foreach my $i (0 .. $#BlobPrecedingP_samples) {
    my $k = $i+4;
    is (BlobPrecedingP($k), $BlobPrecedingP_samples[$i], "BlobPrecedingP");
  }
}

{
  my @BlobB_samples = (0, 0, 0, 0, 4, 10, 20, 38, 68, 122, 212);
  foreach my $k (0 .. $#BlobB_samples) {
    is (BlobB($k), $BlobB_samples[$k], "BlobB");
  }
}
{
  my @BlobA_samples = (0,0,0,0, 1, 3, 6, 14, 33, 71, 150);
  foreach my $k (0 .. $#BlobA_samples) {
    is (BlobA($k), $BlobA_samples[$k], "BlobA");
  }
}

{
  #                 4,  5,  6, ...
  my @BlobN_samples = (7,14,28,53,103,206,412,821,1639,3278,6556,13109);
  foreach my $i (0 .. $#BlobN_samples) {
    my $k = $i+4;
    is (BlobN($k), $BlobN_samples[$i], "BlobN");
  }
}
{
  my @BlobSegments_samples = (0, 0, 0, 0, 4, 11, 22, 47, 100, 203, 406);
  foreach my $k (0 .. $#BlobSegments_samples) {
    is (BlobSegments($k), $BlobSegments_samples[$k], "BlobSegments");
  }
}
{
  my @BlobDeltaLen_samples = (0,0,0,0,0,1,1,3,4,9,9,19,20,41,41,83);
  foreach my $k (0 .. $#BlobDeltaLen_samples) {
    is (BlobDeltaLen($k), $BlobDeltaLen_samples[$k], "BlobDeltaLen");
  }
}

{
  my @BridgeRother_samples = (23, 47, 94, 188, 375, 751, 1502);  # k>=4
  foreach my $i (0 .. $#BridgeRother_samples) {
    my $k = $i + 4;
    is (BridgeRother($k), $BridgeRother_samples[$i], "BridgeRother");
  }
}

{
  my @BlobStart_xy_samples = ([0,0],[0,0],[0,0],[0,0],
                              [-2, 1],
                              [-3, -1],
                              [-2, -4],
                              [1, -6],
                              [6, -5],
                              [11, 1],
                              [10, 12]);
  foreach my $k (4 .. $#BlobStart_xy_samples) {
    my ($got_x,$got_y) = BlobStart_xy($k);
    my ($want_x,$want_y) = @{$BlobStart_xy_samples[$k]};
    is($got_x, $want_x, "BlobStart_xy() x k=$k");
    is($got_y, $want_y, "BlobStart_xy() y k=$k");
  }
}
{
  my @BlobEnd_xy_samples = ([0,0],[0,0],[0,0],[0,0],
                            [-2, 1],
                            [-3, -2],
                            [-1, -5],
                            [4, -6],
                            [10, -1],
                            [11, 10],
                            [1, 21]);
  foreach my $k (4 .. $#BlobEnd_xy_samples) {
    my ($got_x,$got_y) = BlobEnd_xy($k);
    my ($want_x,$want_y) = @{$BlobEnd_xy_samples[$k]};
    is($got_x, $want_x, "BlobEnd_xy() x k=$k");
    is($got_y, $want_y, "BlobEnd_xy() y k=$k");
  }
}

{
  my @BlobPathLen_samples = (0,0,0,0,  0,1,2,3,8,9,18,19,40,41,82,83);
  foreach my $k (0 .. $#BlobPathLen_samples) {
    my $got = BlobPathLen($k);
    my $want = $BlobPathLen_samples[$k];
    is($got, $want, "BlobPathLen_samples k=$k");
  }
}

#------------------------------------------------------------------------------
# blob convex hull

{ my ($x,$y) = BP1_xy(11);     is($x,  0); is($y,42); }
{ my ($x,$y) = BP2_xy(11);     is($x,  5); is($y,37); }
{ my ($x,$y) = BP3_xy(11);     is($x,  5); is($y,32); }
{ my ($x,$y) = BP4_xy(11);     is($x, -1); is($y,21); }
{ my ($x,$y) = BP4dash_xy(11); is($x,  1); is($y,24); }
{ my ($x,$y) = BP5_xy(11);     is($x,-11); is($y,11); }
{ my ($x,$y) = BP6_xy(11);     is($x,-16); is($y,11); }
{ my ($x,$y) = BP7_xy(11);     is($x,-18); is($y,13); }
{ my ($x,$y) = BP8_xy(8);      is($x, 11); is($y,-1); }
{ my ($x,$y) = BP8dash_xy(8);  is($x,  8); is($y, 1); }
{ my ($x,$y) = BP9_xy(8);      is($x, 13); is($y,-3); }


#------------------------------------------------------------------------------
# complex base

{
  my @i_minus_1_power_x = (1, -1,  0, 2);
  my @i_minus_1_power_y = (0, 1,  -2, 2);
  foreach my $k (0 .. $#i_minus_1_power_x) {
    my($got_x,$got_y) = i_minus_1_power($k);
    my $want_x = $i_minus_1_power_x[$k];
    my $want_y = $i_minus_1_power_y[$k];
    is($got_x, $want_x, "i_minus_1_power() x k=$k");
    is($got_y, $want_y, "i_minus_1_power() y k=$k");
  }
}
{
  my @PlusOffsetV_x = (0,0,1,1,-1,-1,-5,-5,3,3,19,19,-13,-13,-77,-77,51,51,307,307);
  my @PlusOffsetV_y = (0,0,1,1,3,3,-1,-1,-9,-9,7,7,39,39,-25,-25,-153,-153,103,103);
  foreach my $k (0 .. $#PlusOffsetV_x) {
    my ($got_x,$got_y) = PlusOffsetV_xy($k);
    my $want_x = $PlusOffsetV_x[$k];
    my $want_y = $PlusOffsetV_y[$k];
    is($got_x, $want_x, "PlusOffsetV_xy() x k=$k");
    is($got_y, $want_y, "PlusOffsetV_xy() y k=$k");
  }
}
{
  my @PlusOffsetH_x = (0,1,1,1,1,-3,-3,-3,-3,13,13,13,13,-51,-51,-51,-51,205,205,205);
  my @PlusOffsetH_y = (0,0,0,2,2,2,2,-6,-6,-6,-6,26,26,26,26,-102,-102,-102,-102,410);
  foreach my $k (0 .. $#PlusOffsetH_x) {
    my($got_x,$got_y) = PlusOffsetH_xy($k);
    my $want_x = $PlusOffsetH_x[$k];
    my $want_y = $PlusOffsetH_y[$k];
    is($got_x, $want_x, "PlusOffsetH_xy() x k=$k");
    is($got_y, $want_y, "PlusOffsetH_xy() y k=$k");
  }
}

#------------------------------------------------------------------------------
# Graphs

{
  my @Diameter_samples = (1,2,4,8,12,18,26,36,52,70,102,136,200,266,394,524);
  foreach my $k (0 .. $#Diameter_samples) {
    is (Diameter($k), $Diameter_samples[$k], "Diameter k=$k");
  }
}

#------------------------------------------------------------------------------
exit 0;
