#!/usr/bin/perl -w

# Copyright 2016, 2019 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  See the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.

use strict;
use 5.006;
use FindBin;
use Math::Geometry::Planar;
use Test::More;

use lib $FindBin::Bin;
use MyPlanar;

use MyTestHelpers;
BEGIN { MyTestHelpers::nowarnings(); }

plan tests => 46;

# uncomment this to run the ### lines
# use Smart::Comments;


#------------------------------------------------------------------------------
# distance_segment_to_segment()

is(MyPlanar::distance_segment_to_segment([ [0,0],[1,0], [0,1],[1,1] ]),  1);
is(MyPlanar::distance_segment_to_segment([ [0,0],[1,0], [-5,1],[10,1] ]),  1);

# DistanceToSegment is +ve on the left and -ve on the right
is(Math::Geometry::Planar::DistanceToSegment([[-5,1],[10,1], [0,2]]),  1);
is(Math::Geometry::Planar::DistanceToSegment([[-5,1],[10,1], [0,0]]),  -1);
is(Math::Geometry::Planar::DistanceToSegment([[10,1],[-5,1], [0,2]]),  -1);
is(Math::Geometry::Planar::DistanceToSegment([[10,1],[-5,1], [0,0]]),  1);


#------------------------------------------------------------------------------
# has_overlap()
# has_boundary_touch()
# gpc_has_overlap()

{
  my $points1 = [ [0,0],[1,0],[0,1] ];
  my $points2 = [ [1,0],[2,0],[1,1] ];
  my $points3 = [ [-1,-1],[5,-1],[-1,5] ];
  my $points4 = [ [0,0],[-1,1],[-1,0] ];
  my $gpc1 = MyPlanar::points_to_gpc($points1);
  my $gpc2 = MyPlanar::points_to_gpc($points2);
  my $gpc3 = MyPlanar::points_to_gpc($points3);
  my $gpc4 = MyPlanar::points_to_gpc($points4);

  ok (  MyPlanar::gpc_has_overlap($gpc1,$gpc1));
  ok (! MyPlanar::gpc_has_overlap($gpc1,$gpc2));
  ok (  MyPlanar::gpc_has_overlap($gpc1,$gpc3));
  ok (! MyPlanar::gpc_has_overlap($gpc1,$gpc4));

  my $planar1 = Math::Geometry::Planar->new; $planar1->points($points1);
  my $planar2 = Math::Geometry::Planar->new; $planar2->points($points2);
  my $planar3 = Math::Geometry::Planar->new; $planar3->points($points3);
  my $planar4 = Math::Geometry::Planar->new; $planar4->points($points4);
  ok (  $planar1->MyPlanar::has_overlap($planar1));
  ok (! $planar1->MyPlanar::has_overlap($planar2));
  ok (  $planar1->MyPlanar::has_overlap($planar3));
  ok (! $planar1->MyPlanar::has_overlap($planar4));

  ok (  $planar1->MyPlanar::has_boundary_touch($planar1));
  ok (  $planar1->MyPlanar::has_boundary_touch($planar2));
  ok (! $planar1->MyPlanar::has_boundary_touch($planar3));
  ok (  $planar1->MyPlanar::has_boundary_touch($planar4));
}


#------------------------------------------------------------------------------
# segment_intersection_segment()

# SegmentIntersection() reckons touching ends as not intersecting.
ok(! Math::Geometry::Planar::SegmentIntersection([[0,0],[1,0], [1,0],[2,0]]));

is_deeply(MyPlanar::segment_intersection_segment([ [6,2],[4,2], [6,2],[7,2] ]),
          [ [6,2],[6,2] ],
          '6,2 point only');

is_deeply(MyPlanar::segment_intersection_segment([ [0,0],[3,0], [1,0],[2,0] ]),
          [ [1,0],[2,0] ]);
is_deeply(MyPlanar::segment_intersection_segment([ [2,0],[1,0], [0,0],[3,0] ]),
          [ [2,0],[1,0] ]);

is_deeply(MyPlanar::segment_intersection_segment([ [1,0],[3,0], [0,0],[2,0] ]),
          [ [1,0],[2,0] ]);
is_deeply(MyPlanar::segment_intersection_segment([ [0,0],[2,0], [1,0],[3,0] ]),
          [ [2,0],[1,0] ]);

is_deeply(MyPlanar::segment_intersection_segment([ [0,0],[1,0], [1,0],[3,0] ]),
          [ [1,0],[1,0] ],
          'single point in common');

is_deeply(MyPlanar::segment_intersection_segment([ [1,-3],[1,3], [-3,2],[3,2] ]),
          [ [1,2],[1,2] ],
          'crossing');

is_deeply(MyPlanar::segment_intersection_segment([ [1,1],[3,1], [1,1],[1,5] ]),
          [ [1,1],[1,1] ],
          'angled endpoint');


#------------------------------------------------------------------------------
# poly_boundary_intersection()

ok (  MyPlanar::segments_are_colinear([ [0,0],[1,0], [2,0],[3,0] ]));
ok (! Math::Geometry::Planar::Colinear([ [0,0],[1,0], [2,1] ]));
ok (  Math::Geometry::Planar::Colinear([ [0,0],[1,0], [3,0] ]));
ok (! MyPlanar::segments_are_colinear([ [0,0],[1,0], [2,1],[3,0] ]));

is_deeply(Math::Geometry::Planar::SegmentIntersection([[0,0],[2,2], [0,2],[2,0]]),
          [1,1]);
is_deeply(Math::Geometry::Planar::SegmentIntersection([[0,0],[1,0], [1,0],[2,0]]),
          0);
is_deeply(Math::Geometry::Planar::SegmentIntersection([[0,0],[1,0], [1,0],[1,1]]),
          0);

ok(MyPlanar::segment_contains_point([[0,0],[2,0], [0,0]]));
ok(MyPlanar::segment_contains_point([[0,0],[2,0], [1,0]]));
ok(MyPlanar::segment_contains_point([[0,0],[2,0], [2,0]]));
ok(! MyPlanar::segment_contains_point([[0,0],[2,0], [3,0]]));
ok(! MyPlanar::segment_contains_point([[2,0],[0,0], [-1,0]]));
ok(! MyPlanar::segment_contains_point([[0,0],[2,0], [1,1]]));


#------------------------------------------------------------------------------
# points_to_gpc()

{
  my $gpc = MyPlanar::points_to_gpc([ [0,0],[1,0],[0,0] ]);
  my @list = MyPlanar::gpc_to_points_list($gpc);
  is_deeply(\@list, [[ [0,0],[1,0],[0,0] ]]);
  ### @list
}
{
  my @points = MyPlanar::poly_union([ [0,0],[1,0],[0,0] ],
                                    [ [1,0],[2,0],[1,0] ]);
  is_deeply(\@points, []);
  ### @points
}


#------------------------------------------------------------------------------
# segments_are_colinear()

ok (  MyPlanar::segments_are_colinear([ [0,0],[1,0], [2,0],[3,0] ]));
ok (! Math::Geometry::Planar::Colinear([ [0,0],[1,0], [2,1] ]));
ok (  Math::Geometry::Planar::Colinear([ [0,0],[1,0], [3,0] ]));
ok (! MyPlanar::segments_are_colinear([ [0,0],[1,0], [2,1],[3,0] ]));

#------------------------------------------------------------------------------
exit 0;
