#!/usr/bin/perl -w

# Copyright 2016, 2017, 2019 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


use 5.004;
use strict;
use List::Util 'min';
use Math::PlanePath::DragonCurve 117; # v.117 for level_to_n_range()
use Math::NumSeq::PlanePathTurn;

use FindBin;
use lib $FindBin::Bin;
use MyDragon 'xy_rotate_plus90','xy_rotate_minus90',
  'Rpred','BlobN','BlobStart_xy','BlobEnd_xy', 'transform_to_horizontal';
use MyLineSegments;
use MyPlanar;
use MyPrintwrap ':all';

# uncomment this to run the ### lines
# use Smart::Comments;


my $level = 10;

print "  % generated by tools/$FindBin::Script\n";
print "  \\newcommand{\\MyLevel}{$level}\n";
my $path = Math::PlanePath::DragonCurve->new;

my ($n_lo, $n_hi) = $path->level_to_n_range($level);
my ($x_hi,$y_hi) = $path->n_to_xy($n_hi);
my ($x_join,$y_join) = $path->n_to_xy($n_hi/2);
### level: "$level  n_hi=$n_hi"

sub transform {
  my ($x,$y) = @_;
  return transform_to_horizontal($x,$y, $level);
}

sub to_frac {
  my ($n) = @_;
  return (int($n)==$n ? $n : (2*$n).'/2');
}
sub to_tfrac {
  my ($n) = @_;
  return (int($n)==$n ? $n : '\\tfrac{'.(2*$n).'}{2}');
}

my $blob_n_start = BlobN($level);
my $blob_n_end = BlobN($level+1)-3;

my $n_middle = $n_hi/2;
my $n_end_opposite = $n_hi - (BlobN($level) - 3);

print "  % blob_n_start=$blob_n_start  middle=$n_middle  n_end_opposite=$n_end_opposite blob_n_end=$blob_n_end\n";
$n_end_opposite > $n_middle or die;

sub nseg_to_blob_side {
  my ($n) = @_;
  if ($n >= $n_end_opposite) {
    $n = $n_hi-1 - $n;
  } elsif ($n >= $n_middle) {
    $n -= $n_middle;
  } else {
    # at n=$n_middle-1 want n=0
    $n = $n_middle-1 - $n;
  }

  if ($n < 7) { return 'bridge'; }
  if ($n < 11) { return 'right'; }  # blob 4  n=7,8,9,10

  for (my $k = 4; ; $k++) {
    my $bn = BlobN($k+1);
    next if $n >= $bn;
    if ($n >= $bn-3) { return 'bridge'; }

    my ($x1,$y1) = BlobStart_xy($k);
    my ($x2,$y2) = BlobEnd_xy($k);

    my $dx = $x2 - $x1;
    my $dy = $y2 - $y1;

    foreach (0, 1) {
      # FIXME: this an undocumented func
      my ($x3,$y3) = $path->n_to_xy($n);
      my $next_dx = $x3 - $x2;
      my $next_dy = $y3 - $y2;
      my $c = Math::NumSeq::PlanePathTurn::_turn_func_LSR
        ($dx,$dy, $next_dx,$next_dy,0);
      if ($c > 0) { return 'left'; }
      if ($c < 0) { return 'right'; }
      $n++; # look at segment end too
    }
    return 'crossing';
  }
}

{
  # curve segments

  my $segments = MyLineSegments->new;
  foreach my $n ($blob_n_start-3 .. $blob_n_end+3-1) {
    my $side = nseg_to_blob_side($n);
    next if $side eq 'crossing';
    next if $side eq 'left' && $n < $n_middle;
    next if $side eq 'left' && $n >= $n_end_opposite;
    next if $side eq 'right' && $n > $n_middle && $n < $n_end_opposite;

    my ($x,$y) = $path->n_to_xy($n);
    my ($x2,$y2) = $path->n_to_xy($n+1);
    $segments->add($x,$y, $x2,$y2);

    ($x,$y) = transform($x,$y);
    ($x2,$y2) = transform($x2,$y2);
  }

  $Printwrap = 2;
  printwrap("\\draw[my dragon] ");
  $Printwrap += 2;
  $segments->printwrap (transform => \&transform);
  printwrap(";\n");

  my ($x,$y) = $path->n_to_xy($blob_n_start-3);
  ($x,$y) = transform($x,$y);
  print "  \\coordinate (segments-start) at ($x,$y);\n";

  ($x,$y) = $path->n_to_xy($blob_n_end+3);
  ($x,$y) = transform($x,$y);
  print "  \\coordinate (segments-end) at ($x,$y);\n";
}

{
  # filled areas

  my $AunfOut = 0;

  my $draw = sub {
    my ($n_start, $n_end, $k, $side) = @_;   # inclusive
    ### $n_start
    ### $n_end
    $n_end > $n_start or die;

    my @points;
    my $segments = MyLineSegments->new;
    foreach my $n ($n_start .. $n_end-1) {
      next unless Rpred($n);
      my ($x1,$y1) = $path->n_to_xy($n);
      my ($x2,$y2) = $path->n_to_xy($n+1);
      $segments->add($x1,$y1, $x2,$y2);
      push @points, [$x1,$y1], [$x2,$y2];
    }
    my $area = MyPlanar::points_to_area(\@points);
    ### $area
    print "% GP-Test-last  $side($k)==",to_frac($area),"\n";
    $AunfOut += $area;

    # @points = map{"$_->[0]+I*$_->[1]"} @points;
    # print "  % ",join(',',@points),"\n";

    $Printwrap = 2;
    printwrap("\\fill[my area] ");
    $Printwrap += 2;
    $segments->printwrap (transform => \&transform);
    printwrap("--cycle;\n");
  };

  foreach my $k (reverse 5 .. $level-3) {   # left sides
    my $bn_start = BlobN($k);
    my $bn_end = BlobN($k+1)-3;
    ### $k
    ### $bn_start
    ### $bn_end
    $draw->($n_middle - $bn_end, $n_middle - $bn_start, $k, 'BlobAcl');
  }
  foreach my $k (4 .. $level-3) {           # right sides
    my $bn_start = BlobN($k);
    my $bn_end = BlobN($k+1)-3;
    $draw->($n_middle + $bn_start, $n_middle + $bn_end, $k, 'BlobAcr');
  }
  foreach my $k ($level-1) {
    my $bn_start = BlobN($k);
    my $bn_end = BlobN($k+1)-3;
    $draw->($n_hi - $bn_end, $n_hi - $bn_start, $k, 'BlobAcl');
  }

  print "% GP-Test-last  AunfOut($level)==",to_frac($AunfOut),"\n";
  print "  \\newcommand{\\MyAunfOut}{",to_tfrac($AunfOut),"}\n";
}


sub blob_crossing_line {
  my ($sx,$sy, $tx,$ty) = @_;
  ($sx,$sy) = transform($sx,$sy);
  ($tx,$ty) = transform($tx,$ty);
  my $len = ($sx-$tx)**2 + ($sy-$ty)**2;
  my $style = ($len < 1 ? "my crossing short" : "my crossing");
  printwrap(" \\draw[$style] ($sx,$sy)--($tx,$ty);");
}

{
  # crossing lines

  $Printwrap = 2;

  # end blobs, back from join
  foreach my $k (4 .. $level-3) {
    my ($sx,$sy) = BlobStart_xy($k);
    my ($tx,$ty) = BlobEnd_xy($k);
    ($sx,$sy) = xy_rotate_minus90($sx,$sy);
    ($tx,$ty) = xy_rotate_minus90($tx,$ty);
    $sx += $x_join;
    $sy += $y_join;
    $tx += $x_join;
    $ty += $y_join;
    blob_crossing_line ($sx,$sy, $tx,$ty);

    # same for second sub-curve
    ($sx,$sy) = xy_rotate_minus90($sx,$sy);
    ($tx,$ty) = xy_rotate_minus90($tx,$ty);
    $sx += $x_hi; $sy += $y_hi;
    $tx += $x_hi; $ty += $y_hi;
    blob_crossing_line ($sx,$sy, $tx,$ty);
  }

  # start blob back from curve end
  foreach my $k ($level-1) {
    my ($sx,$sy) = BlobStart_xy($k);
    my ($tx,$ty) = BlobEnd_xy($k);

    # same for second sub-curve
    ($sx,$sy) = xy_rotate_minus90($sx,$sy);
    ($tx,$ty) = xy_rotate_minus90($tx,$ty);
    $sx += $x_hi; $sy += $y_hi;
    $tx += $x_hi; $ty += $y_hi;
    blob_crossing_line ($sx,$sy, $tx,$ty);
  }

  printwrap("\n");
}



{
  my ($x,$y) = transform($x_join,$y_join);
  print "  \\coordinate (join) at ($x,$y);\n";
}
{
  my ($x,$y) = $path->n_to_xy($blob_n_start);
  ($x,$y) = transform($x,$y);
  print "  \\coordinate (blob-start) at ($x,$y);\n";
}
{
  my ($x,$y) = $path->n_to_xy($blob_n_end);
  ($x,$y) = transform($x,$y);
  print "  \\coordinate (blob-end) at ($x,$y);\n";
}

print "  % end generated\n";
exit 0;
