#!/usr/bin/perl -w

# Copyright 2014, 2015, 2016, 2017, 2018, 2019 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.010;
use strict;
use warnings;
use Math::PlanePath::DragonCurve 117; # v.117 for level_to_n_range()

use FindBin;
use lib $FindBin::Bin;
use MyDragon 'BlobN','BlobSegments',
  'mul','div','xy_rotate_plus90','xy_rotate_minus90','transform_to_horizontal';
use MyLineSegments;
use MyPrintwrap ':all';

# uncomment this to run the ### lines
# use Smart::Comments;


my $level = 16;

print "  % generated by tools/$FindBin::Script\n";

my $path = Math::PlanePath::DragonCurve->new;
my $scale = 2**($level/2 + 1);
print "  \\begin{scope}[scale=1/$scale]\n";

my $n_lo = BlobN($level);
my $n_hi = BlobN($level+1) - 3;

sub transform {
  my ($x,$y) = @_;
  ($x,$y) = transform_to_horizontal($x,$y, $level);
  $x *= 2;
  $y *= 2;
  return ($x,$y);
}

{
  my ($x,$y) = $path->n_to_xy($n_lo);
  ($x,$y) = transform($x,$y);
  print "  % \\coordinate (start) at ($x,$y);\n";
}
{
  my ($x,$y) = $path->n_to_xy($n_hi);
  ($x,$y) = transform($x,$y);
  print "  % \\coordinate (end) at ($x,$y);\n";
}

my $b4 = BlobSegments($level-4);
my $b2 = BlobSegments($level-2);

my @type;

foreach my $n ($n_lo .. $n_hi) {    # blob k-2 segments
  $type[$n] = 'middle-blob';
}

my $b4_first_lo = $n_lo;
my $b4_first_hi = $b4_first_lo + $b4;
foreach my $n ($b4_first_lo .. $b4_first_hi-1) {   # blob k-4 first segments
  $type[$n] = 'end-blob';
}

my $b2_lo = $b4_first_hi + 3;
foreach my $n ($b4_first_hi .. $b2_lo-1) {   # bridge segments
  $type[$n] = 'bridge';
}

my $b4_last_hi = $n_hi;
my $b4_last_lo = $b4_last_hi - $b4;
foreach my $n ($b4_last_lo .. $b4_last_hi-1) {
  $type[$n] = 'end-blob';
}
foreach my $n ($b4_last_lo-3 .. $b4_last_lo-1) {   # bridge segments
  $type[$n] = 'bridge';
}

{
  my ($x,$y) = $path->n_to_xy($b4_first_hi);
  ($x,$y) = transform($x,$y);
  print "  % \\coordinate (blob-4a-end) at ($x,$y);\n";
}
{
  my ($x,$y) = $path->n_to_xy($b2_lo);
  ($x,$y) = transform($x,$y);
  print "  % \\coordinate (blob-2-end) at ($x,$y);\n";
}
{
  my $b2_hi = $b4_last_lo - 3;
  my ($x,$y) = $path->n_to_xy($b2_hi);
  ($x,$y) = transform($x,$y);
  print "  % \\coordinate (blob-2-start) at ($x,$y);\n";
}
{
  my ($x,$y) = $path->n_to_xy($b4_last_lo);
  ($x,$y) = transform($x,$y);
  print "  % \\coordinate (blob-4b-start) at ($x,$y);\n";
}

# BlobPartsTC(8) == 8 - 4*I
# dragon_unpoint(8 - 4*I, -1) == [0, 112]
# dragon_unpoint(8 - 4*I, -1) == [0, 176]
# 
my $twin_lo = 112 * 2**($level-8);
my $twin_hi = $twin_lo + 4*2**($level-4);
foreach my $n ($twin_lo .. $twin_hi-1) {    # twindragon
  $type[$n] = 'twindragon';
}
foreach my $i (0 .. 3) {
  my ($x,$y) = $path->n_to_xy($twin_lo + $i*2**($level-4));
  ($x,$y) = transform($x,$y);
  print "  % \\coordinate (twin-corner-$i) at ($x,$y);\n";
}
{
  my ($x1,$y1) = $path->n_to_xy($twin_lo);
  my ($x2,$y2) = $path->n_to_xy($twin_lo + 4*2**($level-8));
  my $x = ($x1+$x2)/2;
  my $y = ($y1+$y2)/2;
  ($x,$y) = transform($x,$y);
  print "  % \\coordinate (twin-middle) at ($x,$y);\n";
}

print "  % lo=$n_lo,  twin=$twin_lo to $twin_hi,   hi=$n_hi\n";
print "  % segments ",BlobSegments($level),"\n";

my %xyxy_to_type;
sub xyxy_to_type {
  my ($x1,$y1, $x2,$y2) = @_;
  my $key = "$x1,$y1,$x2,$y2";
  return ($xyxy_to_type{$key}
          //= do {
            my $n = $path->xyxy_to_n_either($x1,$y1, $x2,$y2);
            (! defined $n || $n < $n_lo || $n >= $n_hi ? ''
             : $type[$n]);
          });
}
sub xyxy_is_different {
  my ($type, $x1,$y1, $x2,$y2) = @_;
  $type ne '' or die;
  my $other_type = xyxy_to_type($x1,$y1, $x2,$y2);
  ### xyxy_is_different(): "$type   $x1,$y1, $x2,$y2 = $other_type"
  return ($other_type ne $type);
}

my $segments = MyLineSegments->new;
my @prev_mx = (0,0);
my @prev_my = (0,0);
foreach my $n ($n_lo .. $n_hi-1) {
  ### $n
  my ($x1,$y1) = $path->n_to_xy($n);
  my ($x2,$y2) = $path->n_to_xy($n+1);
  my $type = $type[$n];

  my $dx = $x2 - $x1;
  my $dy = $y2 - $y1;
  my ($rx,$ry) = xy_rotate_minus90($dx,$dy);
  foreach my $side (0, 1) {

    if (xyxy_is_different($type,    $x1,$y1, $x1+$rx,$y1+$ry)
        || xyxy_is_different($type, $x2,$y2, $x2+$rx,$y2+$ry)
        || xyxy_is_different($type, $x1+$rx,$y1+$ry, $x2+$rx,$y2+$ry)) {
      my $mx = 2*$x1 + $dx + $rx;
      my $my = 2*$y1 + $dy + $ry;
      ($mx,$my) = mul($mx,$my);
      ### square on: "side=$side  $mx,$my  prev $prev_mx[$side],$prev_my[$side]"
      if ($prev_mx[$side]) {
        $segments->add($prev_mx[$side],$prev_my[$side], $mx,$my);
      }
      $prev_mx[$side] = $mx;
      $prev_my[$side] = $my;
    } else {
      $prev_mx[$side] = 0;
      $prev_my[$side] = 0;
    }
    $rx = -$rx;
    $ry = -$ry;
  }
}


$Printwrap = 2;
printwrap("\\draw[my grey] ");
$Printwrap += 2;
$segments->printwrap (transform => sub {
                        my ($x,$y) = @_;
                        ($x,$y) = div($x,$y);
                        $x /= 4;
                        $y /= 4;
                        ($x,$y) = transform($x,$y);
                        return ($x,$y);
                      });
printwrap(";");
printwrap(" % drawn with " . ($segments->lines_joined_count_points) . " points\n");

print "  \\end{scope}\n";
print "  % end generated\n";
exit 0;



__END__


  # if ($type eq '') {  # bridge
  #   $segments->add($x,$y, $x2,$y2);
  #   next;
  # }
  # if (xyxy_is_boundary($type, $x,$y, $x2,$y2)) {
  #   $segments->add($x,$y, $x2,$y2);
  #   next;
  # }
  # $count_undrawn++;

# $Printwrap = 2;
# printwrap("\\draw[my grey] ");
# $Printwrap += 2;
# $segments->printwrap (transform => \&transform);
# printwrap(";");
# printwrap(" % drawn with " . ($segments->lines_joined_count_points) . " points\n");


# my $count_undrawn = 0;

# my %squares;
# {
#   $Printwrap = 2;
#   printwrap("\\fill foreach \\p in {");
#   $Printwrap += 2;
#   foreach my $str (keys %squares) {
#     my ($mx,$my) = split /,/, $str;
#     ($mx,$my) = transform($mx,$my);
#     $mx = int($mx/2);
#     $my = int($my/2);
#     printwrap("($mx,$my),");
#   }
#   printwrap("}\n");
#   printwrap("{ \\p rectangle +(1,1) };");
#   my $count = scalar keys %squares;
#   printwrap(" % count $count\n");
# }
