#!/usr/bin/perl -w

# Copyright 2014, 2015, 2016, 2017, 2019 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.004;
use strict;
use List::Util 'min','max';
use Math::PlanePath::DragonCurve 118; # v.118 for xyxy_to_n_either()

use FindBin;
use lib $FindBin::Bin;
use MyDragon 'xy_rotate_minus90','transform_to_horizontal';
use MyPrintwrap ':all';


my $level = 8;
if (@ARGV) { $level = shift @ARGV; }

($level & 1) == 0 or die "level must be even";

print "  % generated by tools/$FindBin::Script\n";
print "  \\newcommand{\\MyLevel}{$level}\n";

sub weight_to_colour {
  my ($t) = @_;
  if ($t == 1200) {
    return "";
  }

  # range 10 to 100 parabolic
  {
    $t /= 1200;  # ranging 0 to 1 inclusive
    $t = ($t+.5)**2;  # .5**2 = .25 to 1.5**2 = 2.25
    $t = ($t - .25) / 2;  # 0 to 1
    my $percent = int(10 + 90*$t);
    return "[black!$percent]";
  }

  # range 10 to 100 linear
  {
    $t /= 1200;  # ranging 0 to 1 inclusive
    my $percent = int(10 + 90*$t);
    return "[black!$percent]";
  }
}

my $path = Math::PlanePath::DragonCurve->new;
my ($n_lo, $n_hi) = $path->level_to_n_range($level);

my $wx_min = 0;
my $wy_min = 0;
my $wx_max = 0;
my $wy_max = 0;
my %weight;

sub add {
  my ($wx,$wy, $weight) = @_;
  $weight{$wx}{$wy} += $weight;
  $wx_min = min($wx_min, $wx);
  $wy_min = min($wy_min, $wy);
  $wx_max = max($wx_max, $wx);
  $wy_max = max($wy_max, $wy);
}

sub transform {
  my ($x,$y) = @_;
  ($x,$y) = transform_to_horizontal($x,$y, $level);
  return ($x,$y);
}

foreach my $n ($n_lo .. $n_hi-1) {
  my ($x,$y) = $path->n_to_xy($n);
  my ($dx,$dy) = $path->n_to_dxdy($n);
  if ($n & 1) {
    ($x,$y) = ($x+$dx,$y+$dy);
    ($dx,$dy) = (-$dx,-$dy);
  }
  my ($rx,$ry) = xy_rotate_minus90($dx,$dy);

  add (2*$x -   $dx + $rx, 2*$y -   $dy + $ry,  96);
  add (2*$x +   $dx + $rx, 2*$y +   $dy + $ry,  348);
  add (2*$x + 3*$dx + $rx, 2*$y + 3*$dy + $ry,  6);

  add (2*$x -   $dx - $rx, 2*$y -   $dy - $ry,  24);
  add (2*$x +   $dx - $rx, 2*$y +   $dy - $ry,  102);
  add (2*$x + 3*$dx - $rx, 2*$y + 3*$dy - $ry,  24);
}

my @histogram;
$Printwrap = 2;
my $count_squares = 0;
foreach my $wy ($wy_min .. $wy_max) {
  foreach my $wx ($wx_min .. $wx_max) {
    my $weight = $weight{$wx}{$wy} || next;
    $histogram[$weight]++;
  }
}
my $num_distinct_weights = 0;
foreach my $weight (0 .. $#histogram) {
  if ($histogram[$weight]) {
    $num_distinct_weights++;
  }
}

my $y_lo = 0;
my $y_hi = 0;
foreach my $weight (0 .. $#histogram) {
  next unless $histogram[$weight];
  my $colour = weight_to_colour($weight);
  $Printwrap = 2;
  printwrap("\\fill$colour foreach \\p in {");
  $Printwrap += 2;
  foreach my $wy ($wy_min .. $wy_max) {
    foreach my $wx ($wx_min .. $wx_max) {
      my $wxwy_weight = $weight{$wx}{$wy} || next;
      next unless $wxwy_weight == $weight;

      my ($x1,$y1) = ($wx-1,$wy-1);
      my ($x2,$y2) = ($wx+1,$wy+1);
      ($x1,$y1) = transform($x1,$y1);
      ($x2,$y2) = transform($x2,$y2);
      my $mx = min($x1,$x2);
      my $my = min($y1,$y2);
      $mx /= 2;
      $my /= 2;
      printwrap("($mx,$my),");
      $y_lo = min($y_lo,$my);
      $y_hi = max($y_hi,$my+1);
      $count_squares++;
    }
  }
  printwrap("}");
  printwrap(" { \\p rectangle +(1,1) };\n");
}

my $count_adjacent = 0;
foreach my $wy ($wy_min .. $wy_max) {
  foreach my $wx ($wx_min .. $wx_max) {
    my $weight = $weight{$wx}{$wy} || next;
    $count_adjacent += ($weight == ($weight{$wx+2}{$wy}||0));
    $count_adjacent += ($weight == ($weight{$wx}{$wy+2}||0));
  }
}

{
  my $y_mid = ($y_hi + $y_lo)/2;
  print "  \\begin{scope}[shift=(pos-scale),shift={(0,$y_mid)},scale=\\MyScale,font=\\tiny]\n";
  my $s_lo = - $num_distinct_weights/2;
  my $mx = 0;
  my $my = $s_lo;
  foreach my $weight (0 .. $#histogram) {
    next unless $histogram[$weight];
    my $colour = weight_to_colour($weight);
    printwrap(" \\fill$colour ($mx,$my) rectangle +(1,1);");
    printwrap(" \\node at (".($mx+1).",".($my+.5).")[right]{\$$weight\$};");

    $my++;
  }
  printwrap_newline();
  print "  \\end{scope}\n";
  my $s_hi = $my;
  $s_hi == -$s_lo or die "oops $s_lo to $s_hi";
}

print "  % $count_squares squares, $num_distinct_weights distinct colours ($count_adjacent adjacent)\n";

{
  print "  % weights: ";
  my $join = '';
  foreach my $weight (0 .. $#histogram) {
    next unless $histogram[$weight];
    my $t = $weight / 1200 * 100;
    printf "%s%.1f", $join, $t;
    $join = ', ';
  }
  print "\n";
}

print "  % end generated\n";
exit 0;
