# Copyright 2014, 2015, 2016, 2017, 2018, 2019, 2020, 2021 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  See the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.


# This is various functions related to dragon curve calculations, used for
# development and testing.  Everything likely to change, or move, etc.

# Most functions are named like in dragon.tex.
# Complex number returns are functions Foo_xy returning $x,$y.

package MyDragon;
use 5.010;
use strict;
use warnings;
use Carp 'croak';
use List::Util 'min','max','sum';
use POSIX 'ceil';
use Memoize;
use Math::PlanePath::Base::Digits
  'bit_split_lowtohigh','digit_join_lowtohigh';

# uncomment this to run the ### lines
# use Smart::Comments;

use base 'Exporter';
our @EXPORT_OK = ('DragonCurve_path',
                  'bitlength','FLAT_any_binary','FLAT_zeros','FLAT_ones',

                  'LQ','RQ','B','L','R','RJ',
                  'JA','JN','JNother','A','AR','AL',
                  'LQ3o','RQ3o','BQ3o',
                  'S','D','P',
                  'TA',
                  'BlobN','BlobNend', 'BlobSegments', 'BridgeCount',
                  'BlobP','BlobS','BlobB','BlobA','BlobPrecedingP',
                  'BlobStart_xy','BlobEnd_xy','BlobDelta_xy','BridgeDelta_xy',
                  'BlobDeltaLen', 'BlobPathLen', 'Diameter',
                  'BridgeLother','BridgeRother',
                  'MidSN','MidEN',

                  'turn','Mturn','Paper',
                  'TurnLpred','TurnRpred',
                  'TurnLpred_flat','TurnRpred_flat',

                  'dNextL',
                  'dNextL_flat', '@dNextL_state_table',
                  '@dNextL_state_to_name', '@dNextL_state_to_d',
                  'dNextL_LtoH_flat', '@dNextL_LtoH_state_table',
                  '@dNextL_LtoH_state_to_name', '@dNextL_LtoH_state_to_d',
                  'dNextLsub1_flat', '@dNextLsub1_state_table',
                  '@dNextLsub1_state_to_name', '@dNextLsub1_state_to_d',
                  'dNextLsub1_M_flat', '@dNextLsub1_M_state_table',
                  '@dNextLsub1_M_state_to_name', '@dNextLsub1_M_state_to_d',

                  'dNextR',
                  'dNextR_flat', '@dNextR_state_table','$dNextR_starting',
                  '@dNextR_state_to_name', '@dNextR_state_to_d',

                  'dTurnLeft_flat',
                  '@dTurnLeft_state_table','@dTurnLeft_state_to_name',
                  '@dTurnLeft_state_to_d','@dTurnLeft_state_to_decr',
                  'transmute_initial_0s_to_1s',

                  'dTurnRight_flat',
                  '@dTurnRight_state_table','@dTurnRight_state_to_name',
                  '@dTurnRight_state_to_d','@dTurnRight_state_to_incr',
                  'transmute_lowest_1_to_0',

                  'dir','dir4_flat','@dir4_state_table',
                  'CountOneBits','Gray','DirMaxN','DirRatio',
                  'TurnsL','TurnsR',
                  'BitAboveLowestZero',
                  'BitAboveLowestOne','BitAboveLowestOne_allowing_zero',
                  'BitAboveLowestOne_HtoL_flat',
                  '@BitAboveLowestOne_HtoL_state_table',
                  '@BitAboveLowestOne_HtoL_state_to_result',
                  'BitAboveLowestOne_LtoH_flat',
                  '@BitAboveLowestOne_LtoH_state_table',
                  '@BitAboveLowestOne_LtoH_state_to_result',
                  '@BitAboveLowestOne_LtoH_state_name',

                  'GS_xy','GS_complex','sG_real','sG_imag','GLQ_xy',
                  'to_bigrat','to_bigint',
                  'hull_expand','hull_points_sort_by_angle','HBV',
                  '$hull6_points','seg_to_hull6_points',
                  'blob_hull_points_sort',

                  'points_to_minrectarea',
                  'points_to_centroid','points_to_hull','points_to_bigrat',
                  'point_str','points_str','points_str_gp','point_str_gp',
                  'xy_to_node_name',

                  'mul','bk','div','unfold',
                  'xy_rotate_plus90','xy_rotate_minus90','xy_rotate_180',
                  'xy_to_dir4','xy_to_dir8',
                  'rect_corners_sort',
                  'is_integer','is_pow2',

                  'i_minus_1_mul','i_minus_1_div','i_minus_1_power',
                  'PlusOffsetH_xy','PlusOffsetV_xy','PlusOffsetJ_xy',
                  'PlusToDragon',
                  'NdragonToPlus',

                  'xy_to_tex_str','xy_to_gp_str','chamfer',
                  'transform_to_horizontal',
                  'dragon_Graph','blob_Graph',
                  'twindragon_Graph','twindragon_GraphViz2',
                  'twindragon_turn_tree_Graph',
                  'Graph_xy_to_vertex','Graph_vertex_to_xy','Graph_view',
                  'Graph_layout_diameterline',

                  'Graph_Easy_show',
                  'dir4_to_dx','dir4_to_dy','@dir4_to_dx','@dir4_to_dy',
                  '@dir8_to_dx','@dir8_to_dy',

                  'P1_xy', 'P2_xy', 'P3_xy', 'P4_xy', 'P5_xy', 'P6_xy',
                  'P7_xy', 'P8_xy', 'P9_xy', 'P10_xy','p_xy',
                  'P1N','P5N',

                  'BP1_xy', 'BP2_xy', 'BP3_xy', 'BP4dash_xy', 'BP4_xy',
                  'BP5_xy', 'BP6_xy', 'BP7_xy', 'BP8dash_xy', 'BP8_xy',
                  'BP9_xy', 'BP10_xy',

                  'TAdiameter',
                  'TAVertexToUpper','TAVertexToLower',
                  'TAVertexToLeft','TAVertexToRight',
                  'TAdegreeInf',
                  'WidthS_infinity_k',
                  'TAparentDir0_flat', 'TAparentDir1_flat',
                  'TAparentDir2_flat', 'TAparentDir3_flat',
                  'TAparentDir','TAparent',
                  'TAtospineDir', 'TAtospine',
                  'TA_complex_c',

                  'Rpred','Rpred_of_bits',
                  '@Rpred_state_table','@Rpred_state_to_name',
                  '%Rpred_name_to_state','@Rpred_state_to_sides',
                  'Rpred_flat','NonRpred_flat',
                  'NonRpred_flatregex','$NonRpred_flatregex_str',
                  '$NonRpred_re',
                  'Rnext',
                  '@Rpred_LtoH_state_table',
                  'Rpred_LtoH_flat','NonRpred_LtoH_flat',
                  'Rsides_flat','Rsides',

                  '$Lpred_state_aref','$Lpred_LtoH_state_aref',
                  'Lpred_k_flat','NonLpred_k_flat','Lpred_k_of_bits',
                  'LpredInf_flat','NonLpredInf_flat',
                  'Lpred_k_LtoH_flat','NonLpred_k_LtoH_flat',
                  'Lpred_k','Lpred_k_of_bits','Lnext_k',

                  'Bpred_k_flat','BpredInf_flat','NonBpred_k_flat',
                  'NonBlobPredInf_flat',
                  'fFifthsPred_flat',

                  'Spred_k',
                  'Dpred_k','Dpred_inf_flat',
                  '$DpredRight_runs_fre','$DpredLeft_runs_fre',
                  '$Dpred_runs_flat_fre',
                  'DpredRight_runs_flat','DpredLeft_runs_flat',
                  'SpredRight_runs_flat','SpredLeft_runs_flat',
                  '$DpredRight_k_flat_fre','$DpredLeft_k_flat_fre',
                  '$Dpred_k_flat_fre',
                  'DpredRight_k_flat','DpredLeft_k_flat','Dpred_k_flat',
                  'SpredRight_k_flat','SpredLeft_k_flat','Spred_k_flat',

                  '$DpredFirstR_fre','$DpredFirstL_fre',
                  'DpredFirstR_inf_flat','DpredFirstL_inf_flat',
                  'DpredFirst_inf_flat',
                  '$DpredSecondR_fre','$DpredSecondL_fre',
                  'DpredSecondR_flat','DpredSecondL_flat','DpredSecond_flat',
                  'DpredSecond',

                  'Epred','EpredR','EpredL',
                  'EpredR_flat','EpredR_LtoH_flat',
                  'EpredL_flat','EpredL_LtoH_flat',
                  'Epred_flat','Epred_LtoH_flat',

                  '$conf_before_and_after',
                  '@triple_7ths_segs',
                  '$triple_7ths_union_segs',
                  '$triple_7ths_and_next_segs','$triple_7ths_and_next_within',
                  '@triple_7ths_state_table','@triple_7ths_state_to_conf',

                  '$triple_15ths_union_segs',
                  '$triple_15ths_and_next_segs','$triple_15ths_and_next_within',
                  '@triple_15ths_segs',
                  'triple_15ths_surround_init',
                  '@triple_15ths_state_table','@triple_15ths_state_to_conf',

                  '$triple_5ths_union_segs',
                  '$triple_5ths_nexts_segs',
                  '$triple_5ths_and_next_segs','$triple_5ths_and_next_within',
                  'triple_5ths_surround_init',
                  '@triple_5ths_state_table','@triple_5ths_state_to_conf',

                  '$R_square_segs','$L_square_segs','$B_squares_segs',

                  'fRpred_flat','fNonRpred_flat',
                  'fLpred_flat','fNonLpred_flat',
                  'fBpred_flat','fNonBpred_flat',

                  '$hullstouch_segs',

                  '$hullstouch_segs_outer','$hullstouch_segs_inner',
                  '$hullstouch_segs_all',
                  'fBpred_hullstouch_init',
                  '@fBpred_hullstouch_state_table','@fBpred_hullstouch_state_enclosed_lists',
                  '@fBpred_hullstouch_state_to_conf',
                  '$fNonRpred_hullstouch_flat','$fNonLpred_hullstouch_flat',
                  '$fNonBpred_hullstouch_flat',
                  '$fRpred_hullstouch_flat','$fLpred_hullstouch_flat',
                  '$fBpred_hullstouch_flat',

                  '$surround_segs_outer','$surround_segs_inner',
                  '$surround_segs_all',
                  'fBpred_surround_init',
                  '@fBpred_surround_state_table','@fBpred_surround_state_enclosed_lists',
                  '@fBpred_surround_state_to_conf',
                  '$fNonRpred_surround_flat','$fNonLpred_surround_flat',
                  '$fNonBpred_surround_flat',
                  '$fRpred_surround_flat','$fLpred_surround_flat',
                  '$fBpred_surround_flat',

                  'fBpred_cases_extra',
                  'fRpred_cases_extra_list','fLpred_cases_extra_list',

                  '$hullsaround_segs_outer','$hullsaround_segs_inner',
                  '$hullsaround_segs_all',

                  '@fVisits_adjacent_list','%fVisits_seg_to_index',
                  'fVisits_state_num','fVisits_expand_bit',
                  'fVisits_adjacent_NFA','fVisits_adjacent_DFA',

                  '@CentrelinePred_state_table','@CentrelinePred_state_to_name',
                  'CentrelinePred_flat','CentrelinePred_by_path',
                  'OXpred_flat',

                  'MinusIncPred_dir8_LtoH_flat',
                  'MinusNegPred_LtoH_flat',
                  'MinusNegIncPred_dir8_LtoH_flat',
                  'MinusNegNeighbourPred_dir8_LtoH_flat',

                  '@MinusRot_state_table',
                  '@MinusRot_state_to_dz',
                  'MinusRotPred_LtoH_flat',
                  'MinusRotIncPred_dir8_LtoH_flat',
                  '@MinusRotIncPred_LtoH_dir8_to_state',
                  'MinusRotNeighbourPred_dir8_LtoH_flat',

                  'MinusUnrotPred_LtoH_flat',

                  'MinusDoublePred_LtoH_flat',
                  'MinusDoubleIncPred_dir8_LtoH_flat',
                  'MinusDoubleNeighbourPred_dir8_LtoH_flat',

                  'Pred','NeighbourPred','BoundaryPred',
                  'HangingSquarePred','HangingSquarePred_dir8','NonNSEWpred',

                  'ffixed_u','ffixed_u_parent','ffixed_uDepth',
                  'ffixed_du','ffixed_du_parent', 'ffixed_row_w',
                  'ffixed_string',
                 );
our %EXPORT_TAGS = (all => \@EXPORT_OK);

#------------------------------------------------------------------------------

use constant::defer DragonCurve_path => sub {
  require Math::PlanePath::DragonCurve;
  return  Math::PlanePath::DragonCurve->new;
};


#------------------------------------------------------------------------------

our @dir4_to_dx = (1,0,-1,0);
our @dir4_to_dy = (0,1,0,-1);
our @dir8_to_dx = (1,1, 0,-1, -1,-1, 0,1);
our @dir8_to_dy = (0,1, 1,1,   0,-1, -1,-1);

sub dir4_to_dx {
  my ($dir) = @_;
  return $dir4_to_dx[$dir&3];
}
sub dir4_to_dy {
  my ($dir) = @_;
  return $dir4_to_dy[$dir&3];
}

#------------------------------------------------------------------------------
# Generic

# Return the length of $n in binary, with no high 0s.
# Eg. $n==5 binary 101 is length 3.  $n==0 is length 0.
sub bitlength {
  my ($n) = @_;
  my $k = 0;
  while ($n) { $k++; $n>>=1; }
  return $k;
}

use constant::defer FLAT_any_binary => sub {
  return FLAT::Regex->new('(0|1)*')->as_dfa->MyFLAT::minimize;
};
use constant::defer FLAT_zeros => sub {
  return FLAT::Regex->new('0*')->as_dfa->MyFLAT::minimize
    ->MyFLAT::set_name("all-0s");
};
use constant::defer FLAT_ones => sub {
  return FLAT::Regex->new('1*')->as_dfa->MyFLAT::minimize
    ->MyFLAT::set_name("all-1s");
};


#------------------------------------------------------------------------------
# IPC::Run bits

our @ipc;
END {
  foreach my $h (@ipc) {
    $h->finish;
  }
}

#------------------------------------------------------------------------------

# return $x,$y multiplied by 1+i
sub mul {
  my ($x,$y) = @_;
  # (x,y) + rot+90(x,y), is multiply (i+1)
  return ($x - $y,
          $y + $x);
}
sub bk {
  my ($k) = @_;
  my $y = 0 * $k;  # inherit bigint
  my $x = $y + 1;
  foreach (1 .. $k) { ($x,$y) = mul($x,$y); }
  return ($x, $y);
}

# return $x,$y multiplied by 1-i
sub div {
  my ($x,$y) = @_;
  # (x,y) + rot+90(x,y), is multiply (1-i)
  return ($x + $y,
          $y - $x);
}

# return $x,$y unfolded around pivot $ex,$ey
sub unfold {
  my ($x,$y, $ex,$ey) = @_;
  my $dx = $ex-$x;
  my $dy = $ey-$y;
  ($dx,$dy) = (-$dy,$dx);  # rotate +90
  return ($ex+$dx, $ey+$dy);
}

sub xy_rotate_plus90 {
  my ($x,$y, $reps) = @_;
  $reps //= 1;
  $reps &= 3;
  foreach (1 .. $reps) {
    ($x,$y) = (-$y,$x);  # rotate +90
  }
  return ($x,$y);
}
sub xy_rotate_minus90 {
  my ($x,$y) = @_;
  return ($y,-$x);  # rotate -90
}
sub xy_rotate_180 {
  my ($x,$y) = @_;
  return (-$x,-$y);  # rotate 180
}

# return 0 <= dir < 4
sub xy_to_dir4 {
  my ($x, $y) = @_;
  require Math::NumSeq::PlanePathDelta;
  return Math::NumSeq::PlanePathDelta::_dxdy_to_dir4($x,$y);
}
# return 0..7
sub xy_to_dir8 {
  my ($x, $y) = @_;
  return 2*xy_to_dir4($x,$y);
}

sub rect_corners_sort {
  my ($x1,$y1, $x2,$y2) = @_;
  if ($x1 > $x2) {
    ($x1,$x2) = ($x2,$x1);
  }
  if ($y1 > $y2) {
    ($y1,$y2) = ($y2,$y1);
  }
  return ($x1,$y1, $x2,$y2);
}


#------------------------------------------------------------------------------

sub chamfer {
  my ($delta, $name) = @_;
  $name //= '\MyChamfer';
  if ($delta == 0) { return ''; }
  if ($delta == 1) { return "+$name"; }
  if ($delta == -1) { return "-$name"; }
  return sprintf '%+d*%s', $delta, $name;
}

sub transform_to_horizontal {
  my ($x,$y, $level) = @_;
  $level %= 8;
  while ($level >= 2) {
    ($x,$y) = xy_rotate_minus90($x,$y);
    $level -= 2;
  }
  if ($level) {
    ($x,$y) = div($x,$y);
  }
  return ($x,$y);
}

#------------------------------------------------------------------------------
# State Tables

sub state_table_transitions {
  my ($state_table,$state,$aref) = @_;
  foreach my $bit (@$aref) {  # high to low
    $state = $state_table->[$state]->[$bit];
  }
  return $state;
}


#------------------------------------------------------------------------------
# Turn Sequence

# turn at point $n, 1=left, -1=right.  Must have $n>=1.
sub turn {
  my ($n) = @_;
  unless ($n >= 1) { croak "turn() is for n>=1"; }
  if (BitAboveLowestOne($n)) {
    return -1;
  } else {
    return 1;
  }
}
sub TurnLpred {   # 1=left,0=right
  my ($n) = @_;
  return ($n >= 1 && BitAboveLowestOne($n)==0 ? 1 : 0);
}
sub TurnRpred {   # 0=left,1=right
  my ($n) = @_;
  return ($n >= 1 && BitAboveLowestOne($n) ? 1 : 0);
}
# Paperfolding sequence, 1=left,0=right.  Must have $n>=1.
sub Paper {
  my ($n) = @_;
  return 1 - BitAboveLowestOne($n);
}

# turn of midpoint curve at point $n, 1=left, 0=straight, -1=right.
# Must have $n>=1.
sub Mturn {
  my ($n) = @_;
  unless ($n >= 1) { croak "Mturn() is for n>=1"; }
  my $t = turn($n);
  if ($t == turn($n+1)) {
    return $t;
  } else {
    return 0;
  }
}

sub BitAboveLowestOne_allowing_zero {
  my ($n) = @_;
  return ($n == 0 ? 0 : BitAboveLowestOne($n));
}
sub BitAboveLowestOne {
  my ($n) = @_;
  $n >= 1 or croak "BitAboveLowestOne() is for n>=1";
  until ($n & 1) { $n >>= 1; }
  return (($n>>1) & 1);
}
sub BitAboveLowestZero {
  my ($n) = @_;
  $n >= 0 or croak "BitAboveLowestZero() is for n>=0";
  while ($n & 1) { $n >>= 1; }
  return (($n>>1) & 1);
}

our @BitAboveLowestOne_HtoL_state_table = ([ 0, 1 ],  # 00
                                           [ 0, 3 ],  # 01
                                           [ 2, 1 ],  # 10
                                           [ 2, 3 ]); # 11
our @BitAboveLowestOne_HtoL_state_to_result = (0,0,1,1);

sub BitAboveLowestOne_HtoL_flat {
  my ($bit) = @_;
  require MyFLAT;
  my @accepting_list
    = grep {$BitAboveLowestOne_HtoL_state_to_result[$_] == $bit}
    0 .. $#BitAboveLowestOne_HtoL_state_table
    or croak "BitAboveLowestOne_HtoL_flat() unrecognised bit=$bit";
  return MyFLAT::aref_to_FLAT_DFA
    (\@BitAboveLowestOne_HtoL_state_table,
     name => "BitAboveLowestOne_HtoL of n is $bit",
     starting => 0,
     accepting_list => \@accepting_list);
}
memoize('BitAboveLowestOne_HtoL_flat', LIST_CACHE=>'MERGE');

our @BitAboveLowestOne_LtoH_state_table = ([ 0, 1 ],  # S
                                           [ 2, 3 ],  # T
                                           [ 2, 2 ],  # r0
                                           [ 3, 3 ]); # r1
our @BitAboveLowestOne_LtoH_state_to_result = (0,0,0,1);
our @BitAboveLowestOne_LtoH_state_name = ('S','T','r0','r1');

sub BitAboveLowestOne_LtoH_flat {
  my ($bit) = @_;
  require MyFLAT;
  my @accepting_list
    = grep {$BitAboveLowestOne_LtoH_state_to_result[$_] == $bit}
    0 .. $#BitAboveLowestOne_LtoH_state_table
    or croak "BitAboveLowestOne_LtoH_flat() unrecognised bit=$bit";
  return MyFLAT::aref_to_FLAT_DFA
    (\@BitAboveLowestOne_LtoH_state_table,
     name => "BitAboveLowestOne_LtoH of n is $bit",
     starting => 0,
     accepting_list => \@accepting_list);
}
memoize('BitAboveLowestOne_LtoH_flat', LIST_CACHE=>'MERGE');


#------------------------------------------------------------------------------
# dNextL

sub dNextL {
  my ($n) = @_;
  foreach my $d (1 .. 4) {
    if (TurnLpred($n+$d)) { return $d; }
  }
  die "oops, no next TurnLpred";
}

our @dNextL_state_table = ([ 0, 1 ],  # 0 d1a
                           [ 2, 1 ],  # 1 d1b
                           [ 0, 4 ],  # 2 d2a
                           [ 5, 3 ],  # 3 d2b
                           [ 5, 3 ],  # 4 d3a
                           [ 0, 6 ],  # 5 d3b
                           [ 5, 3 ]); # 6 d4
our @dNextL_state_to_d = (1,1, 2,2, 3,3, 4); # dNextL values
our @dNextL_state_to_name
  = ('d1a', 'd1b', 'd2a', 'd2b', 'd3a', 'd3b', 'd4');

sub dNextL_flat {
  my ($d) = @_;
  require MyFLAT;
  my @accepting_list
    = grep {$dNextL_state_to_d[$_] == $d} 0 .. $#dNextL_state_table
    or croak "dNextL_flat() unrecognised d=$d";
  return MyFLAT::aref_to_FLAT_DFA
    (\@dNextL_state_table,
     name => "dNextL of n is $d",
     starting => 0,
     accepting_list => \@accepting_list);
}
memoize('dNextL_flat', LIST_CACHE=>'MERGE');


our @dNextL_LtoH_state_table = ([  1,  2 ],   # 0 start
                                [ 10,  4 ],   # 1 s0
                                [  3,  5 ],   # 2 s1
                                [ 10,  8 ],   # 3 s01
                                [  6,  4 ],   # 4 s10
                                [  7,  5 ],   # 5 s11
                                [ 11, 12 ],   # 6 s010
                                [ 10, 11 ],   # 7 s011
                                [  9,  8 ],   # 8 s101
                                [ 12, 13 ],   # 9 s0101
                                [ 10, 10 ],   # 10 d1
                                [ 11, 11 ],   # 11 d2
                                [ 12, 12 ],   # 12 d3
                                [ 13, 13 ]);  # 13 d4
our @dNextL_LtoH_state_to_d = (1, 1,1, 1,2,1, 2,1,3, 3, 1,2,3,4);
our @dNextL_LtoH_state_to_name
  = ('start', 's0','s1', 's01','s10','s11', 's010','s011','s101', 's0101',
     'd1','d2','d3','d4');

sub dNextL_LtoH_flat {
  my ($d) = @_;
  require MyFLAT;
  my @accepting_list
    = grep {$dNextL_LtoH_state_to_d[$_] == $d} 0 .. $#dNextL_LtoH_state_table
    or croak "dNextL_LtoH_flat() unrecognised d=$d";
  return MyFLAT::aref_to_FLAT_DFA
    (\@dNextL_LtoH_state_table,
     name => "dNextL of n is $d",
     starting => 0,
     accepting_list => \@accepting_list);
}
memoize('dNextL_LtoH_flat', LIST_CACHE=>'MERGE');


our @dNextLsub1_state_table = ([  0, 4 ],  # 0 d1a
                               [  2, 1 ],  # 1 d1b
                               [  7, 6 ],  # 2 d1c
                               [  5, 3 ],  # 3 d1d
                               [  2, 1 ],  # 4 d2a
                               [ 10, 9 ],  # 5 d2b
                               [  5, 3 ],  # 6 d2c
                               [  0, 8 ],  # 7 d3a
                               [  2, 1 ],  # 8 d3b
                               [  5, 3 ],  # 9 d3c
                               [  0, 8 ]); # 10 d4
our @dNextLsub1_state_to_d = (1,1,1,1, 2,2,2, 3,3,3, 4); # dNextLsub1 values
our @dNextLsub1_state_to_name
  = ('d1a','d1b','d1c','d1d', 'd2a','d2b','d2c', 'd3a','d3b','d3c', 'd4');

sub dNextLsub1_flat {
  my ($d) = @_;
  require MyFLAT;
  my @accepting_list
    = grep {$dNextLsub1_state_to_d[$_] == $d} 0 .. $#dNextLsub1_state_table
    or croak "dNextLsub1_flat() unrecognised d=$d";
  return MyFLAT::aref_to_FLAT_DFA
    (\@dNextLsub1_state_table,
     name => "dNextLsub1 of n is $d",
     starting => 0,
     accepting_list => \@accepting_list);
}
memoize('dNextLsub1_flat', LIST_CACHE=>'MERGE');


our @dNextLsub1_M_state_table = ([  0, 3 ],  # 0 d1a
                                 [  2, 1 ],  # 1 d1b
                                 [  5, 8 ],  # 2 d1c
                                 [  2, 1 ],  # 3 d2a
                                 [  7, 8 ],  # 4 d2b
                                 [  0, 6 ],  # 5 d3a
                                 [  2, 1 ],  # 6 d3b
                                 [  0, 6 ],  # 7 d4
                                 [  4, 8 ]); # 8 m
our @dNextLsub1_M_state_to_d = (1,1,1, 2, 2,3, 3, 4, undef); # dNextLsub1 values
our @dNextLsub1_M_state_to_name
  = ('d1a','d1b','d1c', 'd2a','d2b', 'd3a','d3b', 'd4', 'm');

sub dNextLsub1_M_flat {
  my ($d) = @_;
  require MyFLAT;
  my @accepting_list
    = (grep {($dNextLsub1_M_state_to_d[$_]//-1) == $d}
       0 .. $#dNextLsub1_M_state_table)
    or croak "dNextLsub1_M_flat() unrecognised d=$d";
  return MyFLAT::aref_to_FLAT_DFA
    (\@dNextLsub1_M_state_table,
     name => "dNextLsub1 of n is $d",
     starting => 0,
     accepting_list => \@accepting_list);
}
memoize('dNextLsub1_M_flat', LIST_CACHE=>'MERGE');


#------------------------------------------------------------------------------
# dNextR

sub dNextR {
  my ($n) = @_;
  foreach my $d (1 .. 4) {
    if (TurnRpred($n+$d)) { return $d; }
  }
  die "oops, no next TurnRpred";
}

our @dNextR_state_table = ([ 3, 1 ],  # 0 d1a
                           [ 0, 1 ],  # 1 d1b
                           [ 0, 5 ],  # 2 d2a
                           [ 4, 2 ],  # 3 d2b
                           [ 4, 2 ],  # 4 d3a
                           [ 0, 6 ],  # 5 d3b
                           [ 0, 6 ]); # 6 d4
our @dNextR_state_to_d = (1,1, 2,2, 3,3, 4); # dNextR values
our @dNextR_state_to_name
  = ('d1a','d1b', 'd2a','d2b', 'd3a','d3b', 'd4');
our $dNextR_starting = 4; # d3a

sub dNextR_flat {
  my ($d) = @_;
  require MyFLAT;
  my @accepting_list
    = grep {$dNextR_state_to_d[$_] == $d} 0 .. $#dNextR_state_table
    or croak "dNextR_flat() unrecognised d=$d";
  return MyFLAT::aref_to_FLAT_DFA
    (\@dNextR_state_table,
     name => "dNextR of n is $d",
     starting => $dNextR_starting,
     accepting_list => \@accepting_list);
}
memoize('dNextR_flat', LIST_CACHE=>'MERGE');


#------------------------------------------------------------------------------
# dTurnLeft_flat()

our @dTurnLeft_state_table = ([  0,  7 ],    # 0 d1a
                              [  8,  1 ],    # 1 d1b                    
                              [ 10,  2 ],    # 2 d1c                    
                              [  8,  1 ],    # 3 d2                     
                              [  0,  7 ],    # 4 d3a                    
                              [ 10,  2 ],    # 5 d3b                    
                              [  0,  7 ],    # 6 d4                     
                              [  8,  3 ],    # 7 s1    NF  decr      
                              [  4,  9 ],    # 8 s2    NF  decr      
                              [ 10,  5 ],    # 9 s3    NF  decr     
                              [  6,  9 ]);   # 10 s4   NF  decr     
our @dTurnLeft_state_to_decr = (0,0,0, 0,0,0,0, 1,1,1,1);
our @dTurnLeft_state_to_d = (1,1,1, 2, 3,3, 4,   # dTurnLeft values
                             undef,undef,undef,undef);
our @dTurnLeft_state_to_name
  = ('d1a','d1b','d1c', 'd2', 'd3a','d3b', 'd4', 's1','s2','s3','s4');

sub dTurnLeft_flat {
  my ($d) = @_;
  my @accepting_list
    = grep {($dTurnLeft_state_to_d[$_]//0) == $d} 0 .. $#dTurnLeft_state_table
    or croak "dTurnLeft_flat() unrecognised d=$d";
  return MyFLAT::aref_to_FLAT_DFA
    (\@dTurnLeft_state_table,
     name => "dTurnLeft=$d of bits seen",
     accepting_list => \@accepting_list);
}
memoize('dTurnLeft_flat', LIST_CACHE=>'MERGE');

# $state is '' when have not yet seen a 1, so still changing 0s to 1s.
# $state is 'done' when have seen a 1, so subsequent digits unchanged.
sub transmute_initial_0s_to_1s {
  my ($state,$symbol) = @_;
  unless ($state) {
    if ($symbol eq '1') { $state = 'done'; }   # stop when found 1
    else { $symbol = 1; } # low 0 -> 1
  }
  return ($state,$symbol);
}


#------------------------------------------------------------------------------
# dTurnRight

our @dTurnRight_state_table = ([ 0, 6 ],   # 0 d3a
                               [ 7, 2 ],   # 1 d3b
                               [ 7, 2 ],   # 2 d4 
                               [ 5, 6 ],   # 3 d1a
                               [ 7, 4 ],   # 4 d1b
                               [ 0, 6 ],   # 5 d2
                               [ 7, 1 ],   # 7 s1       incr
                               [ 3, 8 ],   # 6 s2       incr
                               [ 7, 4 ]);  # 8 s2       incr
our @dTurnRight_state_to_incr = (0,0,0, 0,0,0, 1,1,1);
our @dTurnRight_state_to_d
  = (3,3,4, 1,1,2, undef,undef,undef); # dTurnRight values
our @dTurnRight_state_to_name
  = ('d3a','d3b','d4', 'd1a','d1b','d2', 's1','s2','s3');

sub dTurnRight_flat {
  my ($d) = @_;
  require MyFLAT;
  my @accepting_list
    = grep {($dTurnRight_state_to_d[$_]//0) == $d} 0 .. $#dTurnRight_state_table
    or croak "dTurnRight_flat() unrecognised d=$d";
  return MyFLAT::aref_to_FLAT_DFA
    (\@dTurnRight_state_table,
     name => "dTurnRight=$d of bits seen",
     accepting_list => \@accepting_list);
}
memoize('dTurnRight_flat', LIST_CACHE=>'MERGE');

# $state is '' when not yet seen and changed a 1.
# $state is 'done' when have done so, and so subsequent digits unchanged.
sub transmute_lowest_1_to_0 {
  my ($state,$symbol) = @_;
  if (!$state && $symbol eq '1') {
    # found lowest 1, change it to 0, and no more changes
    return ('done', 0);
  }
  return ($state,$symbol);
}


#------------------------------------------------------------------------------
# Boundary

# Return the number of left boundary squares on dragon curve level $k.
sub LQ {
  my ($k) = @_;
  if ($k == 0) { return 1; }
  my $l0 = 1;  # k=1
  my $l1 = 2;  # k=2
  my $l2 = 4;  # k=3
  for (2 .. $k) {
    ($l0,$l1,$l2) = ($l1,$l2,  $l2 + 2*$l0);
  }
  return $l0;
}
BEGIN { memoize('LQ', LIST_CACHE=>'MERGE') }

# Return the number of right boundary squares on dragon curve level $k.
sub RQ {
  my ($k) = @_;
  my $ret = 1;
  foreach my $i (0 .. $k-1) { $ret += LQ($i); }
  return $ret;
}
BEGIN { memoize('RQ', LIST_CACHE=>'MERGE') }

sub LQ3o {
  my ($k) = @_;
  return ($k==0 ? 0 : dJA($k-1));
}
sub RQ3o {
  my ($k) = @_;
  return ($k==0 ? 0 : JA($k-1));
}
sub BQ3o {
  my ($k) = @_;
  return RQ3o($k+1);
}

# Return the left boundary length of dragon curve level $k.
sub L {
  my ($k) = @_;
  if ($k == 0) { return 1; }
  my $l0 = 2;  # k=1
  my $l1 = 4;  # k=2
  my $l2 = 8;  # k=3
  for (2 .. $k) {
    ($l0,$l1,$l2) = ($l1,$l2,  $l2 + 2*$l0);
  }
  return $l0;
}
BEGIN { memoize('L', LIST_CACHE=>'MERGE') }

# Return the right boundary length of dragon curve level $k.
sub R {
  my ($k) = @_;
  my $ret = 1;
  foreach my $i (0 .. $k-1) { $ret += L($i); }
  return $ret;
}
BEGIN { memoize('R', LIST_CACHE=>'MERGE') }

# boundary length count of segments
sub B {
  my ($k) = @_;
  return L($k) + R($k);
}

sub RJ {
  my ($k) = @_;
  return R($k+2)/2 - ceil(2*$k/3);
}


#------------------------------------------------------------------------------
# Convex Hull

# Hull around fractal 0..1, multiplied by 6 so integers.
# Points P10, P9, ..., P1 same order as GP code hull_fractal.
our $hull6_points = [ [ 5, -3],  # P10 for (5 - 3*I)/6  (etc)
                      [ 7,  0],  # P9
                      [ 7,  1],  # P8
                      [ 6,  2],  # P7
                      [ 4,  2],  # P6
                      [-1,  1],  # P5
                      [-2,  0],  # P4
                      [-2, -2],  # P3
                      [ 0, -4],  # P2
                      [ 4, -4],  # P1
                    ];
# hull_fractal*6  \\ P10 to P1

sub hull_expand {
  my ($points, $ex,$ey, %options) = @_;
  ### hull_expand() ...

  if (@$points < 5) {
    ### small ...
    $points = [ points_midpoint($points->[0],$points->[1], 1,4),
                points_midpoint($points->[0],$points->[1], 1,2),
                points_midpoint($points->[0],$points->[1], 3,4),
                @$points ];
  }

  push @$points, map{[ unfold(@$_, $ex,$ey) ]} @$points;
  ### at: "pxy=$ex,$ey  ".points_str($points)

  $points = points_to_hull($points);
  unless ($options{'unsorted'}) {
    $points = hull_points_sort_by_angle($points, $ex,$ey);
  }
  ($ex,$ey) = mul($ex,$ey);
  return ($points, $ex,$ey);
}

sub points_midpoint {
  my ($p1,$p2, $num,$den) = @_;
  $num //= 1;
  $den //= 2;
  return [ $p1->[0]*$num/$den + $p2->[0]*($den-$num)/$den,
           $p1->[1]*$num/$den + $p2->[1]*($den-$num)/$den ];
}

sub hull_points_sort_by_angle {
  my ($points, $ex,$ey) = @_;
  return [ sort {hull_point_cmp_by_angle($a,$b, $ex,$ey)} @$points ];
}
sub hull_point_cmp_by_angle {
  my ($a,$b, $ex,$ey) = @_;
  require Math::NumSeq::PlanePathDelta;
  my $a4 = Math::NumSeq::PlanePathDelta::_dxdy_to_dir4(@$a);
  my $b4 = Math::NumSeq::PlanePathDelta::_dxdy_to_dir4(@$b);
  if ($a->[0] == 0 && $a->[1] == 0) { $a4 = 2.5; }
  if ($b->[0] == 0 && $b->[1] == 0) { $b4 = 2.5; }
  my $o4 = Math::NumSeq::PlanePathDelta::_dxdy_to_dir4($ex,$ey);
  $a4 = ($a4 - $o4 - .1); while ($a4 < 0) { $a4 += 4; }
  $b4 = ($b4 - $o4 - .1); while ($b4 < 0) { $b4 += 4; }
  return ($b4 <=> $a4);
}

sub P1_xy {
  my ($k) = @_;
  my ($x,$y) = bk($k+3);
  my ($px,$py) = p_xy($k+3);
  return (-($x+$px)/3, -($y+$py)/3);
}
sub P2_xy {
  my ($k) = @_;
  my ($x,$y) = bk($k+2);
  my ($px,$py) = p_xy($k+2);
  return (-($x+$px)/3, -($y+$py)/3);
}
sub P3_xy {
  my ($k) = @_;
  my ($x,$y) = bk($k+1);
  my ($px,$py) = p_xy($k+1);
  return (-($x+$px)/3, -($y+$py)/3);
}
sub P4_xy {
  my ($k) = @_;
  my ($x,$y) = bk($k);
  my ($px,$py) = p_xy($k);
  return (-($x+$px)/3, -($y+$py)/3);
}
sub P5_xy {
  my ($k) = @_;
  my ($x,$y) = bk($k-1);
  my ($px,$py) = p_xy($k-1);
  return (-($x+$px)/3, -($y+$py)/3);
}
sub P6_xy {
  my ($k) = @_;
  my ($x,$y) = bk($k);
  ($x,$y) = (2*$x-$y, $x+2*$y);  # (x+I*y)*(2+I)
  my ($px,$py) = p_xy($k+1);
  ($px,$py) = xy_rotate_plus90($px,$py);  # (px+I*py)*I
  return (($x+$px)/3, ($y+$py)/3);
}
sub P7_xy {
  my ($k) = @_;
  my ($x,$y) = bk($k);
  ($x,$y) = (3*$x-$y, $x+3*$y);  # (x+I*y)*(3+I)
  my ($px,$py) = p_xy($k);
  ($px,$py) = xy_rotate_plus90($px,$py);  # (px+I*py)*I
  return (($x+$px)/3, ($y+$py)/3);
}
sub P8_xy {
  my ($k) = @_;
  my ($x,$y) = bk($k-1);
  ($x,$y) = (3*$x-4*$y, 4*$x+3*$y);  # (x+I*y)*(3+4*I)
  my ($px,$py) = p_xy($k-1);
  ($px,$py) = xy_rotate_plus90($px,$py);  # (px+I*py)*I
  return (($x+$px)/3, ($y+$py)/3);
}
sub P9_xy {
  my ($k) = @_;
  my ($x,$y) = bk($k-2);
  ($x,$y) = (-7*$y, 7*$x);             # (x+I*y)*7*I
  my ($px,$py) = p_xy($k-2);
  ($px,$py) = xy_rotate_plus90($px,$py);  # (px+I*py)*I
  return (($x+$px)/3, ($y+$py)/3);
}
sub P10_xy {
  my ($k) = @_;
  my ($x,$y) = bk($k-1);
  ($x,$y) = (4*$x-$y, $x+4*$y);  # (x+I*y)*(4+1*I)
  my ($px,$py) = p_xy($k);
  return (($x+$px)/3, ($y+$py)/3);
}
{
  my @p_xy_vector = ([-1,  0],
                     [-1, -1],
                     [ 0, -2],
                     [-1, -2],
                     [ 1,  0],
                     [ 1,  1],
                     [ 0,  2],
                     [ 1,  2]);
  sub p_xy {
    my ($k) = @_;
    return @{$p_xy_vector[$k % 8]};
  }
}

{
  my @HBV_vector = (15, 17, 22, 26);
  sub HBV {
    my ($k) = @_;
    if ($k==0) { return 2; }
    if ($k==1) { return 3; }
    return $HBV_vector[$k%4] * 2**int($k/4) / 2 - 6;
  }
}

{
  my @P1N_vector = (7, -1, -2, -4);
  sub P1N {
    my ($k) = @_;
    (2**($k+3) + $P1N_vector[$k%4]) / 15;
  }
}
sub P3N {
  my ($k) = @_;
  return P1N($k-2);
}
sub P5N {
  my ($k) = @_;
  return P1N($k-4);
}


#------------------------------------------------------------------------------
# blob convex hull

sub BP1_xy {
  my ($k) = @_;
  if ($k < 4) { croak "No blob k=$k"; }
  return P1_xy($k);
}
sub BP2_xy {        # (3/6-4/6*I)*b^k + 2/6*p(k-2)
  my ($k) = @_;
  if ($k < 4) { croak "No blob k=$k"; }
  my ($x,$y) = bk($k);
  # GP-Test  (x+y*I)*(3-4*I) == (3*x+4*y) + (-4*x+3*y)*I
  ($x,$y) = (3*$x+4*$y, -4*$x+3*$y);
  my ($px,$py) = p_xy($k-2);
  return (($x+2*$px)/6, ($y+2*$py)/6);
}
sub BP3_xy {      # (5/12-7/12*I)*b^k + 4/12*p(k-3)
  my ($k) = @_;
  if ($k < 4) { croak "No blob k=$k"; }
  my ($x,$y) = bk($k);
  # GP-Test  (x+y*I)*(5-7*I) == (5*x+7*y) + (-7*x+5*y)*I
  ($x,$y) = (5*$x+7*$y, -7*$x+5*$y);
  my ($px,$py) = p_xy($k-3);
  return (($x+4*$px)/12, ($y+4*$py)/12);
}
{
  my @BP4_x_vector = (0,  -1, -1,  0,  0, 1, 1, 0);
  my @BP4_y_vector = (-1,  0, -1, -1,  1, 0, 1, 1);
  sub BP4_xy {
    my ($k) = @_;
    if ($k < 4) { croak "No blob k=$k"; }
    my ($x,$y) = BlobStart_xy($k);
    $x += $BP4_x_vector[$k%8];
    $y += $BP4_y_vector[$k%8];
    return ($x,$y);
  }
}
{
  my @BP4dash_x_vector = (0,0, 4,2,  0,0, -4,-2);
  my @BP4dash_y_vector = (0,0, 1,3,  0,0, -1,-3);
  sub BP4dash_xy {
    my ($k) = @_;
    if ($k < 4) { croak "No blob k=$k"; }
    my ($x,$y) = BP4_xy($k);
    $x += $BP4dash_x_vector[$k%8];
    $y += $BP4dash_y_vector[$k%8];
    return ($x,$y);
  }
}
sub BP5_xy {      # (1/3+1/3*I)*b^k + 1/3*I*p(k+2)
  my ($k) = @_;
  if ($k < 4) { croak "No blob k=$k"; }
  my ($x,$y) = bk($k);
  my ($px,$py) = p_xy($k+2);
  ($px,$py) = xy_rotate_plus90($px,$py);  # (px+I*py)*I
  return (($x+$px)/3, ($y+$py)/3);
}
sub BP6_xy {      # (5/12+1/12*I)*b^k + 1/3*I*p(k+1)
  my ($k) = @_;
  if ($k < 4) { croak "No blob k=$k"; }
  my ($x,$y) = bk($k);
  # GP-Test  (x+y*I)*(5+I) == (5*x-y) + (x+5*y)*I
  ($x,$y) = (5*$x-$y, $x+5*$y);
  my ($px,$py) = p_xy($k+1);
  ($px,$py) = xy_rotate_plus90($px,$py);  # (px+I*py)*I
  return (($x+4*$px)/12, ($y+4*$py)/12);
}
sub BP7_xy {
  my ($k) = @_;
  if ($k < 4) { croak "No blob k=$k"; }
  my ($x,$y) = bk($k);
  # GP-Test  (x+y*I)*(6+I) == (6*x-y) + (x+6*y)*I
  ($x,$y) = (6*$x-$y, $x+6*$y);
  my ($px,$py) = p_xy($k);
  ($px,$py) = xy_rotate_plus90($px,$py);  # (px+I*py)*I
  return (($x+4*$px)/12, ($y+4*$py)/12);
}
{
  my @BP8_x_vector = ( 1, -1,  0, -1,  -1, 1,  0,  1);
  my @BP8_y_vector = ( 0,  0,  1,  1,   0, 0, -1, -1);
  sub BP8_xy {
    my ($k) = @_;
    if ($k < 4) { croak "No blob k=$k"; }
    my ($x,$y) = BlobEnd_xy($k);
    $x += $BP8_x_vector[$k%8];
    $y += $BP8_y_vector[$k%8];
    return ($x,$y);
  }
}
{
  my @BP8dash_x_vector = (-3,0,0, 1,  3,0,0,-1);
  my @BP8dash_y_vector = ( 2,0,0,-4, -2,0,0, 4);
  sub BP8dash_xy {
    my ($k) = @_;
    if ($k < 4) { croak "No blob k=$k"; }
    my ($x,$y) = BP8_xy($k);
    $x += $BP8dash_x_vector[$k%8];
    $y += $BP8dash_y_vector[$k%8];
    return ($x,$y);
  }
}
sub BP9_xy {        # (5/6-1/6*I)*b^k + 2/6*p(k+9)
  my ($k) = @_;
  if ($k < 4) { croak "No blob k=$k"; }
  my ($x,$y) = bk($k);
  # GP-Test  (x+y*I)*(5-1*I) == (5*x+1*y) + (-1*x+5*y)*I
  ($x,$y) = (5*$x+1*$y, -1*$x+5*$y);
  my ($px,$py) = p_xy($k+9);
  return (($x+2*$px)/6, ($y+2*$py)/6);
}
sub BP10_xy {
  my ($k) = @_;
  if ($k < 4) { croak "No blob k=$k"; }
  return P10_xy($k);
}

sub blob_hull_points_sort {
  my ($points, $k) = @_;
  if ($k < 4) { croak "No blob k=$k"; }
  my $transformed_points = [ map {[transform_to_horizontal(@$_,$k)]} @$points ];
  ### $transformed_points

  # BP1 is lowest Y, and the rightmost of the two such
  my $BP1_y = min(map {$_->[1]} @$transformed_points);
  my $BP1_x = max(map {$_->[0]} grep {$_->[1]==$BP1_y} @$transformed_points);
  my ($BP1_i) = grep {$transformed_points->[$_]->[0] == $BP1_x
                       && $transformed_points->[$_]->[1] == $BP1_y
                     } 0 .. $#$transformed_points
                     or die;
  $points = [ map {$points->[($BP1_i-$_)%scalar(@$points)]} 0 .. $#$points ];
  return $points;
}


#------------------------------------------------------------------------------
# sequences

{
  my @sG_real = (2, -1, -2, 1);
  sub sG_real {
    my ($k) = @_;
    return (3 * 2**$k + $sG_real[$k % 4]) / 5;
  }
}

{
  my @sG_imag = (-1, -2, 1, 2);
  sub sG_imag {
    my ($k) = @_;
    return (2**$k + $sG_imag[$k % 4]) / 5;
  }
}

sub BlobP {
  my ($k) = @_;
  return BlobS($k) + BlobA($k);
}
sub BlobS {
  my ($k) = @_;
  return ($k < 4 ? 0 : BlobB($k)/2 + 1);
}

# k=5  7+4+2==13
sub BlobPrecedingP {
  my ($k) = @_;
  my $total = 7;  # initial 7 segments with 1 point before each
  foreach my $i (4 .. $k-1) {
    $total += BlobP($i) + 2;
  }
  return $total;
}

{
  my @BlobB_initial = (0, 0, 0, 0, 4, 10, 20, 38, 68, 122);
  sub BlobB {
    my ($k) = @_;
    if ($k < 10) { return $BlobB_initial[$k]; }
    return 2 * BlobB($k-1)
      - 2 * BlobB($k-2)
      + 4 * BlobB($k-3)
      - 3 * BlobB($k-4)
      + 2 * BlobB($k-5)
      - 2 * BlobB($k-6);
  }
  BEGIN { memoize('BlobB', LIST_CACHE=>'MERGE'); }
}

{
  my @BlobA_initial = (0,0,0,0, 1, 3);
  sub BlobA {
    my ($k) = @_;
    if ($k < 6) { return $BlobA_initial[$k]; }
    return 3 * BlobA($k-1)
      - 3 * BlobA($k-2)
      + 5 * BlobA($k-3)
      - 6 * BlobA($k-4)
      + 2 * BlobA($k-5)
      - 4 * BlobA($k-6);
  }
  BEGIN { memoize('BlobA', LIST_CACHE=>'MERGE'); }
}

sub BlobN {
  my ($k) = @_;
  if ($k < 4) { croak "No blob k=$k"; }
  my $ret = 7;
  foreach my $i (5 .. $k) {
    $ret = 2*$ret - (($i-1) % 4 < 2 ? 0 : 3);
  }
  return $ret;
}

sub BlobNend {
  my ($k) = @_;
  if ($k < 4) { croak "No blob k=$k"; }
  return BlobN($k+1)-3;
}

sub BlobSegments {
  my ($k) = @_;
  return ($k < 4 ? 0 : BlobN($k+1) - BlobN($k) - 3);
}
sub BridgeCount {
  my ($k) = @_;
  return ($k<=4 ? 0 : 2*$k-9);
}
sub BridgeLother {
  my ($k) = @_;
  return P3N($k);
}
sub BridgeRother {
  my ($k) = @_;
  return 2*2**$k - P1N($k);
}

{
  my @BlobDeltaLen = (4, 5, 5, 7);
  sub BlobDeltaLen {
    my ($k) = @_;
    if ($k < 4) { return 0; }
    return (2**int(($k+1)/2) - $BlobDeltaLen[$k%4])/3;
  }
}

{
  my @BlobPathLen = (4, 5, 5, 7);
  sub BlobPathLen {
    my ($k) = @_;
    if ($k < 4) { return 0; }
    return ($k%2 ? 1 : 2) * (2**int(($k+1)/2) - $BlobPathLen[$k%4])/3;
  }
}

{
  sub Diameter {
    my ($k) = @_;
    if ($k == 0) { return 1; }
    if ($k % 2 == 0) { return 3*2**int($k/2) + $k - 4 }
    else             { return 4*2**int($k/2) + $k - 3 }
  }
}

{
  my @boff = ([2, 1],
              [1, 3],
              [-2, 4],
              [-3, 2],
              [-2, -1],
              [-1, -3],
              [2, -4],
              [3, -2]);
  sub boff {
    my ($k) = @_;
    return @{$boff[$k % 8]};
  }
}
# BlobStart(k) = 1/3*(conj(b)*b^k + boff(k));
sub BlobStart_xy {
  my ($k) = @_;
  if ($k < 4) {
    croak "BlobStart_xy k=$k < 4";
  }
  my ($x,$y) = bk($k);
  ($x,$y) = ($x+$y, -$x+$y);   # (x+y*I)*conj(b)
  my ($bx,$by) = boff($k);
  return (($x+$bx)/3, ($y+$by)/3);
}

# BlobEnd_xy(k) = 1/3*(2*b^k - I* boff(k-1));
sub BlobEnd_xy {
  my ($k) = @_;
  if ($k < 4) {
    croak "BlobEnd_xy k=$k < 4";
  }
  my ($x,$y) = bk($k);
  my ($bx,$by) = boff($k-1);
  ($bx,$by) = xy_rotate_plus90($bx,$by); # I*boff
  return ((2*$x - $bx)/3, (2*$y - $by)/3);
}

sub BlobDelta_xy {
  my ($k) = @_;
  my ($xs,$ys) = BlobStart_xy($k);
  my ($xe,$ye) = BlobEnd_xy($k);
  return ($xe-$xs, $ye-$ys);
}
sub BridgeDelta_xy {  # after blob $k
  my ($k) = @_;
  my ($xs,$ys) = BlobEnd_xy($k);
  my ($xe,$ye) = BlobStart_xy($k+1);
  return ($xe-$xs, $ye-$ys);
}

sub JA {
  my ($k) = @_;
  ### JA: $k
  my $j0 = 0;  # k=1
  my $j1 = 0;  # k=2
  my $j2 = 0;  # k=3
  my $j3 = 1;  # k=3
  for (1 .. $k) {
    ($j0,$j1,$j2,$j3) = ($j1,$j2,$j3,  2*$j3 - $j2 + 2*$j1 - 2*$j0);
    ### at: "$_ is  $j0, $j1, $j2, $j3"
  }
  return $j0;
}
BEGIN { memoize('JA', LIST_CACHE=>'MERGE') }

sub JN {
  my ($k) = @_;
  return (6*2**$k + 2**($k % 3) )/7;
}
sub JNother {
  my ($k) = @_;
  return (10*2**$k - 3*2**($k % 3) )/7;
}

# Return the enclosed area of dragon curve level $k (number of unit squares).
sub A {
  my ($k) = @_;
  return AL($k) + AR($k);
}

sub AR {
  my ($k) = @_;
  return (2**$k - R($k)) / 4;
}

sub AL {
  my ($k) = @_;
  return (2**$k - L($k)) / 4;
}

sub TA {
  my ($k) = @_;
  return 2*A($k+1) + LQ($k+1);
}

#------------------------------------------------------------------------------
# dir

sub Gray {
  my ($n) = @_;
  return $n ^ ($n >> 1);
}
sub Ungray {
  my ($n) = @_;
  die;
}

sub CountOneBits {
  my ($n) = @_;
  my $count = 0;
  for ( ; $n; $n>>=1) {
    $count += ($n & 1);
  }
  return $count;
}
sub dir {
  my ($n) = @_;
  return CountOneBits(Gray($n));
}
sub DirMaxN {
  my ($k) = @_;
  return ((1<<($k+1)) - 2 + ($k&1)) / 3;
}

our @dir4_state_table = ([0, 1],   # 0
                         [2, 1],   # 1
                         [2, 3],   # 2
                         [0, 3]);  # 3
sub dir4_flat {
  my ($want_dir4) = @_;
  require MyFLAT;
  MyFLAT::aref_to_FLAT_DFA(\@dir4_state_table,
                           name => "dir $want_dir4 mod 4",
                           accepting => $want_dir4);
}
BEGIN { memoize('dir4_flat', LIST_CACHE=>'MERGE'); }

sub DirRatio {
  my ($n) = @_;
  return dir($n) / bitlength($n);
}

sub TurnsL {
  my ($n) = @_;
  return ($n + dir($n)) / 2;
}
sub TurnsR {
  my ($n) = @_;
  return ($n - dir($n)) / 2;
}



#------------------------------------------------------------------------------
# point visits

# count of single-visited points in level $k
sub S {
  my ($k) = @_;
  return B($k)/2 + 1;
}
# count of double-visited points in level $k
sub D {
  my ($k) = @_;
  return A($k);
}

# count of total distinct points
sub P {
  my ($k) = @_;
  return S($k) + A($k);
}

# Return true if $n in level $k is double-visited.
sub Dpred_k {
  my ($k,$n) = @_;
  (0 <= $n && $n <= (1<<$k)) or croak "Dpred_k n=$n out of range k=$k";
  my $str = sprintf '%0*b', $k, $n;
  ### $n
  ### $str
  return $str =~ /^(.* 1   (01+|10+) (0 (01+|10+))* 01 0*
                  |(.* 0)? (01+|10+) (1 (01+|10+))* 11 0*)$/x;
}
sub Spred_k {
  my ($k,$n) = @_;
  return ! Dpred_k($k,$n);
}


#------------------------------------------------------------------------------
# centroid

# GS(k) = (2/5 - 1/5*I)*b^k + (1/10 + 1/5*I)*conj(b/2)^k;
sub GS_xy {
  my ($k) = @_;
  my ($bx,$by) = (to_bigrat(1), to_bigrat(1));
  my ($px,$py) = (to_bigrat(1), 0);
  foreach (1 .. $k) {
    ($px,$py) = mul($px,$py);
  }
  my $twos = to_bigrat(2)**$k;
  my $cx = $px / $twos;
  my $cy = - $py / $twos;
  return (to_bigrat(2)/5 * $px + to_bigrat(1)/5 * $py
          + to_bigrat(1)/10 * $cx - to_bigrat(1)/5 * $cy,

          to_bigrat(2)/5 * $py - to_bigrat(1)/5 * $px
          + to_bigrat(1)/10 * $cy + to_bigrat(1)/5 * $cx);
}

sub GLQ_xy {
  my ($k) = @_;
  if ($k < 2) { return (1/2, 1/2); }
  my ($x,$y) = bk($k);
  return ($x/2, $y/2);
}

#------------------------------------------------------------------------------
# Math::Geometry::Planar bits

sub points_to_bigrat {
  my ($points) = @_;
  return [ map { [ to_bigrat($_->[0]), to_bigrat($_->[1]) ] } @$points ];
}

sub points_to_centroid {
  my ($points) = @_;
  if (@$points < 3) {
    return [ ($points->[0]->[0] + $points->[1]->[0]) / 2,
             ($points->[0]->[1] + $points->[1]->[1]) / 2 ];
  }
  require Math::Geometry::Planar;
  my $planar = Math::Geometry::Planar->new;
  $planar->points($points);
  return $planar->centroid;
}
sub points_to_hull {
  my ($points, %options) = @_;

  require Math::Geometry::Planar;
  my $planar = Math::Geometry::Planar->new;
  $planar->points($points);
  if (@$points >= 5) {
    ### convexhull2: points_str($points)
    $planar = $planar->convexhull2;
    $points = $planar->points;
  }
  return $points;
}

# $points is an arrayref of [$x,$y] pairs.
# Return a string "x,y x,y ...".
sub points_str {
  my ($points) = @_;
  if (! defined $points) { return "undef"; }
  if (@$points == 0) { return "(empty)"; }
  return join(' ',map{point_str($_)} @$points);
}
# $p is an arrayref pair [$x,$y].
# Return a string "x,y".
sub point_str {
  my ($p) = @_;
  return  join(',',@$p);
}

# $points is an arrayref of [$x,$y] pairs.
sub points_str_gp {
  my ($points) = @_;
  return join(', ',map{point_str_gp($_)}@$points);
}
sub point_str_gp {
  my ($p) = @_;
  my ($x,$y) = @$p;
  my $ret = '';
  if ($x != 0 || $y == 0) {
    $ret = "$x";
  }
  if ($y) {
    if ($ret ne '' && $y > 0) {
      $ret .= '+';
    }
    if ($y == 1) {
      $ret .= 'I';
    } elsif ($y == -1) {
      $ret .= '-I';
    } else {
      $ret .= "$y*I";
    }
  }
  return $ret;
}

sub xy_to_node_name {
  my ($x,$y) = @_;
  if ($x==int($x) && $y==int($y)) {
    if ($x < 0) { $x = '-n'.-$x; }
    if ($y < 0) { $y = 'n'.-$y; }
    return "p$x-$y";
  } else {
    return "$x,$y";
  }
}

#------------------------------------------------------------------------------
# numbers

sub to_bigrat {
  my ($n) = @_;
  require Math::BigRat;
  return Math::BigRat->new($n);
}
sub to_bigint {
  my ($n) = @_;
  return bigint_class()->new($n);
}
# Usually better to let the application pre-load and decide globals like
# BigInt back-end, but this and to_bigint() are convenient.
use constant::defer bigint_class => sub {
  unless (Math::BigInt->VERSION) {
    require Math::BigInt;
    eval { Math::BigInt->import (try => 'GMP') };
  }
  return "Math::BigInt";
};

sub is_integer {
  my ($n) = @_;
  return $n == int($n);
}

# return true if $n is a perfect square 1, 4, 9, etc
sub is_square {
  my ($n) = @_;
  my $s = int(sqrt($n));
  my $ss = $s*$s;
  if ($ss > $n) {
    $s--;
    $ss = $s*$s;
  }
  return $ss == $n;
}

# return true if $n is a power of 2, so 1, 2, 4, 8, 16, etc
sub is_pow2 {
  my ($n) = @_;
  (($n >> 1) < $n) or die "is_pow2() infinite";

  while ($n > 1) {
    if ($n & 1) {
      return 0;
    }
    $n >>= 1;
  }
  return ($n == 1);
}

#------------------------------------------------------------------------------
# Complex i+1

# return $x,$y multiplied by 1+i
# GP-Test  (x+I*y)*(I-1) == (-x-y) + (x-y)*I
sub i_minus_1_mul {
  my ($x,$y) = @_;
  return (-$x-$y, $x-$y);
}
sub i_minus_1_power {
  my ($k) = @_;
  my ($x,$y) = (1,0);
  foreach (1 .. $k) { ($x,$y) = i_minus_1_mul($x,$y); }
  return ($x, $y);
}

# return $x,$y multiplied by 1+i
# GP-Test  (x+I*y)/(I-1) == (-x+y)/2 + (-x-y)/2*I
sub i_minus_1_div {
  my ($x,$y) = @_;
  return ((-$x+$y)/2, (-$x-$y)/2);
}

# (-1 + 3*I)/5 * (1 - (2*I)^floor(k/2));
# (-1 + 3*I)*(x+I*y) == -x -3*y   + I*(3*x - y)
sub PlusOffsetV_xy {
  my ($k) = @_;
  my $h = int($k/2);
  my ($x,$y) = (-2**$h, 0);
  ($x,$y) = xy_rotate_plus90($x,$y, $h);
  $x += 1;
  return ((-$x - 3*$y)/ 5,
          (3*$x - $y) / 5);
}

# (1 + 2*I)/5*(1 - (2*I)^ceil(k/2));
# (1 + 2*I)*(x+I*y) == x -2*y   + I*(2*x + y)
sub PlusOffsetH_xy {
  my ($k) = @_;
  my $h = ceil($k/2);
  my ($x,$y) = (-2**$h, 0);
  ($x,$y) = xy_rotate_plus90($x,$y, $h);
  $x += 1;
  ### $x
  ### $y
  return (($x - 2*$y) /5,
          (2*$x + $y) /5);
}

sub PlusOffsetJ_xy {
  my ($k) = @_;
  return ($k % 2 == 0 ? PlusOffsetH_xy($k) : PlusOffsetV_xy($k));
}

sub PlusToDragon {
  my ($x,$y) = @_;
  return i_minus_1_mul($x,$y);
}

sub NdragonToPlus {
  my ($n) = @_;
  my @bits = (0, bit_split_lowtohigh($n));  # 2*$n low to high
  ### NdragonToPlus: $n
  ### @bits
  my $flip = 0;
  foreach my $i (reverse 0 .. $#bits-1) {  # high to low
    if (! $bits[$i]        # 0 low
        && $bits[$i+1]) {  # 1 high
      $flip ^= 1;
    }
    $bits[$i+1] ^= $flip;
  }
  $bits[0] ^= $flip;
  ### @bits
  return digit_join_lowtohigh(\@bits,2);
}

#------------------------------------------------------------------------------
# TeX

sub xy_to_tex_str {
  my ($x,$y) = @_;
  my $str = "$x+${y}i";
  $str =~ s/\+-/-/;
  $str =~ s/\+1i/+i/;
  $str =~ s/\-1i/-i/;
  $str =~ s/\+0i//;
  $str =~ s/0\+//;
  $str =~ s/0\-/-/;
  $str =~ s/(.)([+-])/$1\{$2}/;
  return $str;
}

#------------------------------------------------------------------------------
# GP

sub xy_to_gp_str {
  my ($x,$y) = @_;
  my $str = '';
  if ($x || !$y) { $str .= $x; }
  if ($y) {
    my $join = ($y > 0 && length($str) ? '+' : '');
    my $num = $y==1 ? '' : $y==-1 ? '-' : "$y*";
    $str .= $join . $num . 'I';
  }
  return $str;
}


#------------------------------------------------------------------------------
# Graph.pm
# cf Graph::Convert

# return a Graph.pm graph of dragon curve level $k
sub dragon_Graph {
  my ($k) = @_;
  my $path = DragonCurve_path();
  require Graph::Undirected;
  my $graph = Graph::Undirected->new;
  my ($n_lo, $n_hi) = $path->level_to_n_range($k);
  my ($prev_x,$prev_y) = $path->n_to_xy($n_lo);
  $graph->add_vertex(Graph_xy_to_vertex($prev_x,$prev_y));
  foreach my $n ($n_lo+1 .. $n_hi) {
    my ($x,$y) = $path->n_to_xy($n);
    $graph->add_edge(Graph_xy_to_vertex($prev_x,$prev_y),
                     Graph_xy_to_vertex($x,$y));
    ($prev_x,$prev_y) = ($x,$y);
  }
  return $graph;
}

# return a Graph.pm graph of blob level $k
sub blob_Graph {
  my ($k) = @_;
  my $path = DragonCurve_path();
  require Graph::Undirected;
  my $graph = Graph::Undirected->new;
  if ($k < 4) {
    $graph->add_vertex("0,0");
  } else {
    my $n_lo = BlobN($k);
    my $n_hi = BlobN($k+1) - 3;
    my ($prev_x,$prev_y) = $path->n_to_xy($n_lo);
    $graph->add_vertex(Graph_xy_to_vertex($prev_x,$prev_y));
    foreach my $n ($n_lo+1 .. $n_hi) {
      my ($x,$y) = $path->n_to_xy($n);
      $graph->add_edge(Graph_xy_to_vertex($prev_x,$prev_y),
                       Graph_xy_to_vertex($x,$y));
      ($prev_x,$prev_y) = ($x,$y);
    }
  }
  return $graph;
}

# return a Graph.pm graph of twindragon curve level $k
sub twindragon_Graph {
  my ($k) = @_;
  my $path = DragonCurve_path();
  require Graph::Undirected;
  my $graph = Graph::Undirected->new;
  my ($n_lo, $n_hi) = $path->level_to_n_range($k+1);
  my ($x_hi,$y_hi) = $path->n_to_xy($n_hi);
  my ($prev_x,$prev_y) = $path->n_to_xy($n_lo);
  foreach my $n ($n_lo+1 .. $n_hi) {
    my ($x,$y) = $path->n_to_xy($n);
    $graph->add_edge(Graph_xy_to_vertex($prev_x,$prev_y),
                     Graph_xy_to_vertex($x,$y));
    $graph->add_edge(Graph_xy_to_vertex($x_hi-$prev_x,$y_hi-$prev_y),
                     Graph_xy_to_vertex($x_hi-$x,$y_hi-$y));
    ($prev_x,$prev_y) = ($x,$y);
  }
  return $graph;
}

sub Graph_xy_to_vertex {
  my ($x,$y) = @_;
  return sprintf '%d%+d', $x,$y;
}
sub Graph_vertex_to_xy {
  my ($v) = @_;
  $v =~ /(-?[0-9]+)\+?([-0-9]+)/ or die "bad vertex $v";
  return ($1,$2);
}

# $graph is a Graph.pm object, show it graphically
sub Graph_view {
  my ($graph) = @_;
  # print "$graph";

  require File::Temp;
  my $dot = File::Temp->new (UNLINK => 0, SUFFIX => '.dot');
  my $dot_filename = $dot->filename;

  require Graph::Writer::Dot;
  my $writer = Graph::Writer::Dot->new;
  $writer->write_graph($graph, $dot_filename);
  #  system("cat $dot_filename");

  require File::Temp;
  my $ps = File::Temp->new (UNLINK => 0, SUFFIX => '.ps');
  my $ps_filename = $ps->filename;

  require IPC::Run;
  IPC::Run::run(['dot','-Tps'], '<',$dot_filename, '>',$ps_filename);
  push @ipc, IPC::Run::start(['gv','--scale=1',$ps_filename],'&');
}

# $graph is a Graph.pm tree
# Return a hashref which is $v=>[$x,$y] where $x,$y is the location
# to place vertex $v.
# Vertices on the diameter are a horizontal line.
# Sub-trees from the diameter go upwards and outwards.
sub Graph_layout_diameterline {
  my ($graph) = @_;

  my $layout_subtree;
  $layout_subtree = sub {
    my ($v1, $v2) = @_;
    ### subtree(): "$v1  $v2"

    my @children = sort grep {$_ ne $v1} $graph->neighbours($v2);
    @children <= 2 or die;
    if (! @children) {
      return (0, { $v2 => [0,0] });
    }
    ### @children
    my ($depth, $l1) = $layout_subtree->($v2, shift @children);
    my $l2;
    if (@children) {
      (my $depth2, $l2) = $layout_subtree->($v2, $children[0]);
      ### $depth
      ### $depth2
      if ($depth2 > $depth) {
        $depth = $depth2;
        ($l1, $l2) = ($l2, $l1);  # to $l1 as the deepest
      }
      foreach my $p (values %$l2) {
        $p->[0]++;
        @$p = reverse @$p;  # transpose
      }
    } else {
      $l2 = {};
    }
    foreach my $p (values %$l1) {
      $p->[0]++;
    }
    ### right shift to: $l1
    ### and l2 upwards: $l2
    return ($depth+1, { $v2 => [0, 0], %$l1, %$l2 });
  };

  require MyGraphs;
  my ($eccentricty, @centre) = MyGraphs::Graph_tree_centre_vertices($graph);
  @centre = sort @centre;
  if (@centre <= 1) {
    return { map{$_=>[0,0]} @centre };
  }
  my ($v1,$v2) = @centre;
  my ($d1, $l1) = $layout_subtree->($v1, $v2);
  my ($d2, $l2) = $layout_subtree->($v2, $v1);

  #  v1   v2       v1     v2
  # 0,0   1,0     -1,0   0,0
  #
  foreach my $p (values %$l2) {
    $p->[0] = -$p->[0] - 1;  # -X - 1
  }
  ### $l1
  ### $l2
  return { %$l1, %$l2 };
}

#------------------------------------------------------------------------------
# Graph::Easy

# $graph is a Graph::Easy object, show it graphically
sub Graph_Easy_show {
  my ($graph) = @_;
  # print "$graph";

  require File::Temp;
  my $dot = File::Temp->new (UNLINK => 0, SUFFIX => '.dot');
  my $dot_filename = $dot->filename;

  print $dot $graph->as_graphviz;

  require File::Temp;
  my $ps = File::Temp->new (UNLINK => 0, SUFFIX => '.ps');
  my $ps_filename = $ps->filename;

  require IPC::Run;
  IPC::Run::run(['dot','-Tps'], '<',$dot_filename, '>',$ps_filename);
  push @ipc, IPC::Run::start(['gv','--scale=1',$ps_filename],'&');
}

#------------------------------------------------------------------------------
# Graph::Hashed

{
  # Subset of Graph.pm with hashes for edges.
  # Undirected.
  # $self->{'neighbours'} is a hashref of $from => { $to => 1, ... }
  # where the value at $from is a further hashref of which $to vertices
  # are neighbours of $from.
  # The graph is undirected so there is always entries $from -> $to and
  # back $to -> $from.

  package Graph::Hashed;
  use constant is_directed => 0;
  sub new {
    my $class =  shift;
    return bless { @_ }, $class;
  }
  sub vertices {
    my ($self) = @_;
    ### neighbours: $self->{'neighbours'}
    return keys %{$self->{'neighbours'}};
  }
  sub add_vertex {
    my ($self, $v) = @_;
    $self->{'neighbours'}->{$v} ||= {};
  }
  sub add_edge {
    my ($self, $v1,$v2) = @_;
    ### add_edge: "$v1 to $v2"
    $self->{'neighbours'}->{$v1}->{$v2} = undef;
    $self->{'neighbours'}->{$v2}->{$v1} = undef;
  }
  sub has_edge {
    my ($self, $v1,$v2) = @_;
    return exists $self->{'neighbours'}->{$v1}->{$v2};
  }
  sub neighbours {
    my ($self, $v) = @_;
    return keys %{$self->{'neighbours'}->{$v}};
  }
  sub centre_vertices {
    my ($self) = @_;
    my ($eccentricty, @centre) = MyGraphs::Graph_tree_centre_vertices($self);
    return @centre;
  }
  sub set_graph_attribute {
    my ($self) = @_;
  }
  sub vertex_eccentricity {
    return 'not implemented';
  }
  sub edges {
    my ($self) = @_;
    my $neighbours = $self->{'neighbours'};
    return map { my $v = $_;
                 map {($v cmp $_) >= 0 ? [ $v, $_ ] : ()
                    } keys %{$neighbours->{$v}}
                  } keys %$neighbours;
  }
  sub degree {
    my ($self, $v) = @_;
    return scalar(keys %{$self->{'neighbours'}->{$v}});
  }
}

#------------------------------------------------------------------------------
# GraphViz2.pm

sub twindragon_GraphViz2 {
  my ($k) = @_;
  my $path = DragonCurve_path();
  require GraphViz2;
  my $graphviz2 = GraphViz2->new (directed => 0);
  my ($n_lo, $n_hi) = $path->level_to_n_range($k+1);
  my ($x_hi,$y_hi) = $path->n_to_xy($n_hi);
  foreach my $n ($n_lo .. $n_hi-1) {
    my ($fx,$fy) = $path->n_to_xy($n);
    my ($tx,$ty) = $path->n_to_xy($n+1);
    foreach my $rev (0, 1) {
      if ($rev) {
        ($fx,$fy) = ($x_hi-$fx, $y_hi-$fy);
        ($tx,$ty) = ($x_hi-$tx, $y_hi-$ty);
      }
      my $from = Graph_xy_to_vertex($fx,$fy);
      my $to   = Graph_xy_to_vertex($tx,$ty);
      $graphviz2->add_node(name  => $from,
                           label => '',
                           pin   => 1,
                           pos   => "$fx,$fy",
                           style => 'filled',
                           fixedsize => 1,
                           width  => '.15',
                           height  => '.15',
                           fillcolor => 'black',
                          );
      $graphviz2->add_edge(from => $from,
                           to   => $to);
    }
  }
  return $graphviz2;
}

#------------------------------------------------------------------------------
# Twindragon Area Tree by Bits

sub TAdiameter {
  my ($k) = @_;
  return LQ($k+1) - 1;
}

sub TAVertexToLower {
  my ($n) = @_;
  my $mask = $n ^ ($n+1);
  return ($n & ($mask+1) ? $n - $mask : undef);
}
sub TAVertexToUpper {
  my ($n) = @_;
  if ($n == 0) { return undef; }
  my $mask = $n ^ ($n-1);
  return ($n & ($mask+1) ? undef : $n + $mask);
}
sub TAVertexToRight {
  my ($n) = @_;
  return ($n & 1 ? undef : $n ^ 1);
}
sub TAVertexToLeft {
  my ($n) = @_;
  return ($n & 1 ? $n ^ 1 : undef);
}

sub TAdegreeInf {
  my ($n) = @_;
  return ($n==0 ? 1
          : 2 + BitAboveLowestZero($n) - BitAboveLowestOne($n));
}

sub WidthS_infinity_k {
  my ($d) = @_;
  my $k = 1;
  for(;;) {
    if ($d - RQ($k-1) >=0) { $k++ }
    else { return $k-1; }
  }
}

#------------------------------------------------------------------------------
# twindragon right turns tree

sub twindragon_turn_tree_Graph {
  my ($k) = @_;
  my $path = DragonCurve_path();
  my $graph = Graph->new(undirected => 1);
  $graph->set_graph_attribute (name => "Twindragon Turn Tree $k");
  $graph->set_graph_attribute('xy',1);
  my $n_lo = 7*2**$k;
  my $n_hi = 11*2**$k;
  my $prev;
  foreach my $n ($n_lo .. $n_hi-1) {
    my $turn = ($n==$n_lo ? 1 : turn($n));
    if ($turn == -1) {
      my ($x,$y) = $path->n_to_xy($n);
      my $v = "$x,$y";
      ### $v
      $graph->add_vertex($v);
      if (defined $prev) { $graph->add_edge($v,$prev); }
      $prev = $v;
    } else {
      undef $prev;
    }
  }
  return $graph;
}


#------------------------------------------------------------------------------
# Rpred

our @Rpred_state_table = ([0,1],   # 0    1e          0-based
                          [2,3],   # 1    1o
                          [0,4],   # 2    2eA
                          [5,6],   # 3    2eB
                          [7,8],   # 4    3oA
                          [9,4],   # 5    3oB
                          [5,6],   # 6    3oC
                          [2,10],  # 7    3eA
                          [10,10], # 8    3eB
                          [10,3],  # 9    3eC
                          [10,10], # 10   non
                         );
our @Rpred_state_to_sides = (1,1, 2,2, 3,3,3, 3,3,3, 4);
our @Rpred_state_to_name = ('1e', '1o',
                            '2eA', '2eB',
                            '3oA', '3oB', '3oC',
                            '3eA', '3eB', '3eC',
                            'non');
our %Rpred_name_to_state
  = (map {$Rpred_state_to_name[$_] => $_} 0 .. $#Rpred_state_table);

use constant::defer NonRpred_flat => sub {
  require MyFLAT;
  MyFLAT::aref_to_FLAT_DFA(\@Rpred_state_table,
                           name => 'NonRpred');
};
use constant::defer Rpred_flat => sub {
  return NonRpred_flat()->complement
    ->MyFLAT::set_name('Rpred');
};

# $aref is an arrayref of bits of N
# return true if N is a right boundary segment
sub Rpred_of_bits {
  my ($aref) = @_;
  return state_table_transitions(\@Rpred_state_table, 0, $aref)
    != $Rpred_name_to_state{'non'};
}
# return true if segment $n is on the right boundary
sub Rpred {
  my ($n) = @_;
  return Rpred_of_bits([reverse bit_split_lowtohigh($n)]);
}

# 110001
our $NonRpred_re = qr/^
                      [01]* 1 ((100*|011*) 0)*
                      ( 10* 000 (100)* ( | 1 | 10 | 11)
                      | 01* 101 (001)* ( | 0 | 1 | 00)
                      | 01* 110 (010)* ( | 0 | 01 | 011)
                      | 01* 111
                      )
                     $/x;

$NonRpred_re = qr/^[01]* 1 ((10+|01+) 0)*
                   ( 10+        (001)* 00 (1 [01]?)?
                   | 01+ 0      (100)* 1 (00?|1)?
                   | 01+ 1 (1 | (001)* 0(0(11?)?)?)
                   )
                   $/x;

our $NonRpred_LtoH_re
  = qr/^ [01]*? (00 0+1 | (01|10|11)1+0 )  (0 (0+1|1+0))*  1 [01]* $/x;

# Return a FLAT::Regex matching N not on the right boundary.
# The main FLAT::DFA Rpred_flat() is simpler, but the regex can combine with
# other regex.
our $NonRpred_flatregex_str
 = '(0|1)*
    1 ((100*|011*) 0)*
    ( 10* 000 (100)* ([] | 1 | 10 | 11)
    | 01* 101 (001)* ([] | 0 | 1 | 00)
    | 01* 110 (010)* ([] | 0 | 01 | 011)
    | 01* 111
    )';

use constant::defer NonRpred_flatregex => sub {
  require FLAT;
  my $f = FLAT::Regex->new($NonRpred_flatregex_str);
  # $f->{'name'} = 'Rpred regex';   # FLAT::Regex is arrayref, no room for name
  return $f;
};


our @Rpred_LtoH_state_table = ([1,2],    # rm0
                               [3,4],    # rm1
                               [5,4],    # rm2
                               [6,4],    # rm3
                               [5,7],    # rm4
                               [3,7],    # rm5
                               [6,8],    # rm6
                               [8,7],    # rm7
                               [9,10],   # rm8
                               [6,7],    # rm9
                               [10,10],  # non
                             );
use constant::defer NonRpred_LtoH_flat => sub {
  require MyFLAT;
  MyFLAT::aref_to_FLAT_DFA(\@Rpred_LtoH_state_table, name => 'NonRpred LtoH');
};
# DFA matching bits of right boundary N, taken low to high.
use constant::defer Rpred_LtoH_flat => sub {
  return NonRpred_LtoH_flat()->complement
    ->MyFLAT::set_name('Rpred LtoH');
};


# $n is a segment number on the right boundary.
# Return the next bigger such segment number (for the curve continued
# infinitely).
sub Rnext {
  my ($n) = @_;
  if (turn($n+1) == 1) {
    my @n_list = DragonCurve_path()->n_to_n_list($n+1);
    return max(@n_list);
  } else {
    return $n+1;
  }
}

#------------------------------------------------------------------------------
# Rsides

# Return a FLAT::DFA which matches those N with a unit square of $s many
# sides on the right.  Must have $s = 1, 2, 3 or 4.
sub Rsides_flat {
  my ($s) = @_;
  require MyFLAT;
  MyFLAT::aref_to_FLAT_DFA
      (\@Rpred_state_table,
       accepting_list => [ grep {$Rpred_state_to_sides[$_] == $s}
                           0 .. $#Rpred_state_to_sides ],
       name => "Rsides $s");
}
BEGIN { memoize('Rsides_flat', LIST_CACHE=>'MERGE'); }

# $aref is an arrayref of bits of N, high to low.
# Return true if N is a right boundary segment.
sub Rsides_of_bits {
  my ($aref) = @_;
  return $Rpred_state_to_sides[state_table_transitions(\@Rpred_state_table, 0, $aref)];
}

# Return the number of sides on the unit square on the right of $n.
sub Rsides {
  my ($n) = @_;
  return Rsides_of_bits([reverse bit_split_lowtohigh($n)]);
}


#------------------------------------------------------------------------------
# Lpred

# left boundary states -- high to low
#
# s0  *----->*      =>   *    *      s4,s3
#                         \  /
#                          vv
#
#                          ^^
#                         /  \       s0,s1
# s1  *<-----*      =>   *    *
#
#                         \
#     ^                    ^
#     |             =>    /
# s2  *----->*           *    *      s6,s8
#                         \  /
#                          vv
#
#
#                        /
#     |                 ^   ^^
#     v             =>   \ /  \
# s3  *<-----*            *    *     s2,s1
#
#                             /
#            |               ^
#            v                \
# s4  *----->*      =>    *    *     s4,s7
#                          \  /
#                           vv
#
#                               \
#            ^              ^^   ^                unreached
#            |      =>     /  \ /
# s5  *<-----*            *    *      s0,s9
#
#
#     ^      |             \  /
#     |      v      =>     /  \       non,non
# s6  *----->*            *    *
#                          \  /
#                           vv
#
#                        /      \
#     |      ^          ^   ^^   ^    s2,s8
#     v      |      =>   \ /  \ /
# s7  *<-----*            *    *
#
#                         *    *
#     *------>           / \  /        non,s3
#     |                 ^   ^^
#     v             =>   \ /  \
# s8  *<-----*            *    *
#
#                         *    *
#     *------>             \  / \
#            ^              ^^   ^     s4,non
#            |      =>     /  \ /
# s9  *<-----*            *    *
#
our $Lpred_state_aref = [ [4,3],   # s0
                          [0,1],
                          [6,8],   # s2
                          [2,1],
                          [4,7],   # s4
                          [0,9],
                          [10,10], # s6
                          [2,9],
                          [10,3],  # s8
                          [4,10],
                          [10,10], # s10
                        ];
use constant::defer NonLpred_k_flat => sub {
  require MyFLAT;
  MyFLAT::aref_to_FLAT_DFA($Lpred_state_aref, name => 'NonLpred');
};

# Return a FLAT::DFA matching bits of left boundary N.
# When $k many bits (high to low) are put through it accepts or not
# according as the segment is on the left boundary within curve $k.
use constant::defer Lpred_k_flat => sub {
  return NonLpred_k_flat()->complement
    ->MyFLAT::set_name('Lpred');
};

# $aref is an arrayref of bits of N, high to low.
# Return true if N is a left boundary segment.
sub Lpred_k_of_bits {
  my ($aref) = @_;
  my $state = 0;
  foreach my $bit (@$aref) {  # high to low
    $state = $Lpred_state_aref->[$state]->[$bit];
  }
  return $state != 10;
}

# Return a FLAT::DFA which matches left boundary segment numbers of the
# curve continued infinitely.
use constant::defer NonLpredInf_flat => sub {
  require MyFLAT;
  MyFLAT::aref_to_FLAT_DFA($Lpred_state_aref,
                           starting => 4,
                           name => 'NonLpredInf');
};
# DFA matching bits of left boundary N of curve continued infinitely.
use constant::defer LpredInf_flat => sub {
  require MyFLAT;
  MyFLAT::aref_to_FLAT_DFA($Lpred_state_aref,
                           starting => 4,
                           accepting_list => [0 .. $#$Lpred_state_aref - 1],
                           name => 'LpredInf');
};

our $Lpred_LtoH_state_aref = [ [1,2],   # m0
                               [3,5],
                               [3,4],   # m2
                               [7,5],
                               [3,6],   # m4
                               [7,4],
                               [8,6],   # m6
                               [7,8],   # m7
                               [10,9],  # m8
                               [7,6],
                               [10,10], # non
                             ];
use constant::defer NonLpred_k_LtoH_flat => sub {
  require MyFLAT;
  MyFLAT::aref_to_FLAT_DFA($Lpred_LtoH_state_aref,
                           accepting_list => [8,10],
                           name => 'NonLpred LtoH');
};
use constant::defer Lpred_k_LtoH_flat => sub {
  return NonLpred_k_LtoH_flat()->complement
    ->MyFLAT::set_name('Lpred LtoH');
};

sub Lpred_k {
  my ($k,$n) = @_;
  my @bits = reverse bit_split_lowtohigh($n);
  if (@bits > $k) {
    croak "Lpred_k() n too big for k";
  }
  return Lpred_k_of_bits([((0)x($k-scalar(@bits))), @bits]);
}
sub Lnext_k {
  my ($k,$n) = @_;
  my @n_list;
  if (turn($n+1) == -1) {
    @n_list = DragonCurve_path()->n_to_n_list($n+1);
  } else {
    @n_list = ($n+1);
  }
  my $limit = 1 << $k;
  @n_list = grep {$_<$limit} @n_list;
  return (@n_list ? max(@n_list) : return undef);
}


#------------------------------------------------------------------------------
# Bpred

# Return a FLAT::DFA which matches boundary segment numbers.
# A string of $k many bits (high to low) is accepted if that N is a boundary
# segment (left or right or both).
use constant::defer Bpred_k_flat => sub {
  # union() gives an NFA, want to go to DFA
  return Rpred_flat()->union(Lpred_k_flat())->as_dfa
    ->MyFLAT::set_name('Bpred');
};

# Return a FLAT::DFA which matches non-boundary segment numbers.
# A string of $k many bits (high to low) is accepted if that N is a
# non-boundary segment (neither left nor right boundary).
use constant::defer NonBpred_k_flat => sub {
  # both NonR and NonL, so the intersection of those sets
  return NonRpred_flat()->intersect(NonLpred_k_flat())
    ->MyFLAT::set_name('NonBpred');
};

# Return a FLAT::DFA which matches boundary segment numbers of the curve
# continued infinitely.
use constant::defer BpredInf_flat => sub {
  return Rpred_flat()->union(LpredInf_flat())->as_dfa
    ->MyFLAT::set_name('BpredInf');
};

#------------------------------------------------------------------------------
# other preds

# curve start and all bridges in curve continued infinitely
use constant::defer NonBlobPredInf_flat => sub {
  return Rpred_flat()->intersect(LpredInf_flat())
    ->MyFLAT::set_name('NonBlobPred');
};


#------------------------------------------------------------------------------
# 5ths

# fFifthsPred_flat() returns a FLAT::DFA matching fractional bits of f which
# is both left and right boundary.
# Such an f is 0, 1, (2/5)/2^k, 1-(2/5)/2^k.
# These f are infinite length strings of bits.  The DFA remains accepting as
# long as the bits put through it are the start of such a 5th string.  When
# it sees bits not the start of such a string it is non-accepting, and
# remains so for any further bits.
#
# 2/5 = 6/15 is 0110 repeating, and leading 0s for (2/5)/2^k
# 3/5 = 9/15 is 1001 repeating, and leading 1s for 1-(2/5)/2^k
#
# An empty string or any number of 0s is accepted, since it is either f=0 or
# the start of some (2/5)/2^k for a big k.
#
# An empty string or any number of leading 1s is accepted, since it is
# either f=1 in the form .111... or the start of some 1-(2/5)/2^k for big k.
#
use constant::defer fFifthsPred_flat => sub {
  require FLAT;
  return FLAT::Regex->new ("0* (0110)* ([] | 0 | 01 | 011)
                   | 1* (1001)* ([] | 1 | 10 | 100)")->as_dfa
    ->MyFLAT::set_name('fFifthsPred');
};


#------------------------------------------------------------------------------
# Dpred

# FLAT::Regex patterns for the runs part of Dpred, either left or right turn.
# The bit-above-lowest-1, lowest 1, and possible 0s below are not included.
#
our $DpredLeft_runs_fre
  = "      (0|1)* 1  (100* | 011*) (0 (100* | 011*))*";  # with t=0
our $DpredRight_runs_fre
  = "([] | (0|1)* 0) (100* | 011*) (1 (100* | 011*))*";  # with t=1

use constant::defer DpredLeft_runs_flat => sub {
  require FLAT;
  require MyFLAT;
  return FLAT::Regex->new($DpredLeft_runs_fre )
    ->as_dfa->MyFLAT::set_name('DpredLeft_runs');
};
use constant::defer DpredRight_runs_flat => sub {
  require FLAT;
  require MyFLAT;
  return FLAT::Regex->new($DpredRight_runs_fre )
    ->as_dfa->MyFLAT::set_name('DpredRight_runs');
};
use constant::defer SpredLeft_runs_flat => sub {
  return DpredLeft_runs_flat()->complement
    ->MyFLAT::set_name('SpredLeft_runs');
};
use constant::defer SpredRight_runs_flat => sub {
  return DpredRight_runs_flat()->complement
    ->MyFLAT::set_name('SpredRight_runs');
};

# Dpred_k_flat() returns a FLAT::DFA which matches a string of k many bits
# (with leading 0s) which is a double-visited point N within level k.
#
# DpredRight_k_flat() and DpredLeft_k_flat() similarly, but restricted to
# those N which are a right or left turn respectively.  They are the
# intersection of Dpred_k_flat() with TurnRpred_flat() or TurnLpred_flat()
# respectively.
#
# $Dpred_k_flat_fre etc are FLAT::Regex pattern strings for the same.
#
our $DpredLeft_k_flat_fre  = "$DpredLeft_runs_fre   01 0*";
our $DpredRight_k_flat_fre = "$DpredRight_runs_fre  11 0*";
our $Dpred_k_flat_fre      = "( $DpredRight_k_flat_fre | $DpredLeft_k_flat_fre )";
#
use constant::defer DpredLeft_k_flat => sub {
  require FLAT;
  require MyFLAT;
  return FLAT::Regex->new($DpredLeft_k_flat_fre)->as_dfa->as_min_dfa
    ->MyFLAT::set_name('DpredLeft k');
};
use constant::defer DpredRight_k_flat => sub {
  require FLAT;
  require MyFLAT;
  return FLAT::Regex->new($DpredRight_k_flat_fre)->as_dfa->as_min_dfa
    ->MyFLAT::set_name('DpredRight k');
};
use constant::defer Dpred_k_flat => sub {
  require FLAT;
  require MyFLAT;
  return FLAT::Regex->new($Dpred_k_flat_fre)->as_dfa->as_min_dfa
    ->MyFLAT::set_name('Dpred k');
};

# Spred_k_flat() returns a FLAT::DFA which matches a string of k many bits
# (with leading 0s) which is a single-visited point N within level k.  This
# includes N=0 (string 000...000 of any k many bits, including empty).
#
# SpredRight_k_flat() and SpredLeft_k_flat() similarly, but restricted to
# those N which are a right or left turn respectively.  They are the
# intersection of Spred_k_flat() with TurnRpred_flat() or TurnLpred_flat()
# respectively.  This means they do not match N=0, since there is no turn at
# N=0.
#
# Spred_k_flat() is the complement of Dpred_k_flat().
#
use constant::defer Spred_k_flat => sub {
  return Dpred_k_flat()->complement->as_min_dfa
    ->MyFLAT::set_name('Spred k');
};
use constant::defer SpredLeft_k_flat => sub {
  require MyFLAT;
  require FLAT;
  my $f = FLAT::Regex->new($DpredLeft_runs_fre)->as_dfa->complement
    ->MyFLAT::concat(FLAT::Regex->new('01 0*'));
  # N=2^k without high 0 when Dpred runs empty
  return $f->union(FLAT::Regex->new('1 0*'))->as_dfa->as_min_dfa
    ->MyFLAT::set_name('SpredLeft k');
};
use constant::defer SpredRight_k_flat => sub {
  require MyFLAT;
  require FLAT;
  return DpredRight_runs_flat()->as_dfa->complement
    ->MyFLAT::concat(FLAT::Regex->new('11 0*'))
    ->as_dfa->as_min_dfa
    ->MyFLAT::set_name('SpredRight k');
};

#------------------------------------------------------------------------------

our $DpredFirstL_fre
  = "       (0|1)* 1         011*  (0 (100* | 011*))*  01 0*";  # t=0
our $DpredFirstR_fre
  = "(((0|1)* 0 |[])   0 |[]) 11*  (1 (100* | 011*))*  11 0*";  # t=1

our $DpredSecondL_fre
  = " (0|1)*  1      (100*       ) (0 (100* | 011*))*  01 0*";  # t=0
our $DpredSecondR_fre
  = "((0|1)*  0 |[]) (100*       ) (1 (100* | 011*))*  11 0*";  # t=1

use constant::defer DpredFirstL_inf_flat => sub {
  require FLAT;
  require MyFLAT;
  return FLAT::Regex->new($DpredFirstL_fre )
    ->as_dfa
    ->MyFLAT::minimize
    ->MyFLAT::set_name('DpredFirstL inf');
};
use constant::defer DpredFirstR_inf_flat => sub {
  require FLAT;
  require MyFLAT;
  return FLAT::Regex->new($DpredFirstR_fre )
    ->as_dfa
    ->MyFLAT::minimize
    ->MyFLAT::set_name('DpredFirstR inf');
};
use constant::defer DpredFirst_inf_flat => sub {
  require FLAT;
  require MyFLAT;
  return    DpredFirstL_inf_flat()
    ->union(DpredFirstR_inf_flat())
    ->MyFLAT::minimize
    ->MyFLAT::set_name("DpredFirst inf");
};

use constant::defer DpredSecondL_flat => sub {
  require FLAT;
  require MyFLAT;
  return FLAT::Regex->new($DpredSecondL_fre )->as_dfa
    ->MyFLAT::set_name('DpredSecondL');
};
use constant::defer DpredSecondR_flat => sub {
  require FLAT;
  require MyFLAT;
  return FLAT::Regex->new($DpredSecondR_fre )->as_dfa
    ->MyFLAT::set_name('DpredSecondR');
};
use constant::defer DpredSecond_flat => sub {
  require FLAT;
  require MyFLAT;
  return    DpredSecondR_flat()
    ->union(DpredSecondL_flat())->MyFLAT::minimize
    ->MyFLAT::set_name("DpredSecond");
};

sub DpredSecond {
  my ($n) = @_;
  my @n_list = DragonCurve_path()->n_to_n_list($n);
  return ($#n_list>=1 && $n_list[1] == $n);
}

use constant::defer Dpred_inf_flat => sub {
  return DpredFirst_inf_flat()
    ->union(DpredSecond_flat())
    ->MyFLAT::minimize
    ->MyFLAT::set_name("Dpred inf");
};


#------------------------------------------------------------------------------
# TurnLpred, TurnRpred

# FLAT::DFA matching bits of N which are a left turn.
# This is 0 above the lowest 1 bit.
# Leading 0 bits are allowed, and are understood so that "1" alone is accepted
# as having a 0 above that 1.
# N=0 is not matched, neither as empty string nor as 0...
use constant::defer TurnLpred_flat => sub {
  require FLAT;
  require MyFLAT;
  return FLAT::Regex->new("((0|1)* 0 | []) 1 0*")->as_dfa
    ->MyFLAT::set_name("TurnLpred");
};

# FLAT::DFA matching bits of N which are a right turn.
# This is 1 above the lowest 1 bit.
# Leading 0 bits are allowed.
# N=0 is not matched, neither as empty string nor as 0...
use constant::defer TurnRpred_flat => sub {
  require FLAT;
  require MyFLAT;
  return FLAT::Regex->new(" (0|1)* 1       1 0*")->as_dfa
    ->MyFLAT::set_name("TurnRpred");
};


#------------------------------------------------------------------------------

{
  # Segments are represented by strings "X,Y,DIR" where X,Y are integer
  # coordinates with X+Y even, and DIR = 0 to 3.  DIR is direction for
  # expansion, so on expansion get new segments on the right always.
  #
  # "side" = 0 for a forward segment, so that segment 0,0 dir=0 is even.
  #
  # "side" = 0 for a reverse segment, so that segment 0,0,dir=0 has 0,0
  # towards the end of the curve and 1,0 towards the start.  Consequently
  # the right side of 0,0,dir=0 is the left of the curve, and vice versa.

  package MyDragonConf;
  use Carp 'croak';
  use List::Util 'min','max';
  use List::MoreUtils;

  sub new {
    my $class = shift;   # ($class, segs=>$s, side=>$side)
    if (ref $class) { $class = ref $class; }
    my $self = bless { @_ }, $class;
    $self->{'segs'} //= [];
    $self->{'segshash'} = { map {$_=>1} $self->segs_list };

    my @next_segs = @{$self->{'next_segs'} // []};
    my $want_next_segs = ($self->{'want_next_segs'} ||= 0);
    if (@next_segs > $want_next_segs) {
      $#next_segs = $want_next_segs-1;  # truncate
    }
    $self->{'next_segs'} = \@next_segs;
    $self->validate;
    ### $self
    return $self;
  }

  sub seg_validate {
    my ($self, $seg) = @_;
    unless ($seg =~ /^(-?[1-9][0-9]*|0),(-?[1-9][0-9]*|0),[0123]$/) {
      croak "bad seg ",$seg;
    }
    # x+y must be even
    my ($x,$y,$dir) = split /,/, $seg;
    if (($x+$y) & 1) {
      croak "bad seg location ",$seg;
    }
  }

  {
    my %valid_next_seg = (0 => {'1,1,3' => 1,  '1,-1,1' => 1},
                          1 => {'0,0,1' => 1,  '0,0,3' => 1});

    sub validate {
      my ($self) = @_;
      my $next_segs = $self->{'next_segs'};
      foreach my $seg ($self->segs_list,
                       (defined $next_segs ? @$next_segs : ()),
                      ) {
        $self->seg_validate($seg);
      }
      my $want_next_segs = $self->{'want_next_segs'};
      if ($want_next_segs && @$next_segs > $want_next_segs) {
        croak "next_segs too long";
      }
      if (defined $next_segs && @$next_segs >= 1) {
        $valid_next_seg{$self->{'side'}}->{$next_segs->[0]}
          or croak "bad next_seg $next_segs->[0] for side=$self->{side}";
      }
    }
  }

  # Return a string which is a key representing $self, being its side,
  # segments and next_segs.
  sub key {
    my ($self) = @_;
    my @next_segs = $self->next_segs;
    ### $self
    return ($self->{'key'}
            //= join(' ',
                     "side=$self->{side}",
                     (sort $self->segs_list),
                     ($self->{'want_next_segs'} && @next_segs
                      ? "next=".join(' ',@next_segs) : ())));
  }

  # MyDragonConf->initial(key => value)
  # Return a configuration for the initial k=0 single segment configuration.
  # Additional options are per new(), for example want_next_segs => N.
  sub initial {
    my $class = shift;
    return $class->new (side => 0, segs => ['0,0,0'], @_);
  }

  # Return true if $self equals $other, meaning same set of segments and
  # same side.  If want_next_segs then the list of next_segs() must be the
  # same too.
  sub equals {
    my ($self, $other) = @_;
    return $self->key eq $other->key;
  }

  # $seg is a segment as a string "x,y,dir".
  # Return true if $self contains $seg.
  sub contains {
    my ($self, $seg) = @_;
    return $self->{'segshash'}->{$seg};
  }

  # $self->is_subset($superset)
  # Return true if $self is a subset of the given $superset configuration.
  sub is_subset {
    my ($self, $superset) = @_;
    foreach my $seg ($self->segs_list) {
      if (! $superset->contains($seg)) { return 0; }
    }
    return 1;
  }

  # Return a list of the segments the configuration.
  # Each segment is a string "x,y,dir".
  sub segs_list {
    my ($self) = @_;
    return @{$self->{'segs'}};
  }

  # $conf->num_segs()
  # Return the number of segments in the configuration.
  sub num_segs {
    my ($self) = @_;
    return scalar @{$self->{'segs'}};
  }

  # Return a list of the segments next after the middle 0,0,0 in curve start
  # towards end order.  Middle 0,0,0 itself is not included.
  sub next_segs {
    my ($self) = @_;
    # $self->{'want_next_segs'}
    #   or croak "next_segs() unavailable without want_next_segs";
    return @{$self->{'next_segs'}};
  }

  sub transform_segs {
    my ($self, $transform_func, %options) = @_;
    ### transform_segs: %options
    my $name = $self->{'name'};
    if (defined (my $name_append = delete $options{'name_append'})
        && defined $name) {
      $name .= join(' ',$name,$name_append);
    }

    my $want_next_segs = $self->{'want_next_segs'} || 0;
    if (exists $options{'want_next_segs'}) {
      $want_next_segs = delete $options{'want_next_segs'};
    }
    ### $want_next_segs
    my @next_segs;
    if ($want_next_segs) {
      @next_segs = map {$transform_func->($_)} @{$self->{'next_segs'}};
    }

    ### @next_segs

    return $self = $self->new
      (name   => $name,
       side   => $self->{'side'},
       within => $self->{'within'},
       segs   => [ map {$transform_func->($_)} $self->segs_list ],
       want_next_segs => $want_next_segs,
       next_segs      => \@next_segs);
  }

  # Return a new MyDragonConf with each point $x,$y moved to $x+$dx, $y+$dy.
  sub translate {
    my ($self, $dx,$dy, @options) = @_;
    ### translate ...
    if ($dx==0 && $dy==0) { return $self; }
    return $self->transform_segs
      (sub {
         my ($seg) = @_;
         my ($x,$y,$dir) = split /,/, $seg;
         $x += $dx;
         $y += $dy;
         "$x,$y,$dir";
       },
       name_append => "moved $dx,$dy",
       want_next_segs => 0,
       @options);
  }

  # return a new MyDragonConf with each segment rotated by $ddir, so each
  # $dir becomes $dir+$ddir, mod 4
  sub rotate {
    my ($self, $ddir, @options) = @_;
    $ddir %= 4;
    if ($ddir == 0) { return $self; }
    return $self->transform_segs
      (sub {
         my ($seg) = @_;
         my ($x,$y,$dir) = split /,/, $seg;
         ($x,$y) = MyDragon::xy_rotate_plus90($x,$y, $ddir);
         $dir = ($dir + $ddir) % 4;
         "$x,$y,$dir";
       },
       name_append  => ("rotate ".($ddir==1   ? "+90"
                                   : $ddir==2 ? "180"
                                   : $ddir==3 ? "-90"
                                   : $ddir)),
       want_next_segs => 0,
       @options);
  }

  # return a new MyDragonConf with the segments common to all the given confs
  sub intersection {
    my $self = shift;
    foreach my $other (@_) {
      $self = $self->transform_segs
        (sub {
           my ($seg) = @_;
           return ($other->contains($seg) ? $seg : ());
         });
    }

    return $self;
  }

  # $conf->difference($conf2, $conf3, ...)
  # Return a new MyDragonConf which is $conf with all the segments of $conf2
  # and $conf3 etc removed.
  sub difference {
    my $self = shift;
    foreach my $other (@_) {
      $self =$self->new
        (side   => $self->{'side'},
         within => $self->{'within'},
         segs => [ grep {! $other->contains($_)} $self->segs_list ]);
    }
    return $self;
  }

  sub union {
    my $self = shift;
    my $name = $self->{'name'};
    if (defined $name) { $name = "$name union"; }
    return $self->new (name   => $name,
                       side   => $self->{'side'},
                       within => $self->{'within'},
                       segs   => [ List::MoreUtils::uniq
                                   (map {$_->segs_list} $self,@_) ]);

  }

  sub transform_to_seg {
    my ($self, $base) = @_;
    return $self->transform_segs
      (sub {
         my ($seg) = @_;
         return MyDragon::seg_transform_to($seg, $base);
       },
       name_append    => "transformed $base",
       want_next_segs => 0);
  }

  # $positions is a MyDragonConf whose segments represent positions.
  # 0,0,0 in $self is rotated and translated as necessary to each of $positions.
  # Return the union of each $self at each of $positions.
  sub union_in_positions {
    my ($self, $positions) = @_;
    return union(map {
      my ($x,$y,$dir) = split /,/;
      $self->translate(-$x,-$y)->rotate(-$dir)
    } $positions->segs_list);
  }

  # Return a new MyDragonConf which is ...
  sub next_positions {
    my ($class, $want_next_segs) = @_;
    my %segs = ('0,0,0' => 1);
    my %xy = ('0,0' => 1,
              '1,0' => 1);
    foreach (1 .. $want_next_segs) {
      my @keys = keys %xy;
      foreach my $xy (@keys) {
        my ($x,$y) = split /,/, $xy;
        foreach my $dir (0 .. 3) {
          $segs{MyDragon::seg_normalize("$x,$y,$dir")} = 1;
          my $x2 = $x + $MyDragon::dir4_to_dx[$dir];
          my $y2 = $y + $MyDragon::dir4_to_dy[$dir];
          $xy{"$x2,$y2"} = 1;
        }
      }
    }
    return $class->new (name => "next positions $want_next_segs",
                        side => 0,
                        segs => [keys %segs]);
  }

  sub set_name {
    my ($self, $name) = @_;
    $self->{'name'} = $name;
    return $self;
  }

  # Set a "within" which is another MyDragonConf.
  # ->expand() limits its output to intersection of "within".
  sub set_within {
    my ($self, $within) = @_;
    $self->{'within'} = $within;
    return $self;
  }

  {
    my %expand_next_seg
      = (0 => { 0 => {
                      #  side=0, bit=0
                      #    ----->
                      #    \   /
                      #     v v
                      '1,1,3'  => '1,1,3',
                      '1,-1,1' => '1,1,3',
                     },
                1 => {
                      #  side=0, bit=1
                      #        / | next       ----->
                      #       v  |            \   / ^
                      #       ^  |             v v  | \
                      #        \ |                  |  v    next
                      #          v                  |  ^
                      #    ----->                   | /
                      #    \   /
                      #     v v
                      '1,1,3'  => '0,0,3',
                      '1,-1,1' => '0,0,1',
                     },
              },
         1 => { 0 => {
                      #  side=1, bit=0
                      #  curve end  ----->  curve start
                      #             \   /bit=0
                      #              v v
                      '0,0,3' => '1,-1,1',  # always
                      '0,0,1' => '1,-1,1',
                     },
                1 => {
                      #     side=1, bit=1                curve end  -----> curve start
                      #           ^                               /| \  /
                      #     next  | \                      next  v |  v v
                      #           |  v                           ^ |
                      #           |  ^                            \|
                      #           | /                              v
                      # curve end  -----> curve start
                      #           1\   /
                      #             v v
                      '0,0,1' => '0,0,1',  # unchanged
                      '0,0,3' => '0,0,3',
                     },
              });

    # Return a new MyDragonConf which is segment expansion, with $bit as the
    # new 0--1 segment.
    sub expand {
      my ($self, $bit) = @_;
      ### expand() ...
      ($bit==0 || $bit==1) or die;
      if (! defined $self->{'side'}) {
        croak 'No side for conf expand';
      }

      # state 0 right side enc, state 1 left side enc
      #  0 ------> 1         1 ------> 0
      #   0\     /1           1\     /0
      # 0a  \   /  1b       1a  \   /  0b
      #     1v v0               0v v1
      # ($conf->{'side'} == $bit)   a  0,0 or 1,1
      # ($conf->{'side'} != $bit)   b  0,1 or 1,0a  0,0 or 1,1

      # for bit==side
      my @segs      = map {MyDragon::seg_expand($_)} $self->segs_list;
      my @next_segs = map {MyDragon::seg_expand($_)} '0,0,0', $self->next_segs;

      # for opposite
      if ($self->{'side'} != $bit) {
        @segs      = map { MyDragon::seg_other_expanded($_) } @segs;
        @next_segs = map { MyDragon::seg_other_expanded($_) } @next_segs;
      }

      ### @next_segs
      for (my $i = (1-$self->{'side'})<<1; $i < $#next_segs; $i += 4) {
        ($next_segs[$i],$next_segs[$i+1]) = ($next_segs[$i+1],$next_segs[$i]);
      }
      ### swapped: @next_segs
      splice @next_segs, 0, $bit+1;  # 1 if bit==0, 2 if bit==1

      my $expand = $self->new (side           => $bit,
                               segs           => \@segs,
                               want_next_segs => $self->{'want_next_segs'},
                               next_segs      => \@next_segs);
      if (defined (my $within = $self->{'within'})) {
        $expand = $expand->intersection($within);
        $expand->{'within'} = $within;
      }
      if (defined (my $name = $self->{'name'})) {
        unless ($name =~ /[0-9]$/ || $name eq '') { $name .= ' '; }
        $expand->{'name'} = $name . $bit;
      }
      return $expand;
    }
  }

  # Return true if the configuration has a boundary square side 0=right 1=left.
  #
  sub is_boundary {
    my ($self, $side) = @_;
    if (! defined $self->{'side'}) {
      croak 'No side for conf is_boundary';
    }
    # $self->print;
    my $within = $self->{'within'} // croak 'No within';
    $side ^= $self->{'side'};
  SEG: foreach my $seg ($self->segs_list) {
      ### $seg
      my ($x,$y,$dir) = split /,/, $seg;
      my $dd = (($side ^ $dir) & 1
                ?  1  # side=1 to left, dir+1
                : -1  # side=0 to right, dir-1
               );

      #               dir
      #         x,y  ----->
      #          |          ^
      #    dir1  |          | dir3
      #          v          |
      #            <----- x2,y2
      #              dir2
      my $dir1 = ($dir + $dd) % 4;
      my $dir2 = ($dir  + 2) % 4;
      my $dir3 = ($dir1 + 2) % 4;
      my $x2 = $x + $dir4_to_dx[$dir] + $dir4_to_dx[$dir1];
      my $y2 = $y + $dir4_to_dy[$dir] + $dir4_to_dy[$dir1];

      foreach my $other_seg ("$x,$y,$dir1",
                             "$x2,$y2,$dir2",
                             "$x2,$y2,$dir3") {
        ### $other_seg
        if (! $within->contains($other_seg)) {
          # presence of $other_seg is unspecified, this square cannot be
          # used to test for boundary
          ### not within ...
          next SEG;
        }
        if (! $self->contains($other_seg)) {
          ### absent, so boundary ...
          return 1;
        }
      }
    }
    return 0;
  }

  # bounding box around the segments
  sub bbox {
    my ($self) = @_;
    my $x_min=0;
    my $x_max=1;
    my $y_min=0;
    my $y_max=0;
    foreach my $seg ($self->segs_list) {
      my ($x,$y,$dir) = split /,/, $seg;
      foreach (0,1) {
        $x_min = min($x_min, $x); $x_max = max($x_max, $x);
        $y_min = min($y_min, $y); $y_max = max($y_max, $y);
        $x += $dir4_to_dx[$dir];
        $y += $dir4_to_dy[$dir];
      }
    }
    return ($x_min,$y_min, $x_max,$y_max);
  }
  sub print {
    my ($self) = @_;
    print $self->print_to_str, "\n";
  }
    sub print_to_str {
      my ($self) = @_;
      my $ret = "conf";
      if (defined $self->{'name'}) {
        $ret .= ' ' . $self->{'name'};
      }
      $ret .= " side=" . ($self->{'side'}//'undef') . "\n";

      # sort for consistency of output
      $ret .= "   " . join(' ',sort $self->segs_list) . "\n";

      my ($x_min,$y_min, $x_max,$y_max) = $self->bbox;
      ### $x_min
      ### $x_max
      ### $y_min
      ### $y_max

      foreach my $y (reverse $y_min .. $y_max) {
        my $yp1 = $y+1;
        if ($y<$y_max) {
          # above verticals
          $ret .= "   ";
          foreach my $x ($x_min .. $x_max) {
            if    ($self->contains("$x,$y,1"))   { $ret .= "^     "; }
            elsif ($self->contains("$x,$yp1,3")) { $ret .= "|     "; }
            else                                 { $ret .= "      "; }
          }
          $ret .= "\n";

          $ret .= "   ";
          foreach my $x ($x_min .. $x_max) {
            if    ($self->contains("$x,$y,1"))   { $ret .= "|     "; }
            elsif ($self->contains("$x,$yp1,3")) { $ret .= "|     "; }
            else                                 { $ret .= "      "; }
          }
          $ret .= "\n";

          $ret .= "   ";
          foreach my $x ($x_min .. $x_max) {
            if    ($self->contains("$x,$y,1"))   { $ret .= "|     "; }
            elsif ($self->contains("$x,$yp1,3")) { $ret .= "v     "; }
            else                                 { $ret .= "      "; }
          }
          $ret .= "\n";
        }

        # row
        $ret .= "   ";
        foreach my $x ($x_min .. $x_max) {
          my $xp1 = $x+1;
          $ret .= $x==0&&$y==0 ? "O" : "*";
          if ($self->contains("$x,$y,0"))      { $ret .= $x==0&&$y==0
                                                   ? "====>" : "---->"; }
          elsif ($self->contains("$xp1,$y,2")) { $ret .= "<----"; }
          else                                 { $ret .= "     "; }
        }
        $ret .= "\n";
      }
      $ret =~ s/ +\n/\n/g; # delete trailing spaces
      return $ret;
    }
  }

sub seg_transform_to {
  my ($seg, $base) = @_;
  my ($x,$y,$dir) = split /,/, $seg;
  my ($bx,$by,$bdir) = split /,/, $base;
  $x -= $bx;
  $y -= $by;
  ($x,$y) = xy_rotate_plus90($x,$y, - $bdir);
  $dir = ($dir - $bdir) % 4;
  return "$x,$y,$dir";
}

# $seg is a segment as a string "x,y,dir".
# Return two segments which is expansion on the right.
sub seg_expand {
  my ($seg) = @_;
  my ($x,$y,$dir) = split /,/, $seg;
  my $x2 = $x + $dir4_to_dx[$dir];
  my $y2 = $y + $dir4_to_dy[$dir];
  ($x,$y) = MyDragon::mul($x,$y);
  ($x2,$y2) = MyDragon::mul($x2,$y2);
  my $dir2 = ($dir-1) % 4;
  return ("$x,$y,$dir", "$x2,$y2,$dir2");
}

# $seg is a segment as a string "x,y,dir".
# Return it modified as for taking the second of an seg_expand().
# For example expand_set('0,0,0') is '0,0,0' and '1,1,3'.
# seg_other_expanded() rotates and moves so second '1,1,3' becomes '0,0,0'.
sub seg_other_expanded {
  my ($seg) = @_;
  my ($x,$y,$dir) = split /,/, $seg;
  $x-=1; $y-=1;
  ($x,$y) = xy_rotate_plus90($x,$y);
  $dir = ($dir+1) % 4;
  return "$x,$y,$dir";
}

# $seg is a segment as a string "x,y,dir".
# If x,y is even then return it unchanged.
# If x,y is odd then return x,y moved to the end and dir reversed.
# Even x,y is the normal form for a segment.
sub seg_normalize {
  my ($seg) = @_;
  ### seg_normalize(): $seg
  my ($x,$y,$dir) = split /,/, $seg;
  if (($x+$y) & 1) {
    $x += $dir4_to_dx[$dir];
    $y += $dir4_to_dy[$dir];
    $dir = ($dir+2) % 4;
    return "$x,$y,$dir";
  } else {
    return $seg;
  }
}

# $seg is a segment as a string "x,y,dir".  Must have x,y even.
# Return an arrayref of points like $hull6_points but moved and rotated
# according to $seg location and direction.
# $hull6_points has all coordinates multiplied by 6 so as to be integers.
# The same is in the return here.
sub seg_to_hull6_points {
  my ($seg) = @_;
  my ($x,$y,$dir) = split /,/, $seg;
  require MyPlanar;
  my $points = MyPlanar::points_rotate_plus90($hull6_points, $dir);
  return MyPlanar::points_move($points, 6*$x, 6*$y);
}

# conf before and after side=undef
#    0,0,0 0,0,1 0,0,3 1,1,3 1,-1,1
#    *     *
#    ^     |
#    |     |
#    |     v
#    O====>*
#    |     ^
#    |     |
#    v     |
#    *     *
our $conf_before_and_after
  = MyDragonConf->new(segs => [ '0,0,0', '0,0,1', '0,0,3',
                                '1,1,3', '1,-1,1' ])
  ->set_name('before and after');
# $conf_before_and_after->print;


#------------------------------------------------------------------------------

#    0,0 ======> 1,0
#     |           ^
#     |           |
#     v           |
#    0,-1 <----- 1,-1
#
our $R_square_segs = MyDragonConf->new
  (name => 'R square',
   side => 0,
   segs => [
            "0,0,0",       # below m
            "0,0,3",
            "1,-1,2",
            "1,-1,1",
           ]);

#    0,1 <------ 1,1
#     ^           |
#     |           |
#     |           v
#    0,0 ======> 1,0
#
our $L_square_segs = MyDragonConf->new
  (name => 'L square',
   side => 0,
   segs => [
            "0,0,0",      # above m
            "0,0,1",
            "1,1,2",
            "1,1,3",
           ]);

#    0,1 <------ 1,1
#     ^           |
#     |           |
#     |           v
#    0,0 ======> 1,0
#     |           ^
#     |           |
#     v           |
#    0,-1 <----- 1,-1
#
our $B_squares_segs = $R_square_segs->union($L_square_segs)
  ->set_name('B two squares');  # Bpred enclosing


#------------------------------------------------------------------------------
# 7ths Triples Segments - Sevenths

our @triple_7ths_segs;

#      0,1
#       ^
#    1  |
#      0,0 ====> 1,0
#          3 5
our $triple_7th_1_3_5_segs = MyDragonConf->new
  (name => '7th 1,3,5',
   side => 0,
   segs => [ '0,0,0',
             '0,0,1',
           ]);
$triple_7ths_segs[3] = $triple_7th_1_3_5_segs;
$triple_7ths_segs[5] = $triple_7th_1_3_5_segs;
$triple_7ths_segs[1] = $triple_7th_1_3_5_segs->rotate(-1);

#      0,1
#       ^         | 4
#    2  |         v
#      0,0 ====> 1,0
#             6
#
our $triple_7th_2_4_6_segs = MyDragonConf->new
  (name => '7th 2,4,6',
   side => 0,
   segs => [ "0,0,0",
             "0,0,1",
             "1,1,3",
           ]);
$triple_7ths_segs[6] = $triple_7th_2_4_6_segs;
$triple_7ths_segs[2] = $triple_7th_2_4_6_segs->rotate(-1);
$triple_7ths_segs[4] = $triple_7th_2_4_6_segs->translate(-1,-1)->rotate(1);

our $triple_7ths_union_segs;
{
  # conf triple 7ths union side=undef
  #    0,0,3 0,0,0 1,-1,2 0,0,1 1,-1,1 1,1,3
  #    *     *
  #    ^     |
  #    |     |
  #    |     v
  #    O====>*
  #    |     ^
  #    |     |
  #    v     |
  #    *<----*
  my $u = MyDragonConf->new (side => 0);
  $u = $u->union(@triple_7ths_segs[1..6])->set_name('triple 7ths union');
  $triple_7ths_union_segs = $u;
  #  $u->print;
}

# Union of $triple_7ths_union_segs also at each possible segment before and
# after middle m.
our $triple_7ths_and_next_segs = do {
  # conf triple 7ths and next side=0
  #    0,0,3 0,0,0 1,-1,2 0,0,1 1,-1,1 1,1,3 0,0,2 -1,-1,1 -1,-1,0 1,1,2 -1,1,0 1,-1,0 2,0,3 2,0,2 1,1,0
  #    *---->*<----*---->*
  #          ^     |
  #          |     |
  #          |     v
  #    *<----O====>*<----*
  #    ^     |     ^     |
  #    |     |     |     |
  #    |     v     |     v
  #    *---->*<----*---->*

  $triple_7ths_union_segs
    ->union_in_positions($conf_before_and_after)
    ->set_name('triple 7ths and next');
};
# $triple_7ths_and_next_segs->print;

our $triple_7ths_and_next_within;
{
  # nothing extra
  my $u = $triple_7ths_and_next_segs;
  $u = $u->union (MyDragonConf->new(segs => []));
  $triple_7ths_and_next_within = $u->set_name('triple 7ths and next within');
  # $u->print;
}

{
  package MyDragonConf;
  sub triple_7ths_list {
    my ($self) = @_;
    my @ret;
    foreach my $i (1 .. $#triple_7ths_segs) {
      if ($triple_7ths_segs[$i]->is_subset($self)) {
        push @ret, $i;
      }
    }
    return @ret;
  }

  # in order curve start to curve end
  sub triple_7ths_diffs {
    my ($self) = @_;
    my @ret = MyDragon::list_differences($self->triple_7ths_list);
    if ($self->{'side'}) {
      return reverse @ret;
    } else {
      return @ret;
    }
  }

  # in order curve start to curve end
  sub triple_7ths_diff_to_next {
    my ($self) = @_;
    $self->{'want_next_segs'}
      or croak "triple_7ths_diff_to_next() must have want_next_segs";

    my @this = $self->triple_7ths_list or return ();
    my ($next_seg) = $self->next_segs   or return ();
    my ($x,$y,$dir) = split /,/, $next_seg;
    my @next = $self->translate(-$x,-$y)->rotate(-$dir)->triple_7ths_list
      or return ();
    if ($self->{'side'} == 0) {
      #          end *
      #              |
      #              v
      # start * ---> *
      return 7 - $this[-1] + 7 - $next[-1];
    } else {
      # * end
      # ^
      # |
      # * ---> * start
      return $this[0] + $next[0];
    }
  }
}
sub list_differences {
  return map {$_[$_+1] - $_[$_]} 0 .. $#_-1;
}

our (@triple_7ths_state_to_conf,
     @triple_7ths_state_table,
     @triple_7ths_diff_lists);
{
  use constant::defer triple_7ths_surround_init => sub {
    my %options = @_;
    ### triple_7ths_surround_init() ...
    my $verbose = $options{'verbose'} || 0;
    # $verbose = 2;

    my %conf_to_state;
    my $conf_to_state = sub {
      my ($conf) = @_;
      my $key = $conf->key;
      my $state = $conf_to_state{$key};
      if (! defined $state) {
        push @triple_7ths_state_to_conf, $conf;
        $state = $#triple_7ths_state_to_conf;
        $conf_to_state{$key} = $state;
        # print "new state $state\n";
      }
      return $state;
    };
    $conf_to_state->(MyDragonConf->initial
                     (within => $triple_7ths_and_next_within,
                      want_next_segs => 1));
    ### @triple_7ths_state_to_conf

    {
      my $state = 0;
      my $depth = 0;
      for (;;) {
        $depth++;
        my $last_state = $#triple_7ths_state_to_conf;
        for ( ; $state <= $last_state; $state++) {
          my $from_conf = $triple_7ths_state_to_conf[$state];
          if ($verbose >= 3) {
            print "########################################\n";
            print "from $state  ",$from_conf->key,"\n";
            $from_conf->print;
          }
          foreach my $bit (0,1) {
            my $to_conf = $from_conf->expand($bit);
            my $to_state = $conf_to_state->($to_conf);
            $triple_7ths_state_table[$state]->[$bit] = $to_state;

            if ($verbose >= 3) {
              print "  bit=$bit to $triple_7ths_state_table[$state]->[$bit]  ",$to_conf->key,"\n";
              $to_conf->print;
            }
          }
        }
        my $new = $#triple_7ths_state_to_conf - $last_state;
        if ($verbose) {
          print "depth $depth new $new\n";
        }
        last if $new == 0;
      }

      # exit if $from == 6;
      # print "  $from=>{0=>$triple_7ths_state_table[$from]->[0],1=>$triple_7ths_state_table[$from]->[1]},\n";
      # {
      #   my $f = $from + 1;
      #   my $t0 = $triple_7ths_state_table[$from]->[0] + 1;
      #   my $t1 = $triple_7ths_state_table[$from]->[1] + 1;
      #   print "  $f=>{0=>$t0,1=>$t1},\n";
      # }
    }

    if ($verbose) {
      my $num_states = scalar(@triple_7ths_state_table);
      print "triple_7ths_state_table length $num_states\n";
    }

    foreach my $state (0 .. $#triple_7ths_state_to_conf) {
      my $conf = $triple_7ths_state_to_conf[$state];
      my @diffs = ($conf->triple_7ths_diffs, $conf->triple_7ths_diff_to_next);
      @diffs = List::Util::uniqnum(@diffs);
      foreach my $diff (@diffs) {
        push @{$triple_7ths_diff_lists[$diff]}, $state;
      }
    }
    if ($verbose >= 1) {
      print "diffs occurring ";
      foreach my $diff (0 .. $#triple_7ths_diff_lists) {
        my $aref = $triple_7ths_diff_lists[$diff] || next;
        print "$diff,";
      }
      print "\n";
    }
    if ($verbose >= 2) {
      foreach my $diff (0 .. $#triple_7ths_diff_lists) {
        my $aref = $triple_7ths_diff_lists[$diff] || next;
        print "diff $diff states ",join(',',@$aref),"\n";
      }
    }
  };
}


#------------------------------------------------------------------------------
# 15ths Triples Segments - Fifteenths

our @triple_15ths_segs;

#      0,1
#   1   ^
#       |    3
#      0,0 ====> 1,0
#       |
#  11   v
#      0,-1
#
our $triple_1_3_11_segs = MyDragonConf->new
  (name => '15th 1,3,11',
   side => 0,
   segs => [ '0,0,0',
             '0,0,1',
             '0,0,3',
           ]);
$triple_15ths_segs[3]  = $triple_1_3_11_segs;
$triple_15ths_segs[1]  = $triple_1_3_11_segs->rotate(-1);
$triple_15ths_segs[11] = $triple_1_3_11_segs->rotate(1);

#    0,1
#  2  ^
#     |    6
#    0,0 ====> 1,0
#
#
#    0,-1 <--- 1,-1
#           8
#
our $triple_2_6_8_segs = MyDragonConf->new
  (name => '15th 2,6,8',
   side => 0,
   segs => [ "0,0,0",
             "0,0,1",
             "1,-1,2",
           ]);
$triple_15ths_segs[6] = $triple_2_6_8_segs;
$triple_15ths_segs[2] = $triple_2_6_8_segs->rotate(-1);
$triple_15ths_segs[8] = $triple_2_6_8_segs->translate(-1,1)->rotate(2);

#     *     *
#     ^
#   4 |
#     |  12
#     O====>*
#           ^
#           |  14
#           |
#     *     *
#
our $triple_4_12_14_segs = MyDragonConf->new
  (name => '15th 4,12,14',
   side => 0,
   segs => [ "0,0,0",
             "0,0,1",
             "1,-1,1",
           ]);
$triple_15ths_segs[12] = $triple_4_12_14_segs;
$triple_15ths_segs[4]  = $triple_4_12_14_segs->rotate(-1);
$triple_15ths_segs[14] = $triple_4_12_14_segs->translate(-1,1)->rotate(-1);

#       7,9
#     O====>*
#           ^
#           | 13
#           |
#     *     *
#
our $triple_7_9_13_segs = MyDragonConf->new
  (name => '15th 7,9,13',
   side => 0,
   segs => [ "0,0,0",
             "1,-1,1",
           ]);
$triple_15ths_segs[7]  = $triple_7_9_13_segs;
$triple_15ths_segs[9]  = $triple_7_9_13_segs;
$triple_15ths_segs[13] = $triple_7_9_13_segs->translate(-1,1)->rotate(-1);

our @triple_15ths_configurations = ($triple_1_3_11_segs,
                                    $triple_2_6_8_segs,
                                    $triple_4_12_14_segs,
                                    $triple_7_9_13_segs);

our $triple_15ths_union_segs;
{
  # *     *     *---->*
  #       ^     |
  #       |     |
  #       |     v
  # *<----O====>*     *
  # ^     |     ^
  # |     |     |
  # |     v     |
  # *---->*<----*     *
  #             |
  #             |
  #             v
  # *     *     *     *
  my $u;
  foreach my $conf (@triple_15ths_configurations) {
    foreach my $seg ($conf->segs_list) {
      my ($x,$y,$dir) = split /,/, $seg;
      my $c = $conf->translate(-$x,-$y);
      $c = $c->rotate(-$dir);
      $c->contains('0,0,0') or die;
      # $c->print;
      $u //= $c;
      $u = $u->union($c);
    }
  }
  # $u->print;
  $triple_15ths_union_segs = $u->set_name('triple 15ths union');
}

our $triple_15ths_and_next_segs = do {
  # conf triple 15ths and next side=0
  #    0,0,0 0,0,1 0,0,3 0,0,2 1,-1,2 -1,-1,1 1,-1,3 1,-1,1 -1,-1,0 1,1,3 1,1,0 -1,1,0 -1,-1,2 -1,1,3 1,1,2 -1,1,1 1,-1,0 2,0,3 2,-2,2 2,0,0 2,0,2 2,-2,1 1,1,1 0,2,0 0,2,3
  #    *     *     *---->*     *     *
  #          ^     |     ^
  #          |     |     |
  #          |     v     |
  #    *     *---->*<----*---->*     *
  #          |     ^     |
  #          |     |     |
  #          v     |     v
  #    *     *<----O====>*<----*---->*
  #          ^     |     ^     |
  #          |     |     |     |
  #          |     v     |     v
  #    *<----*---->*<----*---->*     *
  #                      |     ^
  #                      |     |
  #                      v     |
  #    *     *     *     *<----*     *

  $triple_15ths_union_segs
    ->union_in_positions($conf_before_and_after)
    ->set_name('triple 15ths and next');
};
# $triple_15ths_and_next_segs->print;

our $triple_15ths_and_next_within = do {
# conf triple 15ths and next within side=0
#    0,0,0 0,0,1 0,0,3 0,0,2 1,-1,2 -1,-1,1 1,-1,3 1,-1,1 -1,-1,0 1,1,3 1,1,0 -1,1,0 -1,-1,2 -1,1,3 1,1,2 -1,1,1 1,-1,0 2,0,3 2,-2,2 2,0,0 2,0,2 2,-2,1 1,1,1 0,2,0 0,2,3 0,-2,1 0,-2,2
#    *     *     *---->*     *     *
#          ^     |     ^
#          |     |     |
#          |     v     |
#    *     *---->*<----*---->*     *
#          |     ^     |
#          |     |     |
#          v     |     v
#    *     *<----O====>*<----*---->*
#          ^     |     ^     |
#          |     |     |     |
#          |     v     |     v
#    *<----*---->*<----*---->*     *
#                ^     |     ^
#                |     |     |
#                |     v     |
#    *     *<----*     *<----*     *

  $triple_15ths_and_next_segs
    ->union(MyDragonConf->new(segs => ['0,-2,1', '0,-2,2']))
    ->set_name('triple 15ths and next within');
};
# $triple_15ths_and_next_within->print;

{
  package MyDragonConf;
  sub triple_15ths_list {
    my ($self) = @_;
    my @ret;
    foreach my $i (0 .. $#triple_15ths_segs) {
      if (defined $triple_15ths_segs[$i]
          && $triple_15ths_segs[$i]->is_subset($self)) {
        push @ret, $i;
      }
    }
    return @ret;
  }

  # in order curve start to curve end
  sub triple_15ths_diffs {
    my ($self) = @_;
    my @ret = MyDragon::list_differences($self->triple_15ths_list);
    if ($self->{'side'}) {
      return reverse @ret;
    } else {
      return @ret;
    }
  }

  # Return the difference from last 15th triple in $self to the first 15th
  # in the next segment, in curve start to curve end order.
  # Every segment has one or more 15th triple so looking at one next segment
  # suffices.
  #
  sub triple_15ths_diff_to_next {
    my ($self) = @_;
    $self->{'want_next_segs'}
      or croak "triple_15ths_diff_to_next() must have want_next_segs";

    my @this = $self->triple_15ths_list or return ();
    my ($next_seg) = $self->next_segs   or return ();
    my ($x,$y,$dir) = split /,/, $next_seg;
    my @next = $self->translate(-$x,-$y)->rotate(-$dir)->triple_15ths_list
      or return ();
    if ($self->{'side'} == 0) {
      #          end *
      #              |
      #              v
      # start * ---> *
      return 15 - $this[-1] + 15 - $next[-1];
    } else {
      # * end
      # ^
      # |
      # * ---> * start
      return $this[0] + $next[0];
    }
  }
}

our (@triple_15ths_state_to_conf,
     @triple_15ths_state_table,
     @triple_15ths_diff_lists);
{
  use constant::defer triple_15ths_surround_init => sub {
    my %options = @_;
    ### triple_15ths_surround_init() ...
    my $verbose = $options{'verbose'} || 0;
    # $verbose = 2;

    my %conf_to_state;
    my $conf_to_state = sub {
      my ($conf) = @_;
      my $key = $conf->key;
      my $state = $conf_to_state{$key};
      if (! defined $state) {
        push @triple_15ths_state_to_conf, $conf;
        $state = $#triple_15ths_state_to_conf;
        $conf_to_state{$key} = $state;
        # print "new state $state\n";
      }
      return $state;
    };
    $conf_to_state->(MyDragonConf->initial
                     (within => $triple_15ths_and_next_within,
                      want_next_segs => 1));
    ### @triple_15ths_state_to_conf

    {
      my $state = 0;
      my $depth = 0;
      for (;;) {
        $depth++;
        my $last_state = $#triple_15ths_state_to_conf;
        for ( ; $state <= $last_state; $state++) {
          my $from_conf = $triple_15ths_state_to_conf[$state];
          if ($verbose >= 3) {
            print "########################################\n";
            print "from $state  ",$from_conf->key,"\n";
            $from_conf->print;
          }
          foreach my $bit (0,1) {
            my $to_conf = $from_conf->expand($bit);
            my $to_state = $conf_to_state->($to_conf);
            $triple_15ths_state_table[$state]->[$bit] = $to_state;

            if ($verbose >= 3) {
              print "  bit=$bit to $triple_15ths_state_table[$state]->[$bit]  ",$to_conf->key,"\n";
              $to_conf->print;
            }
          }
        }
        my $new = $#triple_15ths_state_to_conf - $last_state;
        if ($verbose) {
          print "depth $depth new $new\n";
        }
        last if $new == 0;
      }

      # exit if $from == 6;
      # print "  $from=>{0=>$triple_15ths_state_table[$from]->[0],1=>$triple_15ths_state_table[$from]->[1]},\n";
      # {
      #   my $f = $from + 1;
      #   my $t0 = $triple_15ths_state_table[$from]->[0] + 1;
      #   my $t1 = $triple_15ths_state_table[$from]->[1] + 1;
      #   print "  $f=>{0=>$t0,1=>$t1},\n";
      # }
    }

    if ($verbose) {
      my $num_states = scalar(@triple_15ths_state_table);
      print "triple_15ths_state_table length $num_states\n";
    }

    foreach my $state (0 .. $#triple_15ths_state_to_conf) {
      my $conf = $triple_15ths_state_to_conf[$state];
      my @diffs = ($conf->triple_15ths_diffs, $conf->triple_15ths_diff_to_next);
      @diffs = List::Util::uniqnum(@diffs);
      foreach my $diff (@diffs) {
        push @{$triple_15ths_diff_lists[$diff]}, $state;
      }
    }
    if ($verbose >= 1) {
      print "diffs occurring ";
      foreach my $diff (0 .. $#triple_15ths_diff_lists) {
        my $aref = $triple_15ths_diff_lists[$diff] || next;
        print "$diff,";
      }
      print "\n";
    }
    if ($verbose >= 2) {
      foreach my $diff (0 .. $#triple_15ths_diff_lists) {
        my $aref = $triple_15ths_diff_lists[$diff] || next;
        print "diff $diff states ",join(',',@$aref),"\n";
      }
    }

    # require MyFLAT;
    # {
    #   # R
    #   my $f = MyFLAT::aref_to_FLAT_DFA
    #     (\@triple_15ths_state_table,
    #      name => 'fNonRpred_surround',
    #      accepting_list => $triple_15ths_state_enclosed_lists[0],
    #     );
    #   $f = MyFLAT::eventually_accepting($f);
    #   $fNonRpred_surround_flat = $f;
    #   if ($verbose >= 2) {
    #     print "$f->{name} accepting ",join(' ',$f->get_accepting),"\n";
    #   }
    #
    #   $f = $f->complement;
    #   $f->{'name'} = 'fRpred_surround';
    #   $fRpred_surround_flat = $f;
    # }
  };
}


#------------------------------------------------------------------------------
# 5ths Triples - Fifths

our $triple_5ths_union_segs = do {
  # The four configurations which have the 5th in the middle.

  # conf triple 5ths union side=0
  #    0,0,0 0,0,1 0,0,3 1,-1,2 1,-1,1
  #    *     *
  #    ^
  #    |
  #    |
  #    O====>*
  #    |     ^
  #    |     |
  #    v     |
  #    *<----*

  # @triple_15ths_configurations has each configuration with the
  # multiple-of-3 segment as middle m, so union of those.
  #
  MyDragonConf->new(side=>0)
    ->union(@triple_15ths_configurations)
    ->set_name('triple 5ths union');
};
# $triple_5ths_union_segs->print;

our $triple_5ths_nexts_segs = do {
  #           R                          R
  #     O====>*     *              O====>*
  #           ^     |                    ^
  #           |     |                    |
  #           |     v                    |
  #     *   E *---->* L          L *<----* E=R
  #        =L                      ^
  #                                |
  #                                |
  #                                *     *

  #    *---->*     *     *
  #          ^     |
  #          |     |
  #          |     v
  #    *     O====>*     *
  #          |     ^     |
  #          |     |     |
  #          v     |     v
  #    *     *<----*---->*
  #          ^
  #          |
  #          |
  #    *     *     *     *

  my $u = MyDragonConf->new (side => 0);
  foreach my $aref (['2,0,3', '1,-1,0', '1,-1,1', '0,0,0'],       # R-L-L
                    ['0,-2,1', '1,-1,2', '1,-1,1', '0,0,0']) {    # R-R-L
    # MyDragonConf->new(segs => $aref)->print;
    foreach my $i (0 .. $#$aref) {  # starting at
      my $first_seg = $aref->[$i];
      my $a = MyDragonConf->new(segs => [ @{$aref}[$i .. $#$aref] ]);
      $a = $a->transform_to_seg($first_seg)->set_name("i=$i");
      # $a->print;
      $u = $u->union($a);
    }
  }
  # $u->print;
  $u->set_name('triple 5ths nexts');
};

our $triple_5ths_and_next_segs = do {
  #    *     *     *     *     *
  #    ^
  #    |
  #    |
  #    *---->*<----*---->*     *
  #    |     ^     |
  #    |     |     |
  #    v     |     v
  #    *<----O====>*<----*---->*
  #    ^     |     ^     |
  #    |     |     |     |
  #    |     v     |     v
  #    *---->*<----*---->*     *
  #          ^     |     ^
  #          |     |     |
  #          |     v     |
  #    *<----*---->*<----*     *
  #
  $triple_5ths_union_segs->union_in_positions($triple_5ths_nexts_segs)
    ->set_name('triple 5ths and nexts');
};
# $triple_5ths_and_next_segs->print;

# with extras needed to expand within this set of segments
our $triple_5ths_and_next_within = do {
  $triple_5ths_and_next_segs
    ->union (MyDragonConf->new(segs => ['-1,-1,2', '1,1,1']))
    ->set_name('triple 5ths and nexts within');
};
# $triple_5ths_and_next_within->print;

our @triple_5ths_segs = (undef,
                         $triple_1_3_11_segs,   # 1*3 = 3
                         $triple_2_6_8_segs,    # 2*3 = 6
                         $triple_7_9_13_segs,   # 3*3 = 9
                         $triple_4_12_14_segs); # 4*3 = 12
{
  package MyDragonConf;
  sub triple_5ths_list {
    my ($self) = @_;
    my @ret;
    foreach my $i (0 .. $#triple_5ths_segs) {
      if (defined $triple_5ths_segs[$i]
          && $triple_5ths_segs[$i]->is_subset($self)) {
        push @ret, $i;
      }
    }
    return @ret;
  }

  # in order curve start to curve end
  sub triple_5ths_diffs {
    my ($self) = @_;
    my @ret = MyDragon::list_differences($self->triple_5ths_list);
    if ($self->{'side'}) {
      return reverse @ret;
    } else {
      return @ret;
    }
  }

  # Return the difference from last 5th triple in $self to the first 5th
  # triple in whichever next segment has some 5th triples.
  # Some segments have no 5th triples.
  #
  sub triple_5ths_diff_to_next {
    my ($self) = @_;
    $self->{'want_next_segs'}
      or croak "triple_5ths_diff_to_next() must have want_next_segs";

    my @this = $self->triple_5ths_list
      or return ();
    my @next_segs = $self->next_segs
      or return ();

    # side=0 towards end of segment
    # side=1 back towards start of segment
    my $ret = ($self->{'side'} ? $this[0] : 5-$this[-1]);
    ### $ret
    foreach my $i (0 .. $#next_segs) {
      my $next_seg = $next_segs[$i];
      my ($x,$y,$dir) = split /,/, $next_seg;
      my @next = $self->translate(-$x,-$y)->rotate(-$dir)->triple_5ths_list;
      ### $i
      ### $next_seg
      ### @next
      if (@next) {
        # i=0 same as side, i=1 opposite, etc
        my $next = (($self->{'side'}+$i)&1 ? $next[0] : 5-$next[-1]);
        ### $next
        return $ret + $next;
      }
      $ret += 5;
    }
    if (@next_segs == $self->{'want_next_segs'}) {
      croak "triple_5ths_diff_to_next() not enough want_next_segs to find next";
    }
    return ();
  }
}

our (@triple_5ths_state_to_conf,
     @triple_5ths_state_table,
     @triple_5ths_diff_lists);
{
  use constant::defer triple_5ths_surround_init => sub {
    my %options = @_;
    ### triple_5ths_surround_init() ...
    my $verbose = $options{'verbose'} || 0;
    # $verbose = 2;

    my %conf_to_state;
    my $conf_to_state = sub {
      my ($conf) = @_;
      my $key = $conf->key;
      my $state = $conf_to_state{$key};
      if (! defined $state) {
        push @triple_5ths_state_to_conf, $conf;
        $state = $#triple_5ths_state_to_conf;
        $conf_to_state{$key} = $state;
        # print "new state $state\n";
      }
      return $state;
    };
    $conf_to_state->(MyDragonConf->initial
                     (within => $triple_5ths_and_next_within,
                      want_next_segs => 3));
    ### @triple_5ths_state_to_conf

    {
      my $state = 0;
      my $depth = 0;
      for (;;) {
        $depth++;
        my $last_state = $#triple_5ths_state_to_conf;
        for ( ; $state <= $last_state; $state++) {
          my $from_conf = $triple_5ths_state_to_conf[$state];
          if ($verbose >= 3) {
            print "########################################\n";
            print "from $state  ",$from_conf->key,"\n";
            $from_conf->print;
          }
          foreach my $bit (0,1) {
            my $to_conf = $from_conf->expand($bit);
            my $to_state = $conf_to_state->($to_conf);
            $triple_5ths_state_table[$state]->[$bit] = $to_state;

            if ($verbose >= 3) {
              print "  bit=$bit to $triple_5ths_state_table[$state]->[$bit]  ",$to_conf->key,"\n";
              $to_conf->print;
            }
          }
        }
        my $new = $#triple_5ths_state_to_conf - $last_state;
        if ($verbose) {
          print "depth $depth new $new\n";
        }
        last if $new == 0;
      }

      # exit if $from == 6;
      # print "  $from=>{0=>$triple_5ths_state_table[$from]->[0],1=>$triple_5ths_state_table[$from]->[1]},\n";
      # {
      #   my $f = $from + 1;
      #   my $t0 = $triple_5ths_state_table[$from]->[0] + 1;
      #   my $t1 = $triple_5ths_state_table[$from]->[1] + 1;
      #   print "  $f=>{0=>$t0,1=>$t1},\n";
      # }
    }

    if ($verbose) {
      my $num_states = scalar(@triple_5ths_state_table);
      print "triple_5ths_state_table length $num_states\n";
    }

    foreach my $state (0 .. $#triple_5ths_state_to_conf) {
      my $conf = $triple_5ths_state_to_conf[$state];
      my @diffs = ($conf->triple_5ths_diffs, $conf->triple_5ths_diff_to_next);
      @diffs = List::Util::uniqnum(@diffs);
      foreach my $diff (@diffs) {
        push @{$triple_5ths_diff_lists[$diff]}, $state;
      }
    }
    if ($verbose >= 1) {
      print "diffs occurring ";
      foreach my $diff (0 .. $#triple_5ths_diff_lists) {
        my $aref = $triple_5ths_diff_lists[$diff] || next;
        print "$diff,";
      }
      print "\n";
    }
    if ($verbose >= 2) {
      foreach my $diff (0 .. $#triple_5ths_diff_lists) {
        my $aref = $triple_5ths_diff_lists[$diff] || next;
        print "diff $diff states ",join(',',@$aref),"\n";
      }
    }

    # require MyFLAT;
    # {
    #   # R
    #   my $f = MyFLAT::aref_to_FLAT_DFA
    #     (\@triple_5ths_state_table,
    #      name => 'fNonRpred_surround',
    #      accepting_list => $triple_5ths_state_enclosed_lists[0],
    #     );
    #   $f = MyFLAT::eventually_accepting($f);
    #   $fNonRpred_surround_flat = $f;
    #   if ($verbose >= 2) {
    #     print "$f->{name} accepting ",join(' ',$f->get_accepting),"\n";
    #   }
    #
    #   $f = $f->complement;
    #   $f->{'name'} = 'fRpred_surround';
    #   $fRpred_surround_flat = $f;
    # }
  };
}


#------------------------------------------------------------------------------
# fBpred by hulls touching

# All segments around 0,0,0 whose convex hull touches or overlaps the hull
# of 0,0,0.
our $hullstouch_segs = MyDragonConf->new
  (name => 'hulls touching or overlapping',
   side => 0,
   segs => [
            "0,0,0",
            "0,0,1",
            "0,0,2",
            "0,0,3",

            "1,1,0",
            "1,1,3",

            "1,-1,1",
            "1,-1,2",
            "1,-1,3",

            "-1,-1,0",
            "-1,-1,1",
            "2,0,2",
           ]);

our $hullstouch_segs_all = $hullstouch_segs;

our @hullstouch_L_squares
  = (MyDragonConf->new(name => 'hullstouch L top',
                       side => 0,
                       segs => [ "0,0,0",   "1,1,2",
                                 "0,0,1",   "1,1,3" ]),
     MyDragonConf->new(name => 'hullstouch L left',
                       side => 0,
                       segs => [ "-1,-1,0",   "0,0,2",
                                 "-1,-1,1",   "0,0,3" ]),
     MyDragonConf->new(name => 'hullstouch L left',
                       side => 0,
                       segs => [ "1,-1,0",   "2,0,2",
                                 "1,-1,1",   "2,0,3" ]),
     MyDragonConf->new(name => 'hullstouch L top',
                       side => 0,
                       segs => [ "0,-2,0",   "1,-1,2",
                                 "0,-2,1",   "1,-1,3" ]),
    );

our @hullstouch_R_squares
  = (MyDragonConf->new(name => 'hullstouch R left',
                       side => 0,
                       segs => [ "0,0,1",   "-1,1,0",
                                 "0,0,2",   "-1,1,3" ]),
     MyDragonConf->new(name => 'hullstouch R right',
                       side => 0,
                       segs => [ "2,0,1",   "1,1,0",
                                 "2,0,2",   "1,1,3" ]),
     MyDragonConf->new(name => 'hullstouch R below',
                       side => 0,
                       segs => [ "1,-1,1",   "0,0,0",
                                 "1,-1,2",   "0,0,3" ]),
    );
# foreach my $conf (@hullstouch_L_squares,@hullstouch_R_squares){$conf->print;}

# $side==0 for R, $side==1 for L, then modified by $conf->{'side'}.
# A nearby square is boundary of respective $side when it has at least one
# segment $conf but not all of its segments in conf.
sub hullstouch_is_boundary {
  my ($conf, $side) = @_;
  $side ^= $conf->{'side'};
  foreach my $square ($side==0 ? @hullstouch_R_squares
                      : $side==1 ? @hullstouch_L_squares
                      : die) {
    my $got = $square->intersection ($conf);
    if ($got->num_segs == 0) {
      # no segments at all around this square, not boundary
      next;
    }
    my $all = $square->intersection ($conf->{'within'});
    if ($got->equals($all)) {
      # all segments (of "within") around this square, not boundary
      next;
    }
    # at least 1 but fewer than all segments, so boundary
    return 1;
  }
  return 0;
}

our (@fBpred_hullstouch_state_to_conf,
     @fBpred_hullstouch_state_table,
     @fBpred_hullstouch_state_enclosed_lists,
     $fNonRpred_hullstouch_flat,$fNonLpred_hullstouch_flat,
     $fNonBpred_hullstouch_flat,
     $fRpred_hullstouch_flat,$fLpred_hullstouch_flat,
     $fBpred_hullstouch_flat);
{
  my @state_is_RLB_enclosed;
  my @state_is_RLB_enclosed_eventually;
  @fBpred_hullstouch_state_enclosed_lists = ([], [], []);

  use constant::defer fBpred_hullstouch_init => sub {
    my %options = @_;
    ### fBpred_hullstouch_init() ...
    my $verbose = $options{'verbose'} || 0;

    my %conf_to_state;
    my $conf_to_state = sub {
      my ($conf) = @_;
      my $key = $conf->key;
      my $state = $conf_to_state{$key};
      if (! defined $state) {
        push @fBpred_hullstouch_state_to_conf, $conf;
        $state = $#fBpred_hullstouch_state_to_conf;
        $conf_to_state{$key} = $state;
        # print "new state $state\n";
      }
      return $state;
    };
    $conf_to_state->(MyDragonConf->initial(within => $hullstouch_segs_all));
    ### @fBpred_hullstouch_state_to_conf
    $fBpred_hullstouch_state_to_conf[0]->is_boundary(0) or die;
    $fBpred_hullstouch_state_to_conf[0]->is_boundary(1) or die;

    {
      # find all reachable configurations
      my $state = 0;
      my $depth = 0;
      for (;;) {
        $depth++;
        my $last_state = $#fBpred_hullstouch_state_to_conf;
        for ( ; $state <= $last_state; $state++) {
          my $from_conf = $fBpred_hullstouch_state_to_conf[$state];
          if ($verbose >= 3) {
            print "########################################\n";
            print "from $state  ",$from_conf->key,"\n";
            $from_conf->print;
          }
          foreach my $bit (0,1) {
            my $to_conf = $from_conf->expand($bit);
            my $to_state = $conf_to_state->($to_conf);
            $fBpred_hullstouch_state_table[$state]->[$bit] = $to_state;

            if ($verbose >= 3) {
              print "  bit=$bit to $fBpred_hullstouch_state_table[$state]->[$bit]  ",$to_conf->key,"\n";
              $to_conf->print;
            }
          }
        }
        my $new = $#fBpred_hullstouch_state_to_conf - $last_state;
        if ($verbose) {
          print "depth $depth new $new\n";
        }
        last if $new == 0;
      }
    }

    if ($verbose) {
      my $num_states = scalar(@fBpred_hullstouch_state_table);
      print "fBpred hullstouch fBpred_hullstouch_state_table length $num_states\n";
    }

    foreach my $i (0 .. $#fBpred_hullstouch_state_to_conf) {
      my $conf = $fBpred_hullstouch_state_to_conf[$i];
      foreach my $side (0, 1) {  # R,L
        $state_is_RLB_enclosed[$i]->[$side]
          = ! hullstouch_is_boundary($conf,$side);
      }
      $state_is_RLB_enclosed[$i]->[2]  # both
        = $state_is_RLB_enclosed[$i]->[0] && $state_is_RLB_enclosed[$i]->[1];
    }
    foreach my $side (0, 1, 2) {
      $fBpred_hullstouch_state_enclosed_lists[$side]
        = [ grep {$state_is_RLB_enclosed[$_]->[$side]}
            0 .. $#fBpred_hullstouch_state_table ];
    }
    if ($verbose >= 2) {
      print "  enclosed R = ",join(',',@{$fBpred_hullstouch_state_enclosed_lists[0]}),
        " count ",scalar(@{$fBpred_hullstouch_state_enclosed_lists[0]}),"\n";
      print "  enclosed L = ",join(',',@{$fBpred_hullstouch_state_enclosed_lists[1]}),
        " count ",scalar(@{$fBpred_hullstouch_state_enclosed_lists[1]}),"\n";
      print "  enclosed B = ",join(',',@{$fBpred_hullstouch_state_enclosed_lists[2]}),
        " count ",scalar(@{$fBpred_hullstouch_state_enclosed_lists[2]}),"\n";
    }
    foreach my $side (2, 1, 0) {
      foreach my $i (@{$fBpred_hullstouch_state_enclosed_lists[$side]}) {
        foreach my $bit (0,1) {
          my $to = $fBpred_hullstouch_state_table[$i]->[$bit];
          unless ($state_is_RLB_enclosed[$to]->[$side]) {
            $fBpred_hullstouch_state_to_conf[$i]->print;
            $fBpred_hullstouch_state_to_conf[$i]->expand($bit)->print;
            print "oops, side=$side enclosed $i bit=$bit didn't remain so";
            # print $conf->is_boundary($side);
            exit 1;
          }
        }
      }
    }

    require MyFLAT;
    {
      # R
      my $f = MyFLAT::aref_to_FLAT_DFA
        (\@fBpred_hullstouch_state_table,
         name => 'fNonRpred_hullstouch',
         accepting_list => $fBpred_hullstouch_state_enclosed_lists[0],
        );
      if ($verbose >= 2) {
        print "$f->{name} extra eventually accepting: ",
          join(' ',$f->MyFLAT::get_eventually_accepting),"\n";
      }
      $f = MyFLAT::eventually_accepting($f);
      $fNonRpred_hullstouch_flat = $f;
      if ($verbose >= 2) {
        print "$f->{name} accepting ",join(' ',$f->get_accepting),"\n";
      }

      $f = $f->complement->MyFLAT::set_name('fRpred hullstouch');
      $fRpred_hullstouch_flat = $f;
    }
    {
      # L
      my $f = MyFLAT::aref_to_FLAT_DFA
        (\@fBpred_hullstouch_state_table, name => 'fNonLpred hullstouch',
         starting => 0,
         accepting_list => $fBpred_hullstouch_state_enclosed_lists[1],
        );
      if ($verbose >= 2) {
        print "$f->{name} extra eventually accepting: ",
          join(' ',$f->MyFLAT::get_eventually_accepting),"\n";
      }
      $f = MyFLAT::eventually_accepting($f);
      $fNonLpred_hullstouch_flat = $f;

      $f = $f->complement->MyFLAT::set_name('fLpred_hullstouch');
      $fLpred_hullstouch_flat = $f;
    }
    {
      # B
      my $f = MyFLAT::aref_to_FLAT_DFA
        (\@fBpred_hullstouch_state_table, name => 'fNonBpred_hullstouch',
         starting => 0,
         accepting_list => $fBpred_hullstouch_state_enclosed_lists[2],
        );
      $f = MyFLAT::eventually_accepting($f);
      $fNonBpred_hullstouch_flat = $f;
      if ($verbose >= 2) {
        print "$f->{name} accepting ",join(' ',$f->get_accepting),"\n";
      }
      $f = $f->complement->MyFLAT::set_name('fBpred_hullstouch');
      $fBpred_hullstouch_flat = $f;
    }
  };
}


#------------------------------------------------------------------------------
# fBpred by surrounding segments

our $surround_segs_outer = MyDragonConf->new
  (name => 'surround outer',
   side => 0,
   segs => [
            # anti-clockwise from -1,-1
            "-1,-1,1",
            "-1,-1,0",
            "0,-2,1",
            "0,-2,0",
            "1,-1,3",
            "1,-1,0",
            "2,0,3",
            "2,0,1",
            "1,1,0",
            "1,1,2",
            "-1,1,0",
            "-1,1,3",
           ]);

our $surround_segs_inner = MyDragonConf->new
  (name => 'surround inner',
   side => 0,
   segs => [
            "0,0,0",

            # internals
            "0,0,1",
            "0,0,2",
            "0,0,3",
            "1,1,3",
            "2,0,2",
            "1,-1,1",
            "1,-1,2",
           ]);

our $surround_segs_all
  = $surround_segs_outer->union($surround_segs_inner)->set_name('surround all');

our (@fBpred_surround_state_to_conf,
     @fBpred_surround_state_table,
     @fBpred_surround_state_enclosed_lists,
     $fNonRpred_surround_flat,$fNonLpred_surround_flat,
     $fNonBpred_surround_flat,
     $fRpred_surround_flat,$fLpred_surround_flat,
     $fBpred_surround_flat);
{
  my @state_is_RLB_enclosed;
  my @state_is_RLB_enclosed_eventually;
  @fBpred_surround_state_enclosed_lists = ([], [], []);

  use constant::defer fBpred_surround_init => sub {
    my %options = @_;
    ### fBpred_surround_init() ...
    my $verbose = $options{'verbose'} || 0;

    my %conf_to_state;
    my $conf_to_state = sub {
      my ($conf) = @_;
      my $key = $conf->key;
      my $state = $conf_to_state{$key};
      if (! defined $state) {
        push @fBpred_surround_state_to_conf, $conf;
        $state = $#fBpred_surround_state_to_conf;
        $conf_to_state{$key} = $state;
        # print "new state $state\n";
      }
      return $state;
    };
    $conf_to_state->(MyDragonConf->initial(within => $surround_segs_all));
    ### @fBpred_surround_state_to_conf
    $fBpred_surround_state_to_conf[0]->is_boundary(0) or die;
    $fBpred_surround_state_to_conf[0]->is_boundary(1) or die;

    {
      my $state = 0;
      my $depth = 0;
      for (;;) {
        $depth++;
        my $last_state = $#fBpred_surround_state_to_conf;
        for ( ; $state <= $last_state; $state++) {
          my $from_conf = $fBpred_surround_state_to_conf[$state];
          if ($verbose >= 3) {
            print "########################################\n";
            print "from $state  ",$from_conf->key,"\n";
            $from_conf->print;
          }
          foreach my $bit (0,1) {
            my $to_conf = $from_conf->expand($bit);
            my $to_state = $conf_to_state->($to_conf);
            $fBpred_surround_state_table[$state]->[$bit] = $to_state;

            if ($verbose >= 3) {
              print "  bit=$bit to $fBpred_surround_state_table[$state]->[$bit]  ",$to_conf->key,"\n";
              $to_conf->print;
            }
          }
        }
        my $new = $#fBpred_surround_state_to_conf - $last_state;
        if ($verbose) {
          print "depth $depth new $new\n";
        }
        last if $new == 0;
      }

      # exit if $from == 6;
      # print "  $from=>{0=>$fBpred_surround_state_table[$from]->[0],1=>$fBpred_surround_state_table[$from]->[1]},\n";
      # {
      #   my $f = $from + 1;
      #   my $t0 = $fBpred_surround_state_table[$from]->[0] + 1;
      #   my $t1 = $fBpred_surround_state_table[$from]->[1] + 1;
      #   print "  $f=>{0=>$t0,1=>$t1},\n";
      # }
    }

    if ($verbose) {
      my $num_states = scalar(@fBpred_surround_state_table);
      print "fBpred surround fBpred_surround_state_table length $num_states\n";
    }

    # print "R_squares\n";
    # foreach my $segs (@R_squares) { segs_print($segs); }
    # print "L_squares\n";
    # foreach my $segs (@L_squares) { segs_print($segs); }

    foreach my $i (0 .. $#fBpred_surround_state_to_conf) {
      my $conf = $fBpred_surround_state_to_conf[$i];
      foreach my $side (0, 1) {
        $state_is_RLB_enclosed[$i]->[$side] = ! $conf->is_boundary($side);
      }
      $state_is_RLB_enclosed[$i]->[2]
        = $state_is_RLB_enclosed[$i]->[0]
        && $state_is_RLB_enclosed[$i]->[1];
    }
    foreach my $side (0, 1, 2) {
      $fBpred_surround_state_enclosed_lists[$side]
        = [ grep {$state_is_RLB_enclosed[$_]->[$side]} 0 .. $#fBpred_surround_state_table ];
    }
    if ($verbose >= 2) {
      print "  enclosed R = ",join(',',@{$fBpred_surround_state_enclosed_lists[0]}),
        " count ",scalar(@{$fBpred_surround_state_enclosed_lists[0]}),"\n";
      print "  enclosed L = ",join(',',@{$fBpred_surround_state_enclosed_lists[1]}),
        " count ",scalar(@{$fBpred_surround_state_enclosed_lists[1]}),"\n";
      print "  enclosed B = ",join(',',@{$fBpred_surround_state_enclosed_lists[2]}),
        " count ",scalar(@{$fBpred_surround_state_enclosed_lists[2]}),"\n";
    }
    foreach my $side (2, 1, 0) {
      foreach my $i (@{$fBpred_surround_state_enclosed_lists[$side]}) {
        foreach my $bit (0,1) {
          my $to = $fBpred_surround_state_table[$i]->[$bit];
          unless ($state_is_RLB_enclosed[$to]->[$side]) {
            $fBpred_surround_state_to_conf[$i]->print;
            $fBpred_surround_state_to_conf[$i]->expand($bit)->print;
            print "oops, side=$side enclosed $i bit=$bit didn't remain so";
            # print $conf->is_boundary($side);
            exit 1;
          }
        }
      }
    }

    # eventually enclosed configurations = those with all of $segs_eventually_B
    # {
    #   my $bad = 0;
    #   my $side = 2;
    #   foreach my $state (0 .. $#fBpred_surround_state_table) {
    #     my $conf = $fBpred_surround_state_to_conf[$state];
    #     my $enc = $state_is_RLB_enclosed_eventually[$state]->[$side];
    #     my $allof = $segs_eventually_B->is_subset($conf) ? 1 : 0;
    #     if ($enc != $allof) {
    #       print "state $state enc $enc  triples $allof\n";
    #       conf_print($conf);
    #       $bad++;
    #     }
    #   }
    #   if ($bad) { exit 1; }
    # }

    require MyFLAT;
    {
      # R
      my $f = MyFLAT::aref_to_FLAT_DFA
        (\@fBpred_surround_state_table,
         name => 'fNonRpred_surround',
         accepting_list => $fBpred_surround_state_enclosed_lists[0],
        );
      $f = MyFLAT::eventually_accepting($f);
      $fNonRpred_surround_flat = $f;
      if ($verbose >= 2) {
        print "$f->{name} accepting ",join(' ',$f->get_accepting),"\n";
      }

      $f = $f->complement->MyFLAT::set_name('fRpred_surround');
      $fRpred_surround_flat = $f;
    }
    {
      # L
      my $f = MyFLAT::aref_to_FLAT_DFA
        (\@fBpred_surround_state_table, name => 'fNonLpred_surround',
         starting => 0,
         accepting_list => $fBpred_surround_state_enclosed_lists[1],
        );
      $f = MyFLAT::eventually_accepting($f);
      $fNonLpred_surround_flat = $f;

      $f = $f->complement->MyFLAT::set_name('fLpred_surround');
      $fLpred_surround_flat = $f;
    }
    {
      # B
      my $f = MyFLAT::aref_to_FLAT_DFA
        (\@fBpred_surround_state_table, name => 'fNonBpred_surround',
         starting => 0,
         accepting_list => $fBpred_surround_state_enclosed_lists[2],
        );
      $f = MyFLAT::eventually_accepting($f);
      $fNonBpred_surround_flat = $f;
      if ($verbose >= 2) {
        print "$f->{name} accepting ",join(' ',$f->get_accepting),"\n";
      }
      $f = $f->complement->MyFLAT::set_name('fBpred_surround');
      $fBpred_surround_flat = $f;
    }
  };
}

#------------------------------------------------------------------------------
# fBpred hulls around

our $hullsaround_segs_outer = MyDragonConf->new
  (name => 'hull outer',
   side => 0,
   segs => [
            "2,0,3",
            "2,0,1",

            "2,-2,2",
            "2,-2,1",

            "0,-2,0",
            "0,-2,2",

            "-1,-1,3",
            "-1,-1,2",

            "-2,0,3",
            "-2,0,0",

            "1,1,2",
            "1,1,1",

            "2,2,3",
            "2,2,2",

            "-1,1,3",
            "-1,1,0",
           ]);

our $hullsaround_segs_inner = MyDragonConf->new
  (name => 'hull inner',
   segs => [
            "0,0,0",
            "0,0,1",
            "0,0,2",
            "0,0,3",

            "1,1,0",
            "1,1,3",

            "1,-1,0",
            "1,-1,1",
            "1,-1,2",
            "1,-1,3",

            "-1,-1,0",
            "-1,-1,1",
            "0,-2,1",
            "2,0,2",
           ]);

our $hullsaround_segs_all
  = $hullsaround_segs_outer->union($hullsaround_segs_inner)
  ->set_name('hull all');

#------------------------------------------------------------------------------
# fBpred by cases

# $side is 'left' or 'right'
# $parity is 0 for 2n or 1 for 2n+1
# $numerator is +/-1 or +/-7 or +/-11
sub fBpred_cases_extra {
  my ($side, $parity, $numerator) = @_;
  print "fBpred_cases_extra $side $parity $numerator\n";

  require MyFLAT;
  require FLAT;

  my $runs_flat = ($side eq 'left'
                   ? SpredLeft_runs_flat() : SpredRight_runs_flat());
  my $turn = $side eq 'left' ? '0' : '1';
  my $lowest = $numerator > 0 ? '1' : '0';
  my $lowest_opposite = 1-$lowest;
  my $even = $parity==0 ? "$lowest_opposite $lowest_opposite *" : '';
  my $fraction_flat
    = MyFLAT::fraction_digits($numerator > 0 ? $numerator : 15+$numerator,
                              15,
                              radix => 2);
  ### re: "$turn $lowest $even"
  my $f = $runs_flat->MyFLAT::concat
    (FLAT::Regex->new($turn . $lowest . $even)->as_dfa,
     $fraction_flat);

  if ($turn eq '0') {
    # for n=1 turn=0 is bit above so can omit
    ### u re: "$lowest $even"
    my $u = FLAT::Regex->new($lowest . $even)->as_dfa
      ->MyFLAT::concat($fraction_flat)
      ->MyFLAT::minimize;
    $f = $f->union($u);

    # MyFLAT::view($u);
    # $u->contains('101011') or die;
  }
  $f = $f->MyFLAT::prefix;

  # $f->contains('101011') or die;
  # MyFLAT::view($f);
  # MyFLAT::view($f);
  # $f->contains('101011') or die;

  $f = $f->as_dfa;
  $f = $f->MyFLAT::minimize;

  $f->{'name'} = sprintf "extra %s 2n%s %+d/15",
    $side, $parity==1 ? '+1' : '', $numerator;
  return $f;
}
BEGIN { memoize('fBpred_cases_extra', LIST_CACHE=>'MERGE'); }

use constant::defer fRpred_cases_extra_list => sub {
  my $extra_7_15_flat
    = MyFLAT::fraction_digits(7,15, radix => 2)
    ->MyFLAT::set_name('extra_7_15');
  return (fBpred_cases_extra('left',0,11),
          fBpred_cases_extra('left',0,-1),
          fBpred_cases_extra('left',1,-8),
          $extra_7_15_flat);
};
use constant::defer fLpred_cases_extra_list => sub {
  return (fBpred_cases_extra('right',0,-11),
          fBpred_cases_extra('right',0,  1),
          fBpred_cases_extra('right',1,  8));
};


#------------------------------------------------------------------------------
# fRpred, fLpred minimized

{
  my @state_table = ([2,1], [4,3], [2,5], [6,3], [8,7], [4,9], [11,10],
                     [13,12], [15,14], [6,16], [17,12], [19,18], [12,12],
                     [4,20], [22,21], [15,23], [24,16], [4,25], [6,21],
                     [12,26], [19,27], [6,9], [8,28], [22,9], [11,29],
                     [19,12], [12,27], [12,25], [13,27], [17,27]);
  my @accepting = (12);
  my $starting = 0;
  use constant::defer fNonRpred_flat => sub {
    require MyFLAT;
    MyFLAT::aref_to_FLAT_DFA
        (\@state_table,
         name => 'fNonRpred',
         starting => $starting,
         accepting_list => \@accepting,
        );
  };
}
use constant::defer fRpred_flat => sub {
  return fNonRpred_flat()->complement
    ->MyFLAT::set_name('fRpred');
};

{
  my @state_table = ([1,2], [1,3], [4,5], [4,6], [7,8], [9,10], [11,12],
                     [7,7], [13,14], [15,14], [16,10], [15,3], [17,7],
                     [7,12], [18,19], [20,3], [20,14], [21,7], [7,22],
                     [23,24], [25,3], [13,7], [26,14], [15,27], [28,24],
                     [25,29], [21,12], [30,19], [20,27], [31,6], [21,22],
                     [21,8]);
  my @accepting = (7);
  my $starting = 0;
  use constant::defer fNonLpred_flat => sub {
    require MyFLAT;
    MyFLAT::aref_to_FLAT_DFA
        (\@state_table,
         name => 'fNonLpred',
         starting => $starting,
         accepting_list => \@accepting,
        );
  };
}
use constant::defer fLpred_flat => sub {
  return fNonLpred_flat()->complement
    ->MyFLAT::set_name('fLpred');
};

# fNonBpred is non-R and non-L both, so intersect
# not minimized
use constant::defer fNonBpred_flat => sub {
  return fNonRpred_flat()->intersect(fNonLpred_flat())->as_dfa
    ->MyFLAT::set_name('fNonBpred');
};
use constant::defer fBpred_flat => sub {
  return fNonBpred_flat()->complement
    ->MyFLAT::set_name('fBpred');
};


#------------------------------------------------------------------------------
# fVisits by adjacent

#                      11
#       *     *     *---->*
#             ^     |
#            1|     |10
#         2   |     v
#       *<----O====>*<----*
#       ^     |     ^   9
#      3|    5|     |8
#       |     v     |
#       *---->*<----*     *
#         4      6  |
#                   |7
#                   v
#       *     *     *     *
our @fVisits_adjacent_list
  = (undef,
     '0,0,1',    # 1
     '0,0,2',
     '-1,-1,1',  # 3
     '-1,-1,0',
     '0,0,3',
     '1,-1,2',   # 6
     '1,-1,3',   # 7
     '1,-1,1',
     '2,0,2',    # 9
     '1,1,3',
     '1,1,0');   # 11
our %fVisits_seg_to_index
  = map { $fVisits_adjacent_list[$_] => $_ } 1 .. $#fVisits_adjacent_list;

sub fVisits_expand_bit {
  my ($s, $bit) = @_;
  my $conf = MyDragonConf->new
    (side => 0,
     segs => [ $fVisits_adjacent_list[$s] ],
     within => $hullstouch_segs_all)
    ->expand($bit);
  # $conf->print;
  my @ret;
  foreach my $seg ($conf->segs_list) {
    my $other = $fVisits_seg_to_index{$seg}
      // die "oops, no index for $seg";
    push @ret, $other;
  }
  return sort {$a<=>$b} @ret;
}

# 0 for non
# 1..11 forward s1 to s11
# 12..22 reverse s1 to s11
sub fVisits_state_num {
  my ($s, $rev) = @_;
  ($s >= 0 && $s <= $#fVisits_adjacent_list) or die "oops";
  return $s + ($s ? $rev * $#fVisits_adjacent_list : 0);
}

# Return a FLAT::NFA.
use constant::defer fVisits_adjacent_NFA => sub {
  require MyFLAT;
  require FLAT::NFA;
  my $f = FLAT::NFA->new->MyFLAT::set_name('fVisits NFA');
  my $last_state = fVisits_state_num($#fVisits_adjacent_list, 1);
  $f->add_states($last_state + 1);
  $f->set_accepting(1 .. $last_state);
  $f->add_transition(0,0, 0);
  $f->add_transition(0,0, 1);
  foreach my $s (1 .. $#fVisits_adjacent_list) {
    foreach my $rev (0,1) {
      my $from_state = fVisits_state_num($s,$rev);
      foreach my $bit (0,1) {
        my @to = fVisits_expand_bit($s, $bit^$rev);
        foreach my $to (@to ? @to : 0) {
          my $to_state = fVisits_state_num($to,$bit);  # new rev = bit
          $f->add_transition($from_state, $to_state, $bit);
        }
      }
    }
  }
  return $f;
};

# Return a FLAT::DFA matching bit strings where segment $s,$rev remains
# always close enough to possibly touch.
sub fVisits_adjacent_DFA {
  my ($s, $rev) = @_;
  my $f = fVisits_adjacent_NFA->clone;
  $f->set_starting(fVisits_state_num($s,$rev));
  return $f->as_dfa
    ->MyFLAT::minimize
    ->MyFLAT::set_name("fVisits " . ($rev?'r':'s') . $s);
}
memoize('fVisits_adjacent_DFA', LIST_CACHE=>'MERGE');

{
  sub fVisits_adjacent_init {
    fBpred_hullstouch_init();
  }
}



#------------------------------------------------------------------------------
# N on centreline by bits

# CentrelinePred_flat() returns a FLAT::DFA which matches bit strings of k
# many bits of n which is on the centreline of curve k start to end.
#
# CentrelinePred_flat('segment') similarly, but matching n which is the
# start of a segment with either start or end on the centreline.
#
#    s3   s2   s1         d2
#      ^   ^   ^           |
#        \ | /             v
#   -------*-------h-------*--------
#        ^ ^ ^           / | \
#      /   |   \       v   v   v
#    e1   e2   e3    a1   a2   a3
#                   =e3        =e1
#      forward          reverse
our @CentrelinePred_state_table = ([0,0],
                                   [4, 7],    # 1 = h start
                                   [1, 10],   # 2 = s1
                                   [2, 11],   # 3 = s2
                                   [3, 11],   # 4 = s3
                                   [11, 9],   # 5 = e1 = a3
                                   [11, 5],   # 6 = e2
                                   [6,  8],   # 7 = e3 = a1
                                   [2,  5],   # 8 = hr
                                   [11, 7],   # 9 = a2
                                   [4, 11],   # 10 = d2
                                   [11,11]);  # 11 = non
our @CentrelinePred_state_to_name = (undef,
                                     'h',  's1','s2','s3', 'e1','e2','e3',
                                     'hr', 'a2','d2',
                                     'non');
{
  my @accepting_segment = (1..10); # all except non
  my @accepting_point = (1,     # h
                         2,3,4, # s1, s2, s3
                         8,     # hr
                         10);   # d2
  sub CentrelinePred_flat {
    my ($type) = @_;
    $type ||= 'point';
    require MyFLAT;
    return MyFLAT::aref_to_FLAT_DFA(\@CentrelinePred_state_table,
                                    name => "N centreline $type",
                                    starting => 1,
                                    accepting_list => ($type eq 'segment'
                                                       ? \@accepting_segment
                                                       : \@accepting_point),
                                   );
  }
  memoize('CentrelinePred_flat', LIST_CACHE=>'MERGE');
}

# Return true if $n is on the centreline of curve $k start to end.
# $type='segment' similarly but a segment which has either segment start or
# end on the centreline.
#
sub CentrelinePred_by_path {
  my ($n, $k, $type) = @_;
  ### CentrelinePred(): "n=$n  k=$k"

  $type ||= 'point';
  my $path = DragonCurve_path();
  my ($n_lo, $n_hi) = $path->level_to_n_range($k);
  if ($n < $n_lo || $n > $n_hi) { return 0; }

  my ($x,$y) = $path->n_to_xy($n);
  ($x,$y) = transform_to_horizontal($x,$y, $k);
  if ($y == 0) { return 1; }

  if ($type eq 'segment') {
    if ($n < $n_lo || $n >= $n_hi) { return 0; }
    ($x,$y) = $path->n_to_xy($n+1);
    ($x,$y) = transform_to_horizontal($x,$y, $k);
    return ($y == 0);
  }
  return 0;
}


#------------------------------------------------------------------------------

# 0*(11+0)+0+
use constant::defer OXpred_flat => sub {
  require FLAT;
  require MyFLAT;
  return FLAT::Regex->new('0*(111*0)(111*0)*00*')->as_dfa->MyFLAT::minimize
    ->MyFLAT::set_name('OXpred');
};


#------------------------------------------------------------------------------

our @MinusInc_LtoH_state_table = ([6, 1],  # 0 = 1    flip
                                  [2, 2],  # 1 = -1-I
                                  [0, 3],  # 2 =  I   flip
                                  [2, 4],  # 3 = -I   flip
                                  [5, 6],  # 4 = -1   flip
                                  [3, 3],  # 5 = 1+I
                                  [6, 6],  # 6 =  0
                                  [0, 0],  # 7 = -1+I
                                  [4, 4],  # 8 =  1-I
                                 );
{
  my @starting = (0,  # dir=0 East
                  5,  # dir=1 NE
                  2,  # dir=2 North
                  7,  # dir=3 NW
                  4,  # dir=4 West
                  1,  # dir=5 SW
                  3,  # dir=6 South
                  8); # dir=7 NW
  # match binary N in k bits which has neighbour in direction $dir8, also
  # within k bits
  sub MinusIncPred_dir8_LtoH_flat {
    my ($dir8) = @_;
    require MyFLAT;
    return MyFLAT::aref_to_FLAT_DFA(\@MinusInc_LtoH_state_table,
                                    name => "MinusIncPred LtoH dir8=$dir8",
                                    starting => $starting[$dir8],
                                    accepting => 6);
  }
}
memoize('MinusIncPred_dir8_LtoH_flat', LIST_CACHE=>'MERGE');


#------------------------------------------------------------------------------

our @MinusNeg_LtoH_state_table = ([0, 1],  # 0 = 0
                                  [2, 4],  # 1 = 1+I
                                  [3, 3],  # 2 = -I  flip
                                  [4, 4],  # 3 =  I  flip
                                  [0, 0],  # 4 =  1  flip
                                  [1, 1],  # 5 = -1  flip

                                  [4, 9],  #  6 = -1+I
                                  [3, 10], #  7 = -1-I
                                  [5, 3],  #  8 =  1-I
                                  [2, 2],  #  9 =  2+I
                                  [8, 8],  # 10 = 1+2*I
                                 );

# match binary N of k bits which is negatable, meaning MinusNeg($n) is also
# within k bits
use constant::defer MinusNegPred_LtoH_flat => sub {
  require MyFLAT;
  return MyFLAT::aref_to_FLAT_DFA(\@MinusNeg_LtoH_state_table,
                                  name => 'MinusNeg LtoH',
                                  accepting => 0)
    ->MyFLAT::minimize; # strip unused states
};
{
  my @starting = (4,  # dir=0 East
                  1,  # dir=1 NE
                  3,  # dir=2 North
                  6,  # dir=3 NW
                  5,  # dir=4 West
                  7,  # dir=5 SW
                  2,  # dir=6 South
                  8); # dir=7 SE
  # match binary N of k bits at location z where the N at -z+$dir8 is also
  # within k bits
  sub MinusNegIncPred_dir8_LtoH_flat {
    my ($dir8) = @_;
    require MyFLAT;
    return MyFLAT::aref_to_FLAT_DFA(\@MinusNeg_LtoH_state_table,
                                    name => "MinusNegIncPred LtoH dir8=$dir8",
                                    starting => $starting[$dir8%8],
                                    accepting => 0)
    ->MyFLAT::minimize; # strip unused states
  }
}
memoize('MinusNegIncPred_dir8_LtoH_flat', LIST_CACHE=>'MERGE');

# match binary N of k bits for which its neighbour in direction $dir8 is
# negatable within k bits
# N itself might be negatable or might not
sub MinusNegNeighbourPred_dir8_LtoH_flat {
  my ($dir8) = @_;
  return MinusIncPred_dir8_LtoH_flat($dir8)
    ->intersect(MinusNegIncPred_dir8_LtoH_flat($dir8+4))
    ->MyFLAT::minimize
    ->MyFLAT::set_name("MinusNegNeighbourPred LtoH dir8=$dir8");
}
memoize('MinusNegNeighbourPred_dir8_LtoH_flat', LIST_CACHE=>'MERGE');


#------------------------------------------------------------------------------

our @MinusRot_state_table = ([0, 3],  # 0 = 0
                                  [6, 3],  # 1 = -1
                                  [4, 0],  # 2 = -I
                                  [0, 2],  # 3 =  1
                                  [3, 5],  # 4 =  I  flip
                                  [1, 0],  # 5 =  1-I
                                  [2, 5],  # 6 =  1+I
                                  [4, 6],  # 7 = -1-I
                                  [3, 9],  # 8 = -1+I
                                  [7, 2],  # 9 = 2
                                 );
our @MinusRot_state_to_dz = ([0, 0],  # 0 = 0
                                  [-1,0],  # 1 = -1
                                  [0,-1],  # 2 = -I
                                  [1, 0],  # 3 =  1
                                  [0, 1],  # 4 =  I
                                  [1,-1],  # 5 =  1-I
                                  [1, 1],  # 6 =  1+I
                                  [-1,-1], # 7 = -1-I
                                  [-1,1],  # 8 = -1+I
                                  [2, 0],  # 9 = 2
                                 );

# accepting bit strings of rotatable N
use constant::defer MinusRotPred_LtoH_flat => sub {
  require MyFLAT;
  return MyFLAT::aref_to_FLAT_DFA(\@MinusRot_state_table,
                                  name => 'MinusRot LtoH',
                                  accepting => 0)
    ->MyFLAT::minimize; # strip unused states
};

our @MinusRotIncPred_LtoH_dir8_to_state = (3,  # dir=0 East
                                           6,  # dir=1 NE
                                           4,  # dir=2 North
                                           8,  # dir=3 NW
                                           1,  # dir=4 West
                                           7,  # dir=5 SW
                                           2,  # dir=6 South
                                           5); # dir=7 SE
sub MinusRotIncPred_dir8_LtoH_flat {
  my ($dir8) = @_;
  require MyFLAT;
  return MyFLAT::aref_to_FLAT_DFA
    (\@MinusRot_state_table,
     name => "MinusRotIncPred LtoH dir8=$dir8",
     starting => $MinusRotIncPred_LtoH_dir8_to_state[$dir8%8],
     accepting => 0)
    ->MyFLAT::minimize; # strip unused states
}
memoize('MinusRotIncPred_dir8_LtoH_flat', LIST_CACHE=>'MERGE');

sub MinusRotNeighbourPred_dir8_LtoH_flat {
  my ($dir8) = @_;
  return MinusIncPred_dir8_LtoH_flat($dir8)
    ->intersect(MinusRotIncPred_dir8_LtoH_flat($dir8+2))
    ->MyFLAT::minimize
    ->MyFLAT::set_name("MinusRotNeighbourPred LtoH dir8=$dir8");
}
memoize('MinusRotNeighbourPred_dir8_LtoH_flat', LIST_CACHE=>'MERGE');

#------------------------------------------------------------------------------

# same as MinusRot but odd states 1,-1,i,-i destinations are 0<->1 swapped
our @MinusUnrot_state_table = ([0, 3],  # 0 = 0
                               [3, 6],  # 1 = -1
                               [0, 4],  # 2 = -I
                               [2, 0],  # 3 =  1
                               [5, 3],  # 4 =  I  flip
                               [1, 0],  # 5 =  1-I
                               [2, 5],  # 6 =  1+I
                               [4, 6],  # 7 = -1-I
                               [3, 9],  # 8 = -1+I
                               [7, 2],  # 9 = 2
                              );

# accepting bit strings of rotatable N
use constant::defer MinusUnrotPred_LtoH_flat => sub {
  require MyFLAT;
  return MyFLAT::aref_to_FLAT_DFA(\@MinusUnrot_state_table,
                                  name => 'MinusUnrot LtoH',
                                  accepting => 0)
    ->MyFLAT::minimize; # strip unused states
};

#------------------------------------------------------------------------------

our @MinusDouble_LtoH_state_table = ([ 0,  7],  #  0 = 0
                                     [ 0,  7],  #  1 = 1
                                     [ 1,  4],  #  2 = I
                                     [ 5,  0],  #  3 = -1
                                     [ 2,  3],  #  4 = -I
                                     [ 4, 13],  #  5 = 1+I
                                     [ 1,  4],  #  6 = -1+I
                                     [ 2,  3],  #  7 = -1-I
                                     [ 3, 14],  #  8 =  1-I
                                     [ 8, 11],  #  9 = 2i
                                     [ 5,  0],  # 10 = -2
                                     [ 6, 10],  # 11 = -2i
                                     [ 8, 11],  # 12 =  1 + 2*I
                                     [ 9,  6],  # 13 = -1 - 2*I
                                     [12,  2],  # 14 = -2 - I

                                     [ 7,18],  # 15   2
                                     [11,20],  # 16   2+2i
                                     [15, 8],  # 17  -2+2i
                                     [ 9, 6],  # 18  -2-2i
                                     [10,21],  # 19   2-2i
                                     [22,23],  # 20  -1-3i
                                     [12, 2],  # 21  -3-I
                                     [15, 8],  # 22  -1+2*I
                                     [24, 1],  # 23  -2+I
                                     [ 4,13],  # 24   2+I
                                    );

# accepting bit strings of doublable N
use constant::defer MinusDoublePred_LtoH_flat => sub {
  require MyFLAT;
  return MyFLAT::aref_to_FLAT_DFA(\@MinusDouble_LtoH_state_table,
                                  name => 'MinusDouble LtoH',
                                  accepting => 0)
    ->MyFLAT::minimize; # strip unused states
};
{
  my @starting = (15,  # dir=0 East
                  16,  # dir=1 NE
                  9,   # dir=2 North
                  17,  # dir=3 NW
                  10,  # dir=4 West
                  18,  # dir=5 SW
                  11,  # dir=6 South
                  19); # dir=7 SE
  sub MinusDoubleIncPred_dir8_LtoH_flat {
    my ($dir8) = @_;
    require MyFLAT;
    return MyFLAT::aref_to_FLAT_DFA(\@MinusDouble_LtoH_state_table,
                                    name => "MinusDoubleIncPred LtoH dir8=$dir8",
                                    starting => $starting[$dir8%8],
                                    accepting => 0)
      ->MyFLAT::minimize; # strip unused states
  }
}
memoize('MinusDoubleIncPred_dir8_LtoH_flat', LIST_CACHE=>'MERGE');

sub MinusDoubleNeighbourPred_dir8_LtoH_flat {
  my ($dir8) = @_;
  return MinusIncPred_dir8_LtoH_flat($dir8)
    ->intersect(MinusDoubleIncPred_dir8_LtoH_flat($dir8))
    ->MyFLAT::minimize
    ->MyFLAT::set_name("MinusDoubleNeighbourPred LtoH dir8=$dir8");
}
memoize('MinusDoubleNeighbourPred_dir8_LtoH_flat', LIST_CACHE=>'MERGE');


my %type_to_pred_info
  = (Complex     => { pred_func     => \&FLAT_any_binary,
                      inc_pred_func => \&MinusIncPred_dir8_LtoH_flat,
                      rot8          => 0,
                      A_func        => sub { 2**$_[0] },
                      B_func        => \&ComplexBoundary,
                    },
     MinusNeg    => { pred_func     => \&MinusNegPred_LtoH_flat,
                      inc_pred_func => \&MinusNegIncPred_dir8_LtoH_flat,
                      rot8          => 4,
                      A_func        => \&MinusNegA,
                      B_func        => \&MinusNegB,
                    },
     MinusRot    => { pred_func     => \&MinusRotPred_LtoH_flat,
                      inc_pred_func => \&MinusRotIncPred_dir8_LtoH_flat,
                      rot8          => 2,
                      A_func        => \&MinusRotA,
                      B_func        => \&MinusRotB,
                    },
     MinusUnrot   => { pred_func    => \&MinusUnrotPred_LtoH_flat,
                      inc_pred_func => undef,
                      unrot8        => -2,
                      A_func        => undef,
                      B_func        => undef,
                    },
     MinusDouble => { pred_func     => \&MinusDoublePred_LtoH_flat,
                      inc_pred_func => \&MinusDoubleIncPred_dir8_LtoH_flat,
                      rot8          => 0,
                      A_func        => \&MinusDoubleA,
                      B_func        => \&MinusDoubleB,
                    },
    );

sub Pred {
  my ($type) = @_;
  return $type_to_pred_info{$type}->{'pred_func'}->();
}

sub NeighbourPred {
  my ($type, $dir8) = @_;
  my $info = $type_to_pred_info{$type};
  return MinusIncPred_dir8_LtoH_flat($dir8)
    ->intersect($info->{'inc_pred_func'}->($dir8 + $info->{'rot8'}))
    ->MyFLAT::minimize
    ->MyFLAT::set_name("${type}NeighbourPred LtoH dir8=$dir8");
}
memoize('NeighbourPred', LIST_CACHE=>'MERGE');

sub BoundaryPred {
  my ($type,$dir8) = @_;
  my $info = $type_to_pred_info{$type};
  return $info->{'pred_func'}->()
    ->intersect(NeighbourPred($type,$dir8)->complement)
    ->MyFLAT::minimize
    ->MyFLAT::set_name("${type} BoundaryPred LtoH dir8=$dir8");
}
memoize('BoundaryPred', LIST_CACHE=>'MERGE');

# N LtoH which have none of NSEW
sub NonNSEWpred {
  my ($type) = @_;
  my $info = $type_to_pred_info{$type};
  my $f = $info->{'pred_func'}->();
  foreach my $dir8 (0,2,4,6) {
    $f = $f->intersect(NeighbourPred($type,$dir8)->complement)
      ->MyFLAT::minimize(verbose=>1);
  }
  return $f->MyFLAT::set_name("${type} NonNSEWpred LtoH");
}
memoize('NonNSEWpred', LIST_CACHE=>'MERGE');

sub HangingSquarePred_dir8 {
  my ($type, $dir8) = @_;
  my $f = NonNSEWpred($type)
    ->intersect(NeighbourPred($type,$dir8))
    ->MyFLAT::minimize(verbose=>1);
  foreach my $non_dir8 (1,3,5,7) {
    if ($non_dir8 != $dir8) {
      $f = $f->intersect(NeighbourPred($type,$non_dir8)->complement)
        ->MyFLAT::minimize(verbose=>1);
    }
  }
  return $f->MyFLAT::set_name("${type} HangingSquarePred dir8=$dir8 LtoH");
}
memoize('HangingSquarePred_dir8', LIST_CACHE=>'MERGE');

sub HangingSquarePred {
  my ($type) = @_;
  HangingSquarePred_dir8($type,1)
    ->union(HangingSquarePred_dir8($type,1))->MyFLAT::minimize(verbose=>1)
    ->union(HangingSquarePred_dir8($type,1))->MyFLAT::minimize(verbose=>1)
    ->union(HangingSquarePred_dir8($type,1))->MyFLAT::minimize(verbose=>1)
    ->MyFLAT::set_name("${type} HangingSquarePred LtoH");
}
memoize('HangingSquarePred', LIST_CACHE=>'MERGE');

#------------------------------------------------------------------------------

sub TA_complex_c {
  my ($k) = @_;
  return ($k == 0 ? 0 : 2**($k-1));

  # {
  #   # left of last segment in side of twindragon
  #   my $n = $n_hi/2 - 1;
  #   my ($x,$y) = $path->n_to_xy($n);
  #   my ($dx,$dy) = $path->n_to_dxdy($n);
  #   my ($lx,$ly) = xy_rotate_plus90($dx,$dy);
  #   my $sqx = 2*$x + $dx + $lx;
  #   my $sqy = 2*$y + $dy + $ly;
  #   $Cv = "$sqx,$sqy";
  # }
}


#------------------------------------------------------------------------------

our @TAparentDir_state_table = ([0,0 ],
                                [ 1, 2 ],   # 1=upper
                                [ 3, 2 ],   # 2=left
                                [ 4, 3 ],   # 3=lower
                                [ 4, 1 ]);  # 4=right

# FLAT::DFA which matches binary n, high to low, which is twindragon area
# tree with tree parent vertex in direction 0 (right)
use constant::defer TAparentDir0_flat => sub {
  require MyFLAT;
  MyFLAT::aref_to_FLAT_DFA (\@TAparentDir_state_table,
                            name => 'TAparentDir0',
                            starting => 1,
                            accepting => 4);
};
use constant::defer TAparentDir1_flat => sub {
  require MyFLAT;
  MyFLAT::aref_to_FLAT_DFA (\@TAparentDir_state_table,
                            name => 'TAparentDir1',
                            starting => 1,
                            accepting => 1);
};
use constant::defer TAparentDir2_flat => sub {
  require MyFLAT;
  MyFLAT::aref_to_FLAT_DFA (\@TAparentDir_state_table,
                            name => 'TAparentDir2',
                            starting => 1,
                            accepting => 2);
};
use constant::defer TAparentDir3_flat => sub {
  require MyFLAT;
  MyFLAT::aref_to_FLAT_DFA (\@TAparentDir_state_table,
                            name => 'TAparentDir3',
                            starting => 1,
                            accepting => 3);
};

sub TAparentDir_transitions_n {
  my ($state, $n) = @_;
  foreach my $bit (reverse bit_split_lowtohigh($n)) {  # high to low
    $state = $TAparentDir_state_table[$state]->[$bit];
  }
  return $state % 4;
}

sub TAstepDir {
  my ($n,$dir) = @_;
  $dir %= 4;
  return ($dir==0   ? TAVertexToRight($n)
          : $dir==1 ? TAVertexToUpper($n)
          : $dir==2 ? TAVertexToLeft($n)
          : $dir==3 ? TAVertexToLower($n)
          : die);
}

sub TAparentDir {
  my ($n) = @_;
  $n >= 1 or croak "TAparentDir() is for n>=1";
  return TAparentDir_transitions_n(1,$n) % 4;
}
sub TAparent {
  my ($n) = @_;
  $n >= 1 or croak "TAparent() is for n>=1";
  return TAstepDir($n, TAparentDir($n));
}

sub TAtospineDir {
  my ($n) = @_;
  $n >= 0 or croak "TAtospineDir() is for n>=1";
  return TAparentDir_transitions_n(4,$n) % 4;
}
sub TAtospine {
  my ($n) = @_;
  $n >= 0 or croak "TAtospine() is for n>=1";
  return TAstepDir($n, TAtospineDir($n));
}


#------------------------------------------------------------------------------

our @EpredL_state_table = ([0,0 ],
                           [ 1,5 ],   # 1 ls0
                           [ 2,6 ],   # 2 ls12
                           [ 2,7 ],   # 3 ls3
                           [ 1,4 ],   # 4 ls0rev
                           [ 2,4 ],   # 5 ls1rev
                           [ 3,5 ],   # 6 ls2rev
                           [ 3,7 ],   # 7 ls3rev
                          );

use constant::defer EpredL_flat => sub {
  require MyFLAT;
  MyFLAT::aref_to_FLAT_DFA (\@EpredL_state_table,
                            name => 'EpredL',
                            starting => 1,
                            accepting_list => [3,7]);
};
sub EpredL_LtoH_flat {
  my ($n) = @_;
  return EpredL_flat()->MyFLAT::reverse
    ->MyFLAT::set_name('EpredL LtoH');
}

use constant::defer EpredR_flat => sub {
  require MyFLAT;
  return DpredSecondL_flat()
    ->MyFLAT::digits_increment(add => -1)
    ->MyFLAT::optional_leading_0s
    ->MyFLAT::minimize
    ->MyFLAT::set_name('EpredR');
};
use constant::defer EpredR_LtoH_flat => sub {
  require MyFLAT;
  return EpredR_flat()->MyFLAT::reverse
    ->MyFLAT::minimize
    ->MyFLAT::set_name('EpredR LtoH');
};

use constant::defer Epred_flat => sub {
  require MyFLAT;
  return EpredR_flat()->union(EpredL_flat())
    ->MyFLAT::minimize
    ->MyFLAT::set_name('Epred');
};
use constant::defer Epred_LtoH_flat => sub {
  require MyFLAT;
  return Epred_flat()->MyFLAT::reverse
    ->MyFLAT::minimize
    ->MyFLAT::set_name('Epred LtoH');
};

sub Epred {
  my ($n) = @_;
  return DpredSecond($n+1);
}
sub EpredL {
  my ($n) = @_;
  return Epred($n) && turn($n+1)==-1;
}
sub EpredR {
  my ($n) = @_;
  return Epred($n) && turn($n+1)==1;
}

#------------------------------------------------------------------------------

sub MidSN {
  my ($k) = @_;
  return 2**($k-1) - BridgeLother($k-1);
}
sub MidEN {
  my ($k) = @_;
  return 3*2**($k-1) - BridgeRother($k-1);
}


#------------------------------------------------------------------------------
# Fixed Point

sub ffixed_u_parent {
  my ($m) = @_;
  my $r = $m % 3;
  my $h = ($m-$r)/3;
  if ($r == 0) { return 2*$h; }
  if ($r == 1) { return 2*$h+1; }
  return 2*$h+1;
}
sub ffixed_uDepth {
  my ($m) = @_;
  my $depth = 0;
  while ($m > 1) {
    $m = ffixed_u_parent($m);
    $depth++;
  }
  return $depth;
}
sub ffixed_u {
  my ($m) = @_;
  if ($m == 0) { return $m*0; }
  if ($m == 1) { return $m*0 + 4; }
  return 2*ffixed_u(ffixed_u_parent($m)) - ($m % 3 == 1 ? 4 : 0);
}

sub ffixed_du_parent {
  my ($m) = @_;
  my $r = $m % 3;
  my $h = ($m-$r)/3;
  if ($r == 0) { return 2*$h; }
  if ($r == 1) { return undef; }
  return 2*$h+1;
}
sub ffixed_du {
  my ($m) = @_;
  if ($m <= 1) { return $m*0 + 4; }
  my $r = $m % 3;
  my $h = ($m-$r)/3;
  if ($r == 0) { return 2*ffixed_du(2*$h) - 4; }
  if ($r == 1) { return $m*0 + 4; }
  return 2*ffixed_du(2*$h+1);
}
sub ffixed_row_w {
  my ($k) = @_;
  my $w = 1;
  foreach (1..$k) {
    $w += ($w>>1) + ($w&1);
  }
  return $w;
}

sub ffixed_string {
  my %options = @_;
  my $num_bits = $options{'num_bits'} || 0;
  my $max_m    = $options{'max_m'}    || 0;
  my $num_runs = $options{'num_runs'} || 0;
  ### $num_bits
  ### $max_m
  ### $num_runs

  my $str = '';
  my $m = 1;
  my $prev_u = 0;
  while (length($str) < $num_bits
         || $m <= $max_m
         || $m < 2*$num_runs) {
    my $u = ffixed_u($m);
    my $du = $u - $prev_u;
    ### at: "m=$m prev_u=$prev_u u=$u du=$du"
    ### $str

    if ($m & 1) {
      $str .= ('1' x ($du-2)).'01';
    } else {
      $str .= '0' x $du;
    }
    $m++;
    $prev_u = $u;
  }
  ### end: $str
  if ($m > $max_m+1 && $m > 2*$num_runs) {
    $str = substr($str,0,$num_bits);
    ### truncate: $str
  }
  return $str;
}


#------------------------------------------------------------------------------

# Local variables:
# compile-command: "perl -w MyDragon.pm"
# End:

1;
__END__
